Package: enrichwith
Title: Methods to Enrich R Objects with Extra Components
Version: 0.5.0
Authors@R: person("Ioannis", "Kosmidis", email = "ioannis.kosmidis@warwick.ac.uk", role = c("aut", "cre"))
Description: Provides the "enrich()" method for augmenting list-like R objects with additional, model-specific components. Methods are currently available for objects of class "family", "link-glm", "lm", "glm", and "betareg". Enriched objects retain their original class and remain compatible with existing methods. For example, enriching a "glm" object produces an "enriched_glm" object that also inherits from "glm". In addition to the standard components, the "enriched_glm" object includes methods for simulation and functions to compute scores, observed and expected information matrices, first-order bias, and other model quantities such as densities, probabilities, and quantiles, which can be evaluated at use-supplied parameter values. The package also provides tools for generating customizable source code templates for the structured implementation of methods to compute new components and enrich arbitrary objects.
Depends: R (>= 3.0.0)
URL: https://github.com/ikosmidis/enrichwith
BugReports: https://github.com/ikosmidis/enrichwith/issues
License: GPL-3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
Suggests: whisker, SuppDists, brglm, brglm2, ggplot2, MASS, knitr,
        rmarkdown, testthat, numDeriv
Enhances: betareg, gnm, stats
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-29 12:57:27 UTC; yiannis
Author: Ioannis Kosmidis [aut, cre]
Maintainer: Ioannis Kosmidis <ioannis.kosmidis@warwick.ac.uk>
Repository: CRAN
Date/Publication: 2026-04-29 15:30:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(coef,enriched_glm)
S3method(coef,enriched_lm)
S3method(enrich,"link-glm")
S3method(enrich,betareg)
S3method(enrich,family)
S3method(enrich,glm)
S3method(enrich,lm)
S3method(get_auxiliary_functions,betareg)
S3method(get_auxiliary_functions,glm)
S3method(get_auxiliary_functions,lm)
S3method(get_bias_function,betareg)
S3method(get_bias_function,glm)
S3method(get_bias_function,lm)
S3method(get_dmodel_function,glm)
S3method(get_enrichment_options,"link-glm")
S3method(get_enrichment_options,betareg)
S3method(get_enrichment_options,family)
S3method(get_enrichment_options,glm)
S3method(get_enrichment_options,lm)
S3method(get_information_function,betareg)
S3method(get_information_function,glm)
S3method(get_information_function,lm)
S3method(get_pmodel_function,glm)
S3method(get_qmodel_function,glm)
S3method(get_score_function,betareg)
S3method(get_score_function,glm)
S3method(get_score_function,lm)
S3method(get_simulate_function,betareg)
S3method(get_simulate_function,glm)
S3method(get_simulate_function,lm)
S3method(print,enrichment_options)
export(create_enrichwith_skeleton)
export(enrich)
export(enriched_glm)
export(get_auxiliary_functions)
export(get_bias_function)
export(get_dmodel_function)
export(get_enrichment_options)
export(get_information_function)
export(get_pmodel_function)
export(get_qmodel_function)
export(get_score_function)
export(get_simulate_function)
import(stats)

[![CRAN_Status_Badge](http://www.r-pkg.org/badges/version/enrichwith)](https://cran.r-project.org/package=enrichwith)
[![R-CMD-check](https://github.com/ikosmidis/enrichwith/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/ikosmidis/enrichwith/actions/workflows/R-CMD-check.yaml)
[![Codecov test coverage](https://codecov.io/gh/ikosmidis/enrichwith/graph/badge.svg)](https://app.codecov.io/gh/ikosmidis/enrichwith)
[![Licence](https://img.shields.io/badge/licence-GPL--3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0.en.html)
[![Contributor
Covenant](https://img.shields.io/badge/Contributor%20Covenant-2.1-4baaaa.svg)](https://contributor-covenant.org/version/2/1/CODE_OF_CONDUCT.html)

# enrichwith

Methods to enrich list-like R objects with extra components

## Rationale

When I develop a piece of statistical methodology it is not uncommon
that it depends on functionality or quantities that a list-like core R
object `object_x` of a certain class could have but doesn't. An
example is objects of class `link-glm` which only provide up to the
the first derivatives of the link function. It is also not uncommon
that the functionality or quantities that I need are hard-coded in
another package in the R ecosystem.  Prominent examples include
implementations of gradients of the log-likelihood and information
matrices for specific model classes.

In such cases, I either contact the developers and ask them to provide
a generic which I can then use, or copy some of the code and adopt it
to what I'm doing. Both options can be rather time-consuming, and
particularly the latter is rarely bug-free.

I believe that

* users and developers should have direct access to useful
  functionality and quantities in the R ecosystem, especially if these include implementations of complex statistical quantities

* quantities and functionality that are specific to a list-like `object_x` should be components of `object_x`

The above motivated me to develop the **enrichwith** R package that
allows `object_x` to be enriched with components corresponding to the
option `enrichment_option` through the following simple call

``` r
enrich(object_x, with = enrichment_option)
```

The call is inspired by [Donald
Knuth's](https://en.wikipedia.org/wiki/Donald_Knuth) [literate
programming](https://en.wikipedia.org/wiki/Literate_programming)
paradigm.

## Purpose and objective

The main objective of **enrichwith** is to allow users and developers
to directly use the enrichment options that other developers have
provided, through a *clean interface*, minimizing the need to adopt
source code of others.

The purpose of **enrichwith** is to provide:

* *useful enrichment options* for core R objects, including objects of
  class `lm`, `glm`, `link-glm` and `family` (see, for example,
  `?enrich.glm`)

* methods for producing *customizable source code templates* for the
  structured implementation of methods to compute new components (see
  `?enrichwith` and `?create_enrichwith_skeleton`)

* generic methods for the *easy enrichment* of the object with those
  components (see, for example, `?enrich` and
  `?get_enrichment_options`)

## Vignettes

The vignettes illustrate the **enrichwith** functionality through
comprehensive, step-by-step case studies. These also include
illustrations from recent research of mine on methods for statistical
learning and inference.

## Installation
Get the development version from GitHub with

``` r
# install.packages("devtools")
devtools::install_github("ikosmidis/enrichwith")
```

## Examples

### `link-glm` objects

Objects of class `link-glm` have as components functions to compute
the link function (`linkfun`), the inverse link function (`linkinv`),
and the 1st derivative of the inverse link function (`mu.eta`).

**enrichwith** comes with a built-in template with the methods for the
enrichment of `link-glm` objects with the 2nd and 3rd derivatives of
the inverse link function.

The `get_enrichment_options` method can be used to check what
enrichment options are available for objects of class `link-glm`.
``` r
library("enrichwith")
standard_link <- make.link("probit")
class(standard_link)
# [1] "link-glm"
get_enrichment_options(standard_link)
# -------
# Option: d2mu.deta
# Description: 2nd derivative of the inverse link function
#   component  compute_function
# 1 d2mu.deta compute_d2mu.deta
# -------
# Option: d3mu.deta
# Description: 3rd derivative of the inverse link function
#   component  compute_function
# 1 d3mu.deta compute_d3mu.deta
# -------
# Option: inverse link derivatives
# Description: 2nd and 3rd derivative of the inverse link function
#   component  compute_function
# 1 d2mu.deta compute_d2mu.deta
# 2 d3mu.deta compute_d3mu.deta
# -------
# Option: all
# Description: all available options
#   component  compute_function
# 1 d2mu.deta compute_d2mu.deta
# 2 d3mu.deta compute_d3mu.deta
```

According to the result of `get_enrichment_options`, the object
`standard_link` can be enriched with the 2nd and 3rd derivative of the
inverse link function through the option "inverse link derivatives".
``` r
enriched_link <- enrich(standard_link, with = "inverse link derivatives")
class(enriched_link)
# [1] "enriched_link-glm" "link-glm"
cat(format(enriched_link$d2mu.deta), sep = "\n")
# function (eta)
# {
#     -eta * pmax(dnorm(eta), .Machine$double.eps)
# }
cat(format(enriched_link$d3mu.deta), sep = "\n")
# function (eta)
# {
#     (eta^2 - 1) * pmax(dnorm(eta), .Machine$double.eps)
# }
```
`enriched_link` is now an "enriched" `link-glm` object, which, as per
the enrichment options above, has the extra components `d2mu.deta` and
`d3mu.deta`, for the calculation of 2nd and 3rd derivatives of the
inverse link function with respect to `eta`, respectively.

### `family` objects

![](https://raw.githubusercontent.com/ikosmidis/enrichwith/master/inst/enrich_family.gif)

N <- 1000

## Poisson
poisson_family <- poisson()
y <- rpois(N, 100)
m <- rpois(N, 100)/N
mu <- y + rnorm(N, 0, 2)
all.equal(2 * m * (y * log(y) -  y - y * log(mu) + mu ),
          poisson_family$dev.resids(y, mu, m),
          tolerance = 1e-08)

## Normal
gaussian_family <- gaussian()
y <- rnorm(N, 10, 4)
m <- rpois(N, 100)
mu <- rnorm(N, 0, 2)

all.equal(m * (y - mu)^2,
          gaussian_family$dev.resids(y, mu, m))


## Binomial
binomial_family <- binomial()
mu <- rbeta(100, 0.9, 0.7)
z <- 1:100
m <- z + rpois(100, 20)
y <- z/m

all.equal(- 2 * m * ((y * log(mu/(1 - mu)) + log(1 - mu)) -
                 (y * log(y/(1 - y)) + log(1 - y))),
          binomial_family$dev.resids(y, mu, m))

## If y = 0 then dev.resids uses -2 * m * log(1 - mu)
all.equal(binomial_family$dev.resids(0, 0.2, 10),
          -2 * 10 * log(1 - 0.2))

## If y = m then dev.resids uses -2 * m * log(mu)
all.equal(-2 * m * log(mu),
          binomial_family$dev.resids(m/m, mu, m))


## Gamma
gamma_family <- Gamma()
mu <- rpois(N, 10)
m <- rpois(N, 100)

mu <- 14
phi <- 0.3333
nu <- 1/phi
rate <- nu/mu

## The actual variance is mu^2/nu
## mu: mean(rgamma(100000, shape = nu, rate = rate))
## var: var(rgamma(100000, shape = nu, rate = rate))

## For gamma theta = -1/mu,
all.equal(-2 * m * (log(y/mu) - (y - mu)/mu),
          Gamma()$dev.resids(y, mu, m))

## ----echo = TRUE, eval = TRUE-------------------------------------------------
clotting <- data.frame(conc = c(5,10,15,20,30,40,60,80,100,
                                5,10,15,20,30,40,60,80,100),
                       time = c(118, 58, 42, 35, 27, 25, 21, 19, 18,
                                69, 35, 26, 21, 18, 16, 13, 12, 12),
                       lot = factor(c(rep(1, 9), rep(2, 9))))

## ----echo = TRUE, eval = TRUE-------------------------------------------------
library("ggplot2")
clottingML <- glm(time ~ log(conc) * lot, family = Gamma, data = clotting)
alpha <- 0.01
pr_out <- predict(clottingML, type = "response", se.fit = TRUE)
new_data <- clotting
new_data$time <- pr_out$fit
new_data$type <- "fitted"
clotting$type <- "observed"
all_data <- rbind(clotting, new_data)
new_data <- within(new_data, {
    low <- pr_out$fit - qnorm(1 - alpha/2) * pr_out$se.fit
    upp <- pr_out$fit + qnorm(1 - alpha/2) * pr_out$se.fit
}) |> merge(all_data, all = TRUE)
ggplot(all_data) +
    geom_point(aes(conc, time, col = type), alpha = 0.8) +
    geom_segment(data = new_data, aes(x = conc, y = low, xend = conc, yend = upp, col = type)) +
    facet_grid(. ~ lot) +
    theme_bw() +
    theme(legend.position = "top")

## ----echo = TRUE, eval = TRUE-------------------------------------------------
library("enrichwith")
enriched_clottingML <- enrich(clottingML, with = "auxiliary functions")
scores_clottingML <- enriched_clottingML$auxiliary_functions$score

## ----echo = TRUE, eval = TRUE-------------------------------------------------
scores_clottingML <- get_score_function(clottingML)

## ----echo = TRUE, eval = TRUE-------------------------------------------------
scores_clottingML()

## ----echo = TRUE, eval = TRUE-------------------------------------------------
info_clottingML <- enriched_clottingML$auxiliary_functions$information

## ----echo = TRUE, eval = TRUE-------------------------------------------------
info_clottingML <- get_information_function(clottingML)

## ----echo = TRUE, eval = TRUE-------------------------------------------------
summary_clottingML <- summary(clottingML)

## ----echo = TRUE, eval = TRUE-------------------------------------------------
summary_std_errors <- coef(summary_clottingML)[, "Std. Error"]
einfo <- info_clottingML(dispersion = summary_clottingML$dispersion)
all.equal(sqrt(diag(solve(einfo)))[1:4], summary_std_errors, tolerance = 1e-05)

## ----echo = TRUE, eval = TRUE-------------------------------------------------
oinfo <- info_clottingML(dispersion = summary_clottingML$dispersion, type = "observed")
all.equal(oinfo[1:4, 1:4], einfo[1:4, 1:4])

## ----echo = TRUE, eval = TRUE-------------------------------------------------
clottingML_log <- update(clottingML, family = Gamma("log"))
summary_clottingML_log <- summary(clottingML_log)
info_clottingML_log <- get_information_function(clottingML_log)
einfo_log <- info_clottingML_log(dispersion = summary_clottingML_log$dispersion, type = "expected")
oinfo_log <- info_clottingML_log(dispersion = summary_clottingML_log$dispersion, type = "observed")
round(einfo_log, 3)
round(oinfo_log, 3)

## ----echo = TRUE, eval = TRUE-------------------------------------------------
clottingML_nested <- update(clottingML, . ~ log(conc))
enriched_clottingML_nested <- enrich(clottingML_nested, with = "mle of dispersion")
coef_full <- coef(clottingML)
coef_hypothesis <- structure(rep(0, length(coef_full)), names = names(coef_full))
coef_hypothesis_short <- coef(enriched_clottingML_nested, model = "mean")
coef_hypothesis[names(coef_hypothesis_short)] <- coef_hypothesis_short
disp_hypothesis <- coef(enriched_clottingML_nested, model = "dispersion")
scores <- scores_clottingML(coef_hypothesis, disp_hypothesis)
info <- info_clottingML(coef_hypothesis, disp_hypothesis)

## ----echo = TRUE, eval = TRUE-------------------------------------------------
(score_statistic <- drop(scores%*%solve(info)%*%scores))

## ----echo = TRUE, eval = TRUE-------------------------------------------------
pchisq(score_statistic, 2, lower.tail = FALSE)

## ----echo = TRUE, eval = TRUE-------------------------------------------------
coef_full[3:4]%*%solve(solve(info)[3:4, 3:4])%*%coef_full[3:4]

## ----echo = TRUE, eval = TRUE-------------------------------------------------
(deviance(clottingML_nested) - deviance(clottingML))/disp_hypothesis

## ----echo = TRUE, eval = TRUE-------------------------------------------------
simulate_clottingML <- get_simulate_function(clottingML)
simulate_clottingML(nsim = 3, seed = 123)

## ----echo = TRUE, eval = TRUE-------------------------------------------------
simulate(clottingML, nsim = 3, seed = 123)

## ----echo = TRUE, eval = TRUE-------------------------------------------------
coefficients <- c(0, 0.01, 0, 0.01)
dispersion <- 0.001
samples <- simulate_clottingML(coefficients = coefficients, dispersion = dispersion, nsim = 500000, seed = 123)

## ----echo = TRUE, eval = TRUE-------------------------------------------------
means <- 1/(model.matrix(clottingML) %*% coefficients)
variances <- dispersion * means^2
max(abs(rowMeans(samples) - means))
max(abs(apply(samples, 1, var) - variances))

## ----echo = TRUE, eval = TRUE-------------------------------------------------
cML_dmodel <- get_dmodel_function(clottingML)
cML_dmodel()

## ----echo = TRUE, eval = TRUE-------------------------------------------------
cML_dmodel(coefficients = c(-0.01, 0.02, -0.01, 0), dispersion = 0.1)

## ----echo = TRUE, eval = TRUE-------------------------------------------------
new_data <- data.frame(conc = 5:10, time = 50:45, lot = factor(c(1, 1, 1, 2, 2, 2)))
cML_dmodel(data = new_data, coefficients = c(-0.01, 0.02, -0.01, 0), dispersion = 0.1)

## ----echo = TRUE, eval = TRUE-------------------------------------------------
cML_qmodel <- get_qmodel_function(clottingML)
cML_pmodel <- get_pmodel_function(clottingML)
probs <- cML_pmodel(data = new_data, coefficients = c(-0.01, 0.02, -0.01, 0), dispersion = 0.1)
cML_qmodel(probs, data = new_data, coefficients = c(-0.01, 0.02, -0.01, 0), dispersion = 0.1)

## ----echo = TRUE, eval = TRUE-------------------------------------------------
new_data <- expand.grid(conc = c(15, 40),
                        lot = factor(1:2),
                        time = seq(0, 50, length = 500))
new_data$density <- cML_dmodel(new_data)
ggplot(data = new_data) + 
     geom_line(aes(time, density)) + facet_grid(conc ~ lot) +
     theme_bw() +
     geom_vline(data = clotting[c(3, 6, 12, 15), ], aes(xintercept = time), lty = 2)

## ----echo = TRUE, eval = TRUE-------------------------------------------------
enriched_clottingML <- enriched_glm(time ~ log(conc) * lot, family = Gamma, data = clotting)
names(enriched_clottingML$auxiliary_functions)
enriched_clottingML$score_mle
enriched_clottingML$expected_information_mle
enriched_clottingML$observed_information_mle
enriched_clottingML$bias_mle
enriched_clottingML$dispersion_mle

#' Enrich objects of class {{class}}
#'
#' Enrich objects of class {{class}} with
#' *BRIEFLY EXPLAIN WITH WHAT*
#'
#'
#' @param object an object of class {{class}}
#' @param with a character vector of options for the enrichment of \code{object}
#' @param ... extra arguments to be passed to the
#'     \code{compute_*} functions
#'
#' @details
#' *ADD DETAILS IF NECESSARY*
#'
#' @return The object \code{object} of class {{class}} with extra
#'     components. \code{get_enrichment_options.{{class}}()} returns
#'     the components and their descriptions.
#'
#' @export
#' @examples
#' ## *ADD AN EXAMPLE*
`enrich.{{class}}` <- function(object, with = "all", ...) {
    if (is.null(with)) {
        return(object)
    }
    dots <- list(...)
    enrichment_options <- get_enrichment_options(object, option = with)
    component <- unlist(enrichment_options$component)
    compute <- unlist(enrichment_options$compute_function)
    for (j in seq.int(length(component))) {
        ccall <- call(compute[j], object = object)
        for (nam in names(dots)) {
            ccall[[nam]] <- dots[[nam]]
        }
        object[[component[j]]] <- eval(ccall)
    }
    if (is.null(attr(object, "enriched"))) {
        attr(object, "enriched") <- TRUE
        classes <- class(object)
        class(object) <- c(paste0("enriched_", classes[1]), classes)
    }
    object
}

#' Available options for the enrichment objects of class {{class}}
#'
#' @param object the object to be enriched
#' @param option a character vector listing the options for enriching
#'     the object
#' @param all_options if \code{TRUE} then output a data frame with the
#'     available enrichment options, their descriptions, the names of
#'     the components that each option results in, and the names of
#'     the corresponding \code{compute_*} functions.
#' @return an object of class \code{enrichment_options}
#'
#' @details A check is being made whether the requested option is
#'     available. No check is being made on whether the functions that
#'     produce the components exist.
#' @examples
#' \dontrun{
#' get_enrichment_options.{{class}}(option = "all")
#' get_enrichment_options.{{class}}(all_options = TRUE)
#' }
#' @export
`get_enrichment_options.{{class}}` <- function(object, option, all_options = missing(option)) {
    ## List the enrichment options that you would like to make
    ## available for objects of class
    out <- list()
    out$option <- {{option}}
    ## Provide the descriptions of the enrichment options
    out$description <- {{description}}
    ## Add all as an option
    out$option <- c(out$option, 'all')
    out$description <- c(out$description, 'all available options')
    out$component <- {{component}}
    out$component[[length(out$component) + 1]] <- unique(unlist(out$component))
    names(out$component) <- names(out$description) <- out$option
    out$compute_function <- lapply(out$component, function(z) paste0('compute_', z))
    class(out) <- 'enrichment_options'
    if (all_options) {
        return(out)
    }
    invalid_options <- !(option %in% out$option)
    if (any(invalid_options)) {
        stop(gettextf('some options have not been implemented: %s', paste0('"', paste(option[invalid_options], collapse = ', '), '"')))
    }

    out <- list(option = option,
                description = out$description[option],
                component = out$component[option],
                compute_function = out$compute_function[option])
    class(out) <- 'enrichment_options'
    out
}

## ----echo = TRUE, eval = TRUE-------------------------------------------------
inverse.gaussian()$dev.resids
inverse.gaussian()$variance
inverse.gaussian()$aic

## ----echo = TRUE, eval = TRUE-------------------------------------------------
library("enrichwith")
dens <- function(y, m = 1, mu, phi, family) {
    object <- enrich(family)
    with(object, {
        c2 <- if (family == "binomial") c2fun(y, m) else c2fun(y)
        exp(m * (y * theta(mu) - bfun(theta(mu)) - c1fun(y))/phi -
            0.5 * afun(-m/phi) + c2)
    })
}

## ----echo = TRUE, eval = TRUE-------------------------------------------------
## Normal
all.equal(dens(y = 0.2, m = 3, mu = 1, phi = 3.22, gaussian()),
          dnorm(x = 0.2, mean = 1, sd = sqrt(3.22/3)))

## Gamma
all.equal(dens(y = 3, m = 1.44, mu = 2.3, phi = 1.3, Gamma()),
          dgamma(x = 3, shape = 1.44/1.3, 1.44/(1.3 * 2.3)))

## Inverse gaussian
all.equal(dens(y = 0.2, m = 7.23, mu = 1, phi = 3.22, inverse.gaussian()),
          SuppDists::dinvGauss(0.2, nu = 1, lambda = 7.23/3.22))

## Binomial
all.equal(dens(y = 0.34, m = 100, mu = 0.32, phi = 1, binomial()),
          dbinom(x = 34, size = 100, prob = 0.32))

## ----echo = TRUE, eval = TRUE-------------------------------------------------
# Get the data from the online supplmementary material of Agresti (2015)
data("endometrial", package = "enrichwith")
modML <- glm(HG ~ NV + PI + EH, family = binomial("probit"), data = endometrial)
theta_mle <- coef(modML)
summary(modML)

## ----echo = TRUE, eval = TRUE, messages = FALSE-------------------------------
library("brglm")
modBR <- brglm(HG ~ NV + PI + EH, family = binomial("probit"), data = endometrial)
theta_brglm <- coef(modBR)
summary(modBR)

## ----echo = TRUE, eval = TRUE-------------------------------------------------
library("enrichwith")
enriched_modML <- enrich(modML, with = "auxiliary functions")

## ----echo = TRUE, eval = TRUE-------------------------------------------------
# Extract the ingredients for the quasi Fisher scoring iteration from the enriched glm object
gradient <- enriched_modML$auxiliary_functions$score # gradient of the log-likelihood
information <- enriched_modML$auxiliary_functions$information # information matrix
bias <- enriched_modML$auxiliary_functions$bias # first-order bias

## ----echo = TRUE, eval = TRUE-------------------------------------------------
# The quasi Fisher scoring iteration using c(theta) = identity
for (k in seq.int(maxit)) {
    s_vector <- gradient(theta_current)
    i_matrix <- information(theta_current, type = "expected")
    b_vector <- bias(theta_current)
    step <- solve(i_matrix) %*% s_vector - b_vector
    theta_current <- theta_current + step
    # A stopping criterion
    if (all(abs(step) < epsilon)) {
        break
    }
}
(theta_e <- drop(theta_current))

## ----echo = TRUE, eval = TRUE-------------------------------------------------
all.equal(theta_e, theta_brglm, check.attributes = FALSE, tolerance = epsilon)

## ----echo = TRUE, eval = TRUE-------------------------------------------------
# The quasi Fisher scoring iteration using c(theta) = solve(i(theta)) %*% j(theta)
for (k in seq.int(maxit)) {
    s_vector <- gradient(theta_current)
    i_matrix <- information(theta_current, type = "expected")
    j_matrix <- information(theta_current, type = "observed")
    b_vector <- bias(theta_current)
    step <- solve(i_matrix) %*% (s_vector - j_matrix %*% b_vector)
    theta_current <- theta_current + step
    # A stopping criterion
    if (all(abs(step) < epsilon)) {
        break
    }
}
(theta_o <- drop(theta_current))

## ----echo = TRUE, eval = TRUE-------------------------------------------------
round(data.frame(theta_mle, theta_e, theta_o), 3)

## ----echo = TRUE, eval = TRUE-------------------------------------------------
se_theta_mle <- sqrt(diag(solve(information(theta_mle, type = "expected"))))
se_theta_e <- sqrt(diag(solve(information(theta_e, type = "expected"))))
se_theta_o <- sqrt(diag(solve(information(theta_o, type = "expected"))))
round(data.frame(z_mle = theta_mle/se_theta_mle,
                 z_br_e = theta_e/se_theta_e,
                 z_br_o = theta_o/se_theta_o), 3)

#' @method compute_{{component}} {{class}}
`compute_{{component}}.{{class}}` <- function(object, ...) {
    ## Write some code to compute the component {{component}} using
    ## the components of object and any other arguments in ...
    ## For example
    cat('component', '{{component}}', 'for objects of class', '{{class}}', "\n")
}


`compute_{{component}}` <- function(object, ...) {
    UseMethod('compute_{{component}}')
}

#' Create a enrichwith skeleton
#'
#'
#' Create an enrichwith skeleton file for the structured
#' implementation of methods to compute new components for objects of
#' a specific class
#'
#' @param class the class of the objects to be enriched
#' @param option a character vector with components the enrichment
#'     options
#' @param description a character vector of length
#'     \code{length(options)} with components the description of the
#'     enrichment options
#' @param component a list of as many character vectors as
#'     \code{length(option)}, specifying the names of the components
#'     that each option will add to the object after enrichment
#' @param path the path where the skeleton file will be created
#' @param filename the name of the skeleton file
#' @param attempt_rename attempt to rename syntactically incorrect
#'     component names? Default is \code{TRUE}
#'
#' @return A file with the necessary functions to use
#'     \code{enrichwith} infrastructure. The skeleton consists of the
#'     following functions
#' \itemize{
#'
#' \item One \code{compute_component.class} function per component
#' name from \code{unique(unlist(component))}. The function takes as
#' input the object to be enriched and returns as output the component
#' to be added to the object.
#'
#' \item The \code{get_enrichment_options.class} function, that takes
#' as input the object to be enriched and an enrichment option, and
#' returns the names of the components that will be appended to the
#' object for this option. This function can also be used to list the
#' available options and print their description.
#'
#' \item The \code{enrich.class} function
#'
#' }
#' @examples
#' \dontrun{
#' # Set the directory where the skeleton is placed
#' my_path <- "~/Downloads"
#' # This is the call that created the enrichment skeleton for glms
#' # that ships with the package
#' create_enrichwith_skeleton(class = "glm",
#'       option = c("auxiliary functions", "score vector",
#'        "mle of dispersion", "expected information",
#'        "observed information", "first-order bias"),
#'       description = c("various likelihood-based quantities
#'         (gradient of the log-likelihood, expected and observed
#'         information matrix and first term in the expansion of
#'         the bias of the mle) and a simulate method as functions
#'         of the model parameters",
#'        "gradient of the log-likelihood at the mle",
#'        "mle of the dispersion parameter",
#'        "expected information matrix evaluated at the mle",
#'        "observed information matrix evaluated at the mle",
#'        "first term in the expansion of the bias of the mle
#'         at the mle"),
#'       component = list("auxiliary_functions", "score_mle",
#'                        "dispersion_mle",
#'                        "expected_information_mle",
#'                        "observed_information_mle",
#'                        "bias_mle"),
#'       path = my_path,
#'       attempt_rename = FALSE)
#'
#' }
#' @export
create_enrichwith_skeleton <- function(class,
                                      option,
                                      description,
                                      component,
                                      path,
                                      filename = paste0(class, "_options.R"),
                                      attempt_rename = TRUE) {

    con <- file(paste(path, filename, sep = "/"), open = "w")

    o_length <- length(option)
    d_length <- length(description)

    message("* Checking supplied class for validity.")
    class_input <- as.character(class)
    ## Just making sure that a syntactically correct name is used for the class
    class_valid <- make.names(class, unique = TRUE, allow_ = FALSE)

    message("* Checking supplied class for validity.")
    option <- as.character(option)

    message("* Checking supplied descriptions for validity.")
    description <- as.character(description)
    if (d_length != o_length) {
        stop("ength(description) is not equal to length(option)")
    }

    message("* Checking supplied component for validity.")
    if (length(component) != o_length) {
        stop("length(component) is not equal to length(option)")
    }
    ## So even if there is only one component, component will become a list
    component <- lapply(component, function(comp) {
        comp_input <- as.character(comp)
        comp_valid <- make.names(comp, unique = TRUE, allow_ = FALSE)
        valid <- comp_input == comp_valid
        if (!all(valid)) {
            if (attempt_rename) {
                warning(gettextf("not syntactically valid component names: %s were renamed to %s",
                                 paste0(paste0("'", comp_input[!valid], "'"), collapse = ", "),
                                 paste0(paste0("'", comp_valid[!valid], collapse = ", "))))
            }
            else {
                ## close(con)
                warning(gettextf("not syntactically valid component names: %s",
                                 paste0(paste0("'", comp_input[!valid], "'"), collapse = ", ")))
            }
        }
        if (attempt_rename) comp_valid else comp_input
    })
    dat <- list(class = class_input,
                option = gsub("\"","'", deparse(option, width.cutoff = 500)),
                description = gsub("\"","'", deparse(description, width.cutoff = 500)),
                component = gsub("\"","'", deparse(component, width.cutoff = 500)))

    ## enrich function
    message("* Setting up enrich.", class_input, " function")
    ## Get _options template
    template_path <- system.file("templates", "template_enrich.R", package = "enrichwith",  mustWork = TRUE)
    ## Read template and replace according to class/options/description/component
    template_out <- whisker::whisker.render(readLines(template_path), data = dat)
    ## Write into _options file
    writeLines(template_out, con = con)

    ## get_enrichment_options fuction
    message("* Setting up get_enrichment_options.", class_input, " function")
    ## Get _options template
    template_path <- system.file("templates", "template_get_enrichment_options.R", package = "enrichwith",  mustWork = TRUE)
    ## Read template and replace according to class/options/description/component
    template_out <- whisker::whisker.render(readLines(template_path), data = dat)
    ## Write into _options file
    writeLines(template_out, con = con)

    ## compute_component functions
    message("* Setting up compute_component.", class_input, " function")
    components <- unique(unlist(component))
    template_path <- system.file("templates", "template_compute_component.R", package = "enrichwith",  mustWork = TRUE)
    for (j in seq.int(length(components))) {
        template_out <-  whisker::whisker.render(readLines(template_path),
                                                 data = list(component = components[j],
                                                             class = class_input))
        writeLines(template_out, con = con)
    }

    call <- match.call()

    writeLines(paste("\n\n\n\n", "## ## Call that produced the enrichwith template for the current script:\n", paste("##", deparse(call), collapse = "\n"), "\n", collapse = "", sep = ""), con = con)


    close(con)


}

#' Enrich objects of class \code{\link{family}}
#'
#' Enrich objects of class \code{\link{family}} with family-specific
#' mathematical functions
#'
#' @param object an object of class \code{\link{family}}
#' @param with a character vector with enrichment options for \code{object}
#' @param ... extra arguments to be passed to the \code{compute_*}
#'     functions
#'
#' @details
#'
#' \code{\link[stats]{family}} objects specify characteristics of the
#' models used by functions such as \code{\link[stats]{glm}}. The
#' families implemented in the \code{stats} package include
#' \code{\link[stats]{binomial}}, \code{\link[stats]{gaussian}},
#' \code{\link[stats]{Gamma}}, \code{\link[stats]{inverse.gaussian}},
#' and \code{\link[stats]{poisson}}, which are all special cases of
#' the exponential family of distributions that have probability mass
#' or density function of the form \deqn{f(y; \theta, \phi) =
#' \exp\left\{\frac{y\theta - b(\theta) - c_1(y)}{\phi/m} -
#' \frac{1}{2}a\left(-\frac{m}{\phi}\right) + c_2(y)\right\} \quad y
#' \in Y \subset \Re\,, \theta \in \Theta \subset \Re\, , \phi >
#' 0}{f(y, theta, phi) = exp((y * theta - b(theta) - c_1(y))/(phi/m) -
#' a(-m/phi)/2 + c_2(y))} where \eqn{m > 0}{m > 0} is an observation
#' weight, and \eqn{a(.)}{a(.)}, \eqn{b(.)}{b(.)},
#' \eqn{c_1(.)}{c_1(.)} and \eqn{c_2(.)}{c_2(.)} are sufficiently
#' smooth, real-valued functions.
#'
#' The current implementation of \code{\link[stats]{family}} objects
#' includes the variance function (\code{variance}), the deviance
#' residuals (\code{dev.resids}), and the Akaike information criterion
#' (\code{aic}). See, also \code{\link{family}}.
#'
#' The \code{enrich} method can further enrich exponential
#' \code{\link{family}} distributions with \eqn{\theta}{theta} in
#' terms of \eqn{\mu}{mu} (\code{theta}), the functions
#' \eqn{b(\theta)}{b(theta)} (\code{bfun}), \eqn{c_1(y)}{c_1(y)}
#' (\code{c1fun}), \eqn{c_2(y)}{c_2(y)} (\code{c2fun}),
#' \eqn{a(\zeta)}{a(zeta)} (\code{fun}), the first two derivatives of
#' \eqn{V(\mu)}{V(mu)} (\code{d1variance} and \code{d2variance},
#' respectively), and the first four derivatives of
#' \eqn{a(\zeta)}{a(zeta)} (\code{d1afun}, \code{d2afun},
#' \code{d3afun}, \code{d4afun}, respectively).
#'
#' Corresponding enrichment options are also available for
#' \code{\link[stats]{quasibinomial}},
#' \code{\link[stats]{quasipoisson}} and \code{\link[gnm]{wedderburn}}
#' families.
#'
#' The \code{\link[stats]{quasi}} families are enriched with
#' \code{d1variance} and \code{d2variance}.
#'
#' See \code{\link{enrich.link-glm}} for the enrichment of
#' \code{\link[=make.link]{link-glm}} objects.
#'
#' @return The object \code{object} of class \code{\link{family}} with
#'     extra components. \code{get_enrichment_options.family()}
#'     returns the components and their descriptions.
#'
#' @seealso \code{\link{enrich.link-glm}}, \code{\link{make.link}}
#'
#' @export
#' @examples
#'
#' ## An example from ?glm to illustrate that things still work with
#' ## enriched families
#' counts <- c(18,17,15,20,10,20,25,13,12)
#' outcome <- gl(3,1,9)
#' treatment <- gl(3,3)
#' print(d.AD <- data.frame(treatment, outcome, counts))
#' glm.D93 <- glm(counts ~ outcome + treatment, family = enrich(poisson()))
#' anova(glm.D93)
#' summary(glm.D93)
`enrich.family` <- function(object, with = "all", ...) {
    if (is.null(with)) {
        return(object)
    }
    dots <- list(...)
    enrichment_options <- get_enrichment_options(object, option = with)
    component <- unlist(enrichment_options$component)
    compute <- unlist(enrichment_options$compute_function)
    for (j in seq.int(length(component))) {
        ccall <- call(compute[j], object = object)
        for (nam in names(dots)) {
            ccall[[nam]] <- dots[[nam]]
        }
        object[[component[j]]] <- eval(ccall)
    }
    if (is.null(attr(object, "enriched"))) {
        attr(object, "enriched") <- TRUE
        classes <- class(object)
        class(object) <- c(paste0("enriched_", classes[1]), classes)
    }
    object
}



#' Available options for the enrichment objects of class family
#'
#' @param object the object to be enriched
#' @param option a character vector listing the options for enriching
#'     the object
#' @param all_options if \code{TRUE} then output a data frame with the
#'     available enrichment options, their descriptions, the names of
#'     the components that each option results in, and the names of
#'     the corresponding \code{compute_*} functions.
#' @return an object of class \code{enrichment_options}
#'
#' @details A check is being made whether the requested option is
#'     available. No check is being made on whether the functions that
#'     produce the components exist.
#' @examples
#' \dontrun{
#' get_enrichment_options.family(option = "all")
#' get_enrichment_options.family(all_options = TRUE)
#' }
#' @export
`get_enrichment_options.family` <- function(object, option, all_options = missing(option)) {
    ## List the enrichment options that you would like to make
    ## available for objects of class
    out <- list()
    out$option <- c('theta', 'bfun', 'c1fun', 'c2fun', 'd1variance', 'd2variance', 'afun', 'd1afun', 'd2afun', 'd3afun', 'd4fun', 'variance derivatives', 'function a derivatives')
    ## Provide the descriptions of the enrichment options
    out$description <- c('natural parameter', 'cumulant transform', 'c1 function', 'c2 function', '1st derivative of the variance function', '2nd derivative of the variance function', 'a function', '1st derivative of the a function', '2nd derivative of the a function', '3rd derivative of the a function', '4th derivative of the a function', '1st and 2nd derivative of the variance function', '1st, 2nd and 3rd derivative of the a function')
    ## Add all as an option
    out$option <- c(out$option, 'all')
    out$description <- c(out$description, 'all available options')
    out$component <- list('theta', 'bfun', 'c1fun', 'c2fun', 'd1variance', 'd2variance', 'afun', 'd1afun', 'd2afun', 'd3afun', 'd4afun', c('d1variance', 'd2variance'), c('d1afun', 'd2afun', 'd3afun'))
    out$component[[length(out$component) + 1]] <- unique(unlist(out$component))
    names(out$component) <- names(out$description) <- out$option
    out$compute_function <- lapply(out$component, function(z) paste0('compute_', z))
    class(out) <- 'enrichment_options'
    if (all_options) {
        return(out)
    }
    invalid_options <- !(option %in% out$option)
    if (any(invalid_options)) {
        stop(gettextf('some options have not been implemented: %s', paste0('"', paste(option[invalid_options], collapse = ', '), '"')))
    }

    out <- list(option = option,
                description = out$description[option],
                component = out$component[option],
                compute_function = out$compute_function[option])
    class(out) <- 'enrichment_options'
    out
}

#' @method compute_theta family
`compute_theta.family` <- function(object, ...) {
    family <- object$family
    switch(family,
           "poisson" = function(mu) {
               log(mu)
           },
           "quasipoisson" = function(mu) {
               log(mu)
           },
           "gaussian" = function(mu) {
               mu
           },
           "binomial" = function(mu) {
               log(mu/(1 - mu))
           },
           "quasibinomial" = function(mu) {
               log(mu/(1 - mu))
           },
           "wedderburn" = function(mu) {
               log(mu/(1 - mu))
           },
           "Gamma" = function(mu) {
               -1/mu
           },
           "inverse.gaussian" = function(mu) {
               -1/2*mu^2
           })
}

`compute_theta` <- function(object, ...) {
    UseMethod('compute_theta')
}

#' @method compute_bfun family
`compute_bfun.family` <- function(object, ...) {
    family <- object$family
    switch(family,
           "poisson" = function(theta) {
               exp(theta)
           },
           "quasipoisson" = function(theta) {
               exp(theta)
           },
           "gaussian" = function(theta) {
               theta^2/2
           },
           "binomial" = function(theta) {
               log(1 + exp(theta))
           },
           "quasibinomial" = function(theta) {
               log(1 + exp(theta))
           },
           "wedderburn" = function(theta) {
               log(1 + exp(theta))
           },
           "Gamma" = function(theta) {
               -log(-theta)
           },
           "inverse.gaussian" = function(theta) {
               -sqrt(-2*theta)
           })
}

`compute_bfun` <- function(object, ...) {
    UseMethod('compute_bfun')
}


#' @method compute_c1fun family
`compute_c1fun.family` <- function(object, ...) {
    family <- object$family
    switch(family,
           "poisson" = function(y) {
        0
    },
    "quasipoisson" = function(y) {
        0
    },
    "gaussian" = function(y) {
        y^2/2
    },
    "binomial" = function(y) {
        0
    },
    "quasibinomial" = function(y) {
        0
    },
    "wedderburn" = function(y) {
        0
    },
    "Gamma" = function(y) {
        -log(y)
    },
    "inverse.gaussian" = function(y) {
        1/(2*y)
    })
}

`compute_c1fun` <- function(object, ...) {
    UseMethod('compute_c1fun')
}


#' @method compute_c2fun family
`compute_c2fun.family` <- function(object, ...) {
    family <- object$family
    switch(family,
           "poisson" = function(y) {
        -log(gamma(y+1))
    },
    "quasipoisson" = function(y) {
        -log(gamma(y+1))
    },
    "gaussian" = function(y) {
        -log(2*pi)/2
    },
    "binomial" = function(y, m) {
        log(choose(m, m * y))
    },
    "quasibinomial" = function(y, m) {
        log(choose(m, m * y))
    },
    "wedderburn" = function(y, m) {
        log(choose(m, m * y))
    },
    "Gamma" = function(y) {
        -log(y)
    },
    "inverse.gaussian" = function(y) {
        -0.5 * log(2 * pi * y^3)
    })
}

`compute_c2fun` <- function(object, ...) {
    UseMethod('compute_c2fun')
}

#' @method compute_d1variance family
`compute_d1variance.family` <- function(object, ...) {
    family <- object$family
    switch(family,
           "poisson" = function(mu) {
               rep(1, length(mu))
           },
           "quasipoisson" = function(mu) {
               rep(1, length(mu))
           },
           "gaussian" = function(mu) {
               rep(0, length(mu))
                                },
           "binomial" = function(mu) {
               1 - 2 * mu
                     },
           "quasibinomial" = function(mu) {
               1 - 2 * mu
           },
           "wedderburn" = function(mu) {
               2 * mu * (1 - mu) * (1 - 2 * mu)
           },
           "Gamma" = function(mu) {
               2 * mu
           },
           "inverse.gaussian" = function(mu) {
               3 * mu^2
           },
           "quasi" = switch(object$varfun,
                            "constant" = function(mu) {
                                rep(0, length(mu))
                            },
                            "mu(1-mu)" = function(mu) {
                                1 - 2 * mu
                            },
                            "mu" = function(mu) {
                                rep(1, length(mu))
                            },
                            "mu^2" = function(mu) {
                                2 * mu
                            },
                            "mu^3" = function(mu) {
                                3 * mu^2
                            },
                            stop(sQuote(object$varfun), " variance function not supported")),
           stop(sQuote(family), " family not supported"))
}


`compute_d1variance` <- function(object, ...) {
    UseMethod('compute_d1variance')
}


#' @method compute_d2variance family
`compute_d2variance.family` <- function(object, ...) {
    family <- object$family
    switch(family,
           "poisson" = function(mu) {
               rep(0, length(mu))
           },
           "quasipoisson" = function(mu) {
               rep(0, length(mu))
           },
           "gaussian" = function(mu) {
                             rep(0, length(mu))
           },
           "binomial" = function(mu) {
               rep(-2, length(mu))
           },
           "quasibinomial" = function(mu) {
               rep(-2, length(mu))
           },
           "wedderburn" = function(mu) {
               2 - 12 * mu * ( 1- mu)
           },
           "Gamma" = function(mu) {
               rep(2, length(mu))
           },
           "inverse.gaussian" = function(mu) {
               6 * mu
           },
           "quasi" = switch(object$varfun,
                            "constant" = function(mu) {
                                rep(0, length(mu))
                            },
                            "mu(1-mu)" = function(mu) {
                                rep(-2, length(mu))
                            },
                            "mu" = function(mu) {
                                rep(0, length(mu))
                                          },
                            "mu^2" = function(mu) {
                                rep(2, length(mu))
                            },
                            "mu^3" = function(mu) {
                                6 * mu
                            },
                            stop(sQuote(object$varfun), " variance function not supported")),
           stop(sQuote(family), " family not supported"))
}


`compute_d2variance` <- function(object, ...) {
    UseMethod('compute_d2variance')
}

#' @method compute_afun family
`compute_afun.family` <- function(object, ...) {
    family <- object$family
    switch(family,
           "gaussian" = function(zeta) {
        -log(-zeta)
    },
    "Gamma" = function(zeta) {
        2*log(gamma(-zeta)) + 2*zeta*log(-zeta)
    },
    "inverse.gaussian" = function(zeta) {
        -log(-zeta)
    },
    "poisson" = function(zeta) {
        0
    },
    "quasipoisson" = function(zeta) {
        0
    },
    "binomial" = function(zeta) {
        0
    },
    "quasibinomial" = function(zeta) {
        0
    },
    "wedderburn" = function(zeta) {
        0
    })
}


`compute_afun` <- function(object, ...) {
    UseMethod('compute_afun')
}


#' @method compute_d1afun family
`compute_d1afun.family` <- function(object, ...) {
    family <- object$family
    switch(family,
           "gaussian" = function(zeta) {
        -1/zeta
    },
    "Gamma" = function(zeta) {
        -2*psigamma(-zeta, 0) + 2*log(-zeta) + 2
        ## this is the expectation of dev.resids + 2, because of the way dev.resids is implemented
    },
    "inverse.gaussian" = function(zeta) {
        -1/zeta
    },
    "poisson" = function(zeta) {
        0
    },
    "quasipoisson" = function(zeta) {
        0
    },
    "binomial" = function(zeta) {
        0
    },
    "quasibinomial" = function(zeta) {
        0
    },
    "wedderburn" = function(zeta) {
        0
    })
}


`compute_d1afun` <- function(object, ...) {
    UseMethod('compute_d1afun')
}


#' @method compute_d2afun family
`compute_d2afun.family` <- function(object, ...) {
    family <- object$family
    switch(family,
           "gaussian" = function(zeta) {
        1/zeta^2
    },
           "Gamma" = function(zeta) {
        2*psigamma(-zeta, 1) + 2/zeta
    },
           "inverse.gaussian" = function(zeta) {
        1/zeta^2
    },
    "poisson" = function(zeta) {
        0
    },
    "quasipoisson" = function(zeta) {
        0
    },
    "binomial" = function(zeta) {
        0
    },
    "quasibinomial" = function(zeta) {
        0
    },
    "wedderburn" = function(zeta) {
        0
    })
}


`compute_d2afun` <- function(object, ...) {
    UseMethod('compute_d2afun')
}

#' @method compute_d3afun family
`compute_d3afun.family` <- function(object, ...) {
    family <- object$family
    switch(family,
           "gaussian" = function(zeta) {
        -2/zeta^3
    },
    "Gamma" = function(zeta) {
        -2*psigamma(-zeta, 2) - 2/zeta^2
    },
    "inverse.gaussian" = function(zeta) {
        -2/zeta^3
    },
    "poisson" = function(zeta) {
        0
    },
    "quasipoisson" = function(zeta) {
        0
    },
    "binomial" = function(zeta) {
        0
    },
    "quasibinomial" = function(zeta) {
        0
    },
    "wedderburn" = function(zeta) {
        0
    })
}

`compute_d3afun` <- function(object, ...) {
    UseMethod('compute_d3afun')
}


#' @method compute_d4afun family
`compute_d4afun.family` <- function(object, ...) {
    family <- object$family
    switch(family,
           "gaussian" = function(zeta) {
        6/zeta^4
    },
    "Gamma" = function(zeta) {
        2 * psigamma(-zeta, 3) + 4/zeta^3
    },
    "inverse.gaussian" = function(zeta) {
        6/zeta^4
    },
    "poisson" = function(zeta) {
        0
    },
    "quasipoisson" = function(zeta) {
        0
    },
    "binomial" = function(zeta) {
        0
    },
    "quasibinomial" = function(zeta) {
        0
    },
    "wedderburn" = function(zeta) {
        0
    })
}


`compute_d4afun` <- function(object, ...) {
    UseMethod('compute_d4afun')
}

## ## Call that produced the initial enrichwith template for the current script:
## create_enrichwith_skeleton(class = "family", option = c("d1variance",
##     "d2variance", "d1afun", "d2afun", "d3afun", "variance derivatives",
##     "function a derivatives"), description = c("1st derivative of the variance function",
##     "2nd derivative of the variance function", "1st derivative of the a function",
##     "2nd derivative of the a function", "3rd derivative of the a function",
##     "1st and 2nd derivative of the variance function", "1st, 2nd and 3rd derivative of the a function"),
##     component = list("d1variance", "d2variance", "d1afun", "d2afun",
##         "d3afun", c("d1variance", "d2variance"), c("d1afun",
##             "d2afun", "d3afun")), path = "~/Downloads", attempt_rename = TRUE)

#' @export
print.enrichment_options <- function(x, ...) {
    o_length <- length(x$option)
    for (o in seq.int(o_length)) {
        cat('-------\n')
        cat('Option:', x$option[o], "\n")
        cat('Description:', x$description[o], "\n")
        mat <- data.frame(component = x$component[[o]], compute_function = x$compute_function[[o]])
        print(mat)
    }
}

#' Fitting generalized linear models enriched with useful components
#'
#' \code{\link{enriched_glm}} fits generalized linear models using
#' \code{\link{glm}} and then enriches the resulting object with all
#' enrichment options.
#'
#' @inheritParams stats::glm
#' @param ... other arguments passed to \code{\link{glm}}
#'
#' @details
#'
#' \code{enriched_glm} has the same interface as \code{\link{glm}}
#'
#' @return
#'
#' An object of class \code{enriched_glm} that contains all the
#' components of a \code{\link{glm}} object, along with a set of
#' auxiliary functions (score function, information matrix, a simulate
#' method, first term in the expansion of the bias of the maximum
#' likelihood estimator, and dmodel, pmodel, qmodel), the maximum
#' likelihood estimate of the dispersion parameter, the expected or
#' observed information at the maximum likelihood estimator, and the
#' first term in the expansion of the bias of the maximum likelihood
#' estimator.
#'
#' See \code{\link{enrich.glm}} for more details and links for the
#' auxiliary functions.
#'
#'
#' @examples
#' \dontrun{
#' # A Gamma example, from McCullagh & Nelder (1989, pp. 300-2)
#' clotting <- data.frame(
#'    u = c(5,10,15,20,30,40,60,80,100, 5,10,15,20,30,40,60,80,100),
#'    time = c(118,58,42,35,27,25,21,19,18,69,35,26,21,18,16,13,12,12),
#'    lot = factor(c(rep(1, 9), rep(2, 9))))
#'
#' # Fit a generalized linear model
#' cML <- enriched_glm(time ~ lot*log(u), data = clotting, family = Gamma("log"))
#'
#' # Evaluate the densities at the data points in clotting at the
#' # maximum likelihood estimates
#' cML_dmodel <- get_dmodel_function(cML) # same as cML$auxiliary_functions$dmodel
#' cML_dmodel()
#'
#' # Evaluate the densities at supplied data points
#' new_data <- data.frame(u = c(15:17, 15:17),
#'                        time = c(30:32, 15:17),
#'                        lot = factor(c(1, 1, 1, 2, 2, 2)))
#' cML_dmodel(data = new_data)
#'
#' # Get pmodel and qmodel function
#' cML_pmodel <- get_pmodel_function(cML) # same as cML$auxiliary_functions$pmodel
#' cML_qmodel <- get_qmodel_function(cML) # same as cML$auxiliary_functions$qmodel
#'
#' # The following should return c(30:32, 15:17)
#' probs <- cML_pmodel(data = new_data)
#' cML_qmodel(probs, data = new_data)
#'
#' # Evaluate the observed information matrix at the MLE
#' cML_info <- get_information_function(cML)
#' cML_info(type = "observed")
#'
#' # Wald tests based on the observed information at the
#' # moment based esimator of the dispersion
#' dispersion <- summary(cML)$dispersion
#' cML_vcov_observed <- solve(cML_info(dispersion = dispersion, type = "observed"))
#' lmtest::coeftest(cML, vcov = cML_vcov_observed)
#'
#' # Wald tests based on the expected information at the
#' # moment based esimator of the dispersion
#' cML_vcov_expected <- solve(cML_info(dispersion = dispersion, type = "expected"))
#' lmtest::coeftest(cML, vcov = cML_vcov_expected)
#' # Same statistics as coef(summary(cML))[, "t value"]
#'
#' }
#'
#'
#' @export
enriched_glm <- function(formula, family = gaussian, ...) {
    .Deprecated(msg = '`enriched_glm()` is deprecated and will be defunct in the next release. Use the call `enrich(glm(...), with = "all") or `glm(...) |> enrich(with = "all")` instead.')
    fit <- glm(formula, family = family, ...)
    enrich(fit, with = "all")
}

#' Histology grade and risk factors for 79 cases of endometrial cancer
#'
#' @format A data frame with 79 rows and 4 variables:
#' \describe{
#'
#' \item{NV}{neovasculization with coding 0 for absent and 1 for present}
#'
#' \item{PI}{pulsality index of arteria uterina}
#'
#' \item{EH}{endometrium height}
#'
#' \item{HG}{histology grade with coding 0 for low grade and 1 for high grade}
#'
#' }
#'
#' @source The packaged data set was downloaded in \code{.dat} format
#'     from \url{https://users.stat.ufl.edu/~aa/glm/data/}. The latter
#'     link provides the data sets used in Agresti (2015).
#'
#'     The endometrial data set was first analyzed in Heinze and
#'     Schemper (2002), and was originally provided by Dr
#'     E. Asseryanis from the Medical University of Vienna.
#'
#'
#' Agresti, A. 2015. *Foundations of Linear and Generalized Linear
#' Models*.  Wiley Series in Probability and Statistics. Wiley.
#'
#' Heinze G, Schemper M (2002). A Solution to the Problem of
#' Separation in Logistic Regression. *Statistics in Medicine*,
#' **21**, 2409–2419. \doi{10.1002/sim.1047}.
#'
"endometrial"

#' Enrich objects of class betareg
#'
#' Enrich objects of class \code{\link[betareg]{betareg}} with any or all of a
#' set of auxiliary functions, the expected information at the maximum
#' likelihood estimator, and the first term in the expansion of the
#' bias of the maximum likelihood estimator.
#'
#'
#' @param object an object of class \code{\link[betareg]{betareg}}
#' @param with  a character vector of options for the enrichment of \code{object}
#' @param ... extra arguments to be passed to the
#'     \code{compute_*} functions
#'
#' @details
#' The \code{auxiliary_functions} component consists of any or all of the following functions:
#' \itemize{
#' \item \code{score}: the log-likelihood derivatives as a function of the model parameters; see \code{get_score_function.betareg}
#' \item \code{information}: the expected information as a function of the model parameters; see \code{\link{get_information_function.betareg}}
#' \item \code{bias}: the first-order term in the expansion of the bias of the maximum likelihood estimator as a function of the model parameters; see \code{\link{get_bias_function.betareg}}
#' \item \code{simulate}: a \code{\link{simulate}} function for \code{\link[betareg]{betareg}} objects that can simulate variates from the model at user-supplied parameter values for the regression parameters (default is the maximum likelihood estimates); see \code{\link{get_simulate_function.betareg}}
#' }
#'
#' @return The object \code{object} of class \code{\link[betareg]{betareg}}
#'     with extra components. \code{get_enrichment_options.betareg()}
#'     returns the components and their descriptions.
#'
#' @export
#' @examples
#'
#' \dontrun{
#' if (require("betareg")) {
#'
#'    data("GasolineYield", package = "betareg")
#'    gy <- betareg(yield ~ batch + temp, data = GasolineYield)
#'
#'    # Get a function that evaluates the expected information for gy at supplied parameter values
#'    gy_info <- get_information_function(gy)
#'.   # compare standard errors with what `summary` returns
#'    all.equal(sqrt(diag(solve(gy_info())))[1:11],
#'              coef(summary(gy))$mean[, 2], check.attributes = FALSE)
#'.   # evaluating at different parameter values
#'    gy_info(rep(1, length = 12))
#'
#'    # Get a function that evaluates the first-order bias of gy at supplied parameter values
#'    gy_bias <- get_bias_function(gy)
#'    # compare with internal betareg implementation of bias correction
#'    gy_bc <- update(gy, type = "BC")
#'    all.equal(gy_bias(coef(gy)), gy_bc$bias, check.attributes = FALSE)
#'
#'  }
#'}
#'
`enrich.betareg` <- function(object, with = "all", ...) {
    if (is.null(with)) {
        return(object)
    }
    dots <- list(...)
    enrichment_options <- get_enrichment_options(object, option = with)
    component <- unlist(enrichment_options$component)
    compute <- unlist(enrichment_options$compute_function)
    for (j in seq.int(length(component))) {
        ccall <- call(compute[j], object = object)
        for (nam in names(dots)) {
            ccall[[nam]] <- dots[[nam]]
        }
        object[[component[j]]] <- eval(ccall)
    }
    if (is.null(attr(object, "enriched"))) {
        attr(object, "enriched") <- TRUE
        classes <- class(object)
        class(object) <- c(paste0("enriched_", classes[1]), classes)
    }
    object
}

#' Available options for the enrichment objects of class betareg
#'
#' @param object the object to be enriched
#' @param option a character vector listing the options for enriching
#'     the object
#' @param all_options if \code{TRUE} then output a data frame with the
#'     available enrichment options, their descriptions, the names of
#'     the components that each option results in, and the names of
#'     the corresponding \code{compute_*} functions.
#' @return an object of class \code{enrichment_options}
#'
#' @details A check is being made whether the requested option is
#'     available. No check is being made on whether the functions that
#'     produce the components exist.
#' @examples
#' \dontrun{
#' get_enrichment_options.betareg(option = "all")
#' get_enrichment_options.betareg(all_options = TRUE)
#' }
#' @export
`get_enrichment_options.betareg` <- function(object, option, all_options = missing(option)) {
    ## List the enrichment options that you would like to make
    ## available for objects of class
    out <- list()
    out$option <- c('auxiliary functions', 'score vector', 'expected information', 'first-order bias')
    ## Provide the descriptions of the enrichment options
    out$description <- c('various likelihood-based quantities (gradient of the log-likelihood, expected information matrix and first term in the expansion of the bias of the mle) and a simulate method as functions of the model parameters', 'gradient of the log-likelihood at the mle', 'expected information matrix evaluated at the mle', 'first term in the expansion of the bias of the mle at the mle')
    ## Add all as an option
    out$option <- c(out$option, 'all')
    out$description <- c(out$description, 'all available options')
    out$component <- list('auxiliary_functions', 'score_mle', 'expected_information_mle', 'bias_mle')
    out$component[[length(out$component) + 1]] <- unique(unlist(out$component))
    names(out$component) <- out$option
    out$compute_function <- lapply(out$component, function(z) paste0('compute_', z))
    class(out) <- 'enrichment_options'
    if (all_options) {
        return(out)
    }
    invalid_options <- !(option %in% out$option)
    if (any(invalid_options)) {
        stop(gettextf('some options have not been implemented: %s', paste0('"', paste(option[invalid_options], collapse = ', '), '"')))
    }
    out <- list(option = option,
                description = out$description[out$option == option],
                component = out$component[option],
                compute_function = out$compute_function[option])
    class(out) <- 'enrichment_options'
    out
}

#' @method compute_auxiliary_functions betareg
`compute_auxiliary_functions.betareg` <- function(object, ...) {
    if (is.null(object$model)) {
        object <- update(object, model = TRUE)
    }
    y <-  object$y
    x <- model.matrix(object, model = "mean")
    z <- model.matrix(object, model = "precision")
    n <- NROW(x)
    k <- NCOL(x)
    weights <- object$weights
    if (is.null(weights))
        weights <- rep.int(1, n)
    nobs <- sum(weights > 0)
    offset <- object$offset
    offset <- lapply(offset, function(off) {
        if (is.null(off)) rep.int(0, n) else off
    })
    m <- NCOL(z)
    if (m < 1L)
        stop("dispersion regression needs to have at least one parameter")
    phi_const <- (m == 1L) && isTRUE(all.equal(as.vector(z[, 1L]), rep.int(1, n)))
    linkmean <- enrich(object$link$mean)
    linkprec <- enrich(object$link$precision)
    linkinv <- linkmean$linkinv
    mu.eta <- linkmean$mu.eta

    dmu.deta <- linkmean$d2mu.deta
    phi_linkinv <- linkprec$linkinv
    phi_mu.eta <- linkprec$mu.eta
    phi_dmu.deta <- linkprec$d2mu.deta
    ystar <- qlogis(y)
    u <- log(1 - y)
    score <- function(coefficients, contributions = FALSE) {
        if (missing(coefficients)) {
            coefficients <- coef(object, model = "full")
        }
        beta <- coefficients[seq.int(length.out = k)]
        gamma <- coefficients[seq.int(length.out = m) + k]
        eta <- as.vector(x %*% beta + offset[[1L]])
        phi_eta <- as.vector(z %*% gamma + offset[[2L]])
        mu <- linkinv(eta)
        phi <- phi_linkinv(phi_eta)
        mustar <- digamma(mu * phi) - digamma((1 - mu) * phi)
        psi1 <- trigamma(mu * phi)
        psi2 <- trigamma((1 - mu) * phi)
        tbar_ubar <- ystar - mustar
        ubar <- u - digamma((1 - mu) * phi) + digamma(phi)
        rval <- cbind(phi * tbar_ubar * mu.eta(eta) * weights * x,
        (mu * tbar_ubar + ubar) * phi_mu.eta(phi_eta) * weights * z)
        if (contributions)
            rval
        else
            colSums(rval)
    }

    information <- function(coefficients, QR = TRUE, CHOL = FALSE,
                            type = c("expected", "observed")) {
        if (missing(coefficients)) {
            coefficients <- coef(object, model = "full")
        }
        type <- match.arg(type)
        beta <- coefficients[seq.int(length.out = k)]
        gamma <- coefficients[seq.int(length.out = m) + k]
        eta <- as.vector(x %*% beta + offset[[1L]])
        phi_eta <- as.vector(z %*% gamma + offset[[2L]])
        mu <- linkinv(eta)
        phi <- phi_linkinv(phi_eta)
        mustar <- digamma(mu * phi) - digamma((1 - mu) * phi)
        psi1 <- trigamma(mu * phi)
        psi2 <- trigamma((1 - mu) * phi)
        a <- psi1 + psi2
        b <- psi1 * mu^2 + psi2 * (1 - mu)^2 - trigamma(phi)
        D1 <- mu.eta(eta)
        D2 <- phi_mu.eta(phi_eta)
        wbb <- phi^2 * a * D1^2
        wpp <- b * D2^2
        wbp <- phi * (mu * a - psi2) * D1 * D2
        kbb <- if (k > 0L)
            crossprod(sqrt(weights) * sqrt(wbb) * x)
        else crossprod(x)
        kpp <- if (m > 0L)
            crossprod(sqrt(weights) * sqrt(wpp) * z)
        else crossprod(z)
        kbp <- if (k > 0L & m > 0L)
            crossprod(weights * wbp * x, z)
        else crossprod(x, z)
        out <- cbind(rbind(kbb, t(kbp)), rbind(kbp, kpp))
        if (type == "observed") {
            tbar_ubar <- ystar - mustar
            ubar <- u - digamma((1 - mu) * phi) + digamma(phi)
            D1dash <- dmu.deta(eta)
            D2dash <- phi_dmu.deta(phi_eta)
            kbb <- if (k > 0L)
                       crossprod(phi * D1dash * tbar_ubar * x, x)
                   else crossprod(x)
            kpp <- if (m > 0L)
                       crossprod(D2dash * (mu * tbar_ubar + ubar) * z, z)
                   else crossprod(z)
            kbp <- if (k > 0L & m > 0L)
                       crossprod(D1 * D2 * tbar_ubar * x, z)
                   else crossprod(x, z)
            out <- out - cbind(rbind(kbb, t(kbp)), rbind(kbp, kpp))
        }
        colnames(out) <- names(coef(object))
        if (CHOL)
            out <- chol(out)
        attr(out, "coefficients") <- coefficients
        out

    }

    bias <- function(coefficients) {
        if (missing(coefficients)) {
            coefficients <- coef(object, model = "full")
        }
        beta <- coefficients[seq.int(length.out = k)]
        gamma <- coefficients[seq.int(length.out = m) + k]
        eta <- as.vector(x %*% beta + offset[[1L]])
        phi_eta <- as.vector(z %*% gamma + offset[[2L]])
        mu <- linkinv(eta)
        phi <- phi_linkinv(phi_eta)
        mustar <- digamma(mu * phi) - digamma((1 - mu) * phi)
        psi1 <- trigamma(mu * phi)
        psi2 <- trigamma((1 - mu) * phi)
        InfoInv <- try(solve(information(coefficients)), silent = TRUE)
        D1 <- mu.eta(eta)
        D2 <- phi_mu.eta(phi_eta)
        D1dash <- dmu.deta(eta)
        D2dash <- phi_dmu.deta(phi_eta)
        dPsi1 <- psigamma(mu * phi, 2)
        dPsi2 <- psigamma((1 - mu) * phi, 2)
        kappa2 <- psi1 + psi2
        kappa3 <- dPsi1 - dPsi2
        psi3 <- psigamma(phi, 1)
        dPsi3 <- psigamma(phi, 2)
        PQsum <- function(t) {
            if (t <= k) {
                Xt <- x[, t]
                bb <- if (k > 0L)
                  crossprod(x, weights * phi^2 * D1 * (phi *
                    D1^2 * kappa3 + D1dash * kappa2) * Xt * x)
                else crossprod(x)
                bg <- if ((k > 0L) & (m > 0L))
                  crossprod(x, weights * phi * D1^2 * D2 * (mu *
                    phi * kappa3 + phi * dPsi2 + kappa2) * Xt *
                    z)
                else crossprod(x, z)
                gg <- if (m > 0L)
                  crossprod(z, weights * phi * D1 * D2^2 * (mu^2 *
                    kappa3 - dPsi2 + 2 * mu * dPsi2) * Xt * z) +
                    crossprod(z, weights * phi * D1 * D2dash *
                      (mu * kappa2 - psi2) * Xt * z)
                else crossprod(z)
            }
            else {
                Zt <- z[, t - k]
                bb <- if (k > 0L)
                  crossprod(x, weights * phi * D2 * (phi * D1^2 *
                    mu * kappa3 + phi * D1^2 * dPsi2 + D1dash *
                    mu * kappa2 - D1dash * psi2) * Zt * x)
                else crossprod(x)
                bg <- if ((k > 0L) & (m > 0L))
                  crossprod(x, weights * D1 * D2^2 * (phi * mu^2 *
                    kappa3 + phi * (2 * mu - 1) * dPsi2 + mu *
                    kappa2 - psi2) * Zt * z)
                else crossprod(x, z)
                gg <- if (m > 0L)
                  crossprod(z, weights * D2^3 * (mu^3 * kappa3 +
                    (3 * mu^2 - 3 * mu + 1) * dPsi2 - dPsi3) *
                    Zt * z) + crossprod(z, weights * D2dash *
                    D2 * (mu^2 * kappa2 + (1 - 2 * mu) * psi2 -
                    psi3) * Zt * z)
                else crossprod(z)
            }
            pq <- rbind(cbind(bb, bg), cbind(t(bg), gg))
            sum(diag(InfoInv %*% pq))/2
        }
        if (inherits(InfoInv, "try-error")) {
            bias <- rep.int(NA_real_, k + m)
        }
        else {
            bias <- drop(-InfoInv %*% sapply(1:(k + m), PQsum))
        }
        bias
    }

    simulate <- function(coefficients, nsim = 1, seed = NULL) {
        if (missing(coefficients)) {
            coefficients <-  coef(object, model = "full")
        }
        else {
            if (!isTRUE(identical(length(coefficients), length(coef(object, model = "full"))))) {
                stop("`coefficients` does not have the right length")
            }
        }
        if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
            runif(1)
        if (is.null(seed))
            RNGstate <- get(".Random.seed", envir = .GlobalEnv)
        else {
            R.seed <- get(".Random.seed", envir = .GlobalEnv)
            set.seed(seed)
            RNGstate <- structure(seed, kind = as.list(RNGkind()))
            on.exit(assign(".Random.seed", R.seed, envir = .GlobalEnv))
        }
        beta <- coefficients[seq.int(length.out = k)]
        gamma <- coefficients[seq.int(length.out = m) + k]
        eta <- as.vector(x %*% beta + offset[[1L]])
        phi_eta <- as.vector(z %*% gamma + offset[[2L]])
        mu <- linkinv(eta)
        phi <- phi_linkinv(phi_eta)
        ps <- mu * phi
        qs <- (1 - mu) * phi
        matrix(1 - rbeta(n * nsim, qs, ps), n, nsim)
    }
    return(list(score = score,
                information = information,
                bias = bias,
                simulate = simulate))
}


`compute_auxiliary_functions` <- function(object, ...) {
    UseMethod('compute_auxiliary_functions')
}

#' @method compute_score_mle betareg
`compute_score_mle.betareg` <- function(object, ...) {
    get_score_function(object)()
}


`compute_score_mle` <- function(object, ...) {
    UseMethod('compute_score_mle')
}

#' @method compute_expected_information betareg
`compute_expected_information_mle.betareg` <- function(object, ...) {
    get_information_function(object)()
}


`compute_expected_information_mle` <- function(object, ...) {
    UseMethod('compute_expected_information_mle')
}

#' @method compute_bias_mle betareg
`compute_bias_mle.betareg` <- function(object, ...) {
    get_bias_function(object)()
}


`compute_bias_mle` <- function(object, ...) {
    UseMethod('compute_bias_mle')
}

#' Function to compute/extract auxiliary functions from objects of
#' class \code{betreg}/\code{enriched_betareg}
#'
#' @param object an object of class \code{betareg} or\code{enriched_betareg}
#' @param ... currently not used
#'
#' @details
#'
#' See \code{\link{enrich.betareg}} for details.
#'
#' @export
get_auxiliary_functions.betareg <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        enriched_object <- enrich(object, with = "auxiliary functions")
        enriched_object$auxiliary_functions
    }
    else {
        object$auxiliary_functions
    }
}

#' Function to compute/extract a simulate function for response
#' vectors from an object of class \code{betareg}/\code{enriched_betareg}
#'
#' @param object an object of class \code{betareg} or\code{enriched_betareg}
#' @param ... currently not used
#'
#' @details
#' The computed/extracted simulate function has arguments
#' \describe{
#'
#' \item{coefficients}{the regression coefficients at which the
#' response vectors are simulated. If missing then the maximum
#' likelihood estimates are used}
#'
#' \item{nsim}{number of response vectors to simulate.  Defaults to \code{1}}
#'
#' \item{seed}{an object specifying if and how the random number
#' generator should be initialized ('seeded'). It can be either
#' \code{NULL} or an integer that will be used in a call to
#' \code{set.seed} before simulating the response vectors.  If set,
#' the value is saved as the \code{seed} attribute of the returned
#' value.  The default, \code{NULL} will not change the random
#' generator state, and return \code{.Random.seed} as the \code{seed}
#' attribute, see \code{Value}}
#'
#' }
#'
#' @export
get_simulate_function.betareg <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        get_auxiliary_functions(object)$simulate
    }
    else {
        object$auxiliary_functions$simulate
    }
}

#' Function to compute/extract a function that returns the scores
#' (derivatives of the log-likelihood) for an object of class
#' \code{betareg}/\code{enriched_betareg}
#'
#' @param object an object of class \code{betareg} or\code{enriched_betareg}
#' @param ... currently not used
#'
#' @details
#' The computed/extracted function has arguments
#' \describe{
#'
#' \item{coefficients}{the regression coefficients at which the scores
#' are computed. If missing then the maximum likelihood estimates are
#' used}
#'
#' }
#'
#' @export
get_score_function.betareg <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        get_auxiliary_functions(object)$score
    }
    else {
        object$auxiliary_functions$score
    }
}

#' Function to compute/extract a function that returns the information
#' matrix for an object of class \code{betareg}/\code{enriched_betareg}
#'
#' @param object an object of class \code{betareg} or\code{enriched_betareg}
#' @param ... currently not used
#'
#' @details
#' The computed/extracted function has arguments
#' \describe{
#'
#' \item{coefficients}{the regression coefficients at which the
#' information matrix is evaluated. If missing then the maximum
#' likelihood estimates are used}
#'
#'
#' \item{type}{should the function return the 'expected' or 'observed' information? Default is \code{expected}}
#'
#' \item{QR}{Currently not used}
#'
#' \item{CHOL}{If \code{TRUE}, then the Cholesky decomposition of the information matrix at the coefficients is returned}
#'
#' }
#'
#' @export
get_information_function.betareg <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        get_auxiliary_functions(object)$information
    }
    else {
        object$auxiliary_functions$information
    }
}


#' Function to compute/extract a function that returns the first term
#' in the expansion of the bias of the MLE for the parameters of an
#' object of class \code{betareg}/\code{enriched_betareg}
#'
#' @param object an object of class \code{betareg} or\code{enriched_betareg}
#' @param ... currently not used
#'
#' @details
#' The computed/extracted function has arguments
#' \describe{
#'
#' \item{coefficients}{the regression coefficients at which the
#' first-order bias is evacuated. If missing then the maximum
#' likelihood estimates are used}
#' }
#'
#' @export
get_bias_function.betareg <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        get_auxiliary_functions(object)$bias
    }
    else {
        object$auxiliary_functions$bias
    }
}



## ## Call that produced the original version of the enrichwith template for the current script:
## enrichwith:::create_enrichwith_skeleton(class = "betareg", option = c("auxiliary functions",
##     "score vector", "mle of dispersion", "expected information",
##     "first-order bias"), description = c("various likelihood-based quantities (gradient of the log-likelihood, expected information matrix and first term in the expansion of the bias of the mle) and a simulate method as functions of the model parameters",
##     "gradient of the log-likelihood at the mle", "mle of the dispersion parameter",
##     "expected information matrix evaluated at the mle",
##     "first term in the expansion of the bias of the mle at the mle"),
##     component = list("auxiliary_functions", "score_mle", "dispersion_mle",
##         "expected_information_mle",
##         "bias_mle"), path = "~/Downloads", attempt_rename = FALSE)

#' Enrich objects of class \code{\link[=make.link]{link-glm}}
#'
#'
#' Enrich objects of class \code{\link[=make.link]{link-glm}} with
#' further derivatives of \code{linkinv} with respect to \code{eta}.
#'
#' @param object an object of class \code{\link[=make.link]{link-glm}}
#' @param with a character vector with enrichment options for \code{object}
#' @param ... extra arguments to be passed to the \code{compute_*}
#'     functions
#'
#' @details
#' The \code{enrich.link-glm} method supports \code{logit},
#' \code{probit}, \code{cauchit}, \code{cloglog}, \code{identity},
#' \code{log}, \code{sqrt}, \code{1/mu^2}, \code{inverse}, as well as
#' the \code{\link{power}} family of links.
#'
#' @return The object \code{object} of class \code{\link[=make.link]{link-glm}}
#'     with extra components. \code{get_enrichment_options.link-glm()}
#'     returns the components and their descriptions.
#'
#' @examples
#'
#' # Example
#' elogit <- enrich(make.link("logit"), with = "inverse link derivatives")
#' str(elogit)
#' elogit$d2mu.deta
#' elogit$d3mu.deta
#' @export
`enrich.link-glm` <- function(object, with = "all", ...) {
    if (is.null(with)) {
        return(object)
    }
    dots <- list(...)
    enrichment_options <- get_enrichment_options(object, option = with)
    component <- unlist(enrichment_options$component)
    compute <- unlist(enrichment_options$compute_function)
    for (j in seq.int(length(component))) {
        ccall <- call(compute[j], object = object)
        for (nam in names(dots)) {
            ccall[[nam]] <- dots[[nam]]
        }
        object[[component[j]]] <- eval(ccall)
    }
    if (is.null(attr(object, "enriched"))) {
        attr(object, "enriched") <- TRUE
        classes <- class(object)
        class(object) <- c(paste0("enriched_", classes[1]), classes)
    }
    object
}



#' Available options for the enrichment objects of class link-glm
#'
#' @param object the object to be enriched
#' @param option a character vector listing the options for enriching
#'     the object
#' @param all_options if \code{TRUE} then output a data frame with the
#'     available enrichment options, their descriptions, the names of
#'     the components that each option results in, and the names of
#'     the corresponding \code{compute_*} functions.
#' @return an object of class \code{enrichment_options}
#'
#' @details A check is being made whether the requested option is
#'     available. No check is being made on whether the functions that
#'     produce the components exist.
#' @examples
#' \dontrun{
#' `get_enrichment_options.link-glm`(option = "all")
#' `get_enrichment_options.link-glm`(all_options = TRUE)
#' }
#' @export
`get_enrichment_options.link-glm` <- function(object, option, all_options = missing(option)) {
    ## List the enrichment options that you would like to make
    ## available for objects of class
    out <- list()
    out$option <- c('d2mu.deta', 'd3mu.deta', 'inverse link derivatives')
    ## Provide the descriptions of the enrichment options
    out$description <- c('2nd derivative of the inverse link function', '3rd derivative of the inverse link function', '2nd and 3rd derivative of the inverse link function')
    ## Add all as an option
    out$option <- c(out$option, 'all')
    out$description <- c(out$description, 'all available options')
    out$component <- list('d2mu.deta', 'd3mu.deta', c('d2mu.deta', 'd3mu.deta'))
    out$component[[length(out$component) + 1]] <- unique(unlist(out$component))
    names(out$component) <- names(out$description) <- out$option
    out$compute_function <- lapply(out$component, function(z) paste0('compute_', z))
    class(out) <- 'enrichment_options'
    if (all_options) {
        return(out)
    }
    invalid_options <- !(option %in% out$option)
    if (any(invalid_options)) {
        stop(gettextf('some options have not been implemented: %s', paste0('"', paste(option[invalid_options], collapse = ', '), '"')))
    }

    out <- list(option = option,
                description = out$description[option],
                component = out$component[option],
                compute_function = out$compute_function[option])
    class(out) <- 'enrichment_options'
    out
}

#' @method compute_d2mu.deta link-glm
`compute_d2mu.deta.link-glm` <- function(object, ...) {
    mu.eta <- object$mu.eta
    linkinv <- object$linkinv
    link <- object$name
    if (grepl("mu\\^", link) & (link != "1/mu^2")) {
        d2mu.deta <- function(eta) {
            (1/lambda) * (1/lambda - 1) * eta^(1/lambda - 2)
        }
        ## set the environment so lambda can be found
        environment(d2mu.deta) <- environment(object$linkfun)
        return(d2mu.deta)
    }
    switch(link,
           "logit" = function(eta) {
               mu.eta(eta) * (1 - 2 * linkinv(eta))
           },
           "probit" = function(eta) {
               -eta * pmax(dnorm(eta),.Machine$double.eps)
           },
           "cauchit" = function(eta) {
               -2 * pi * eta * pmax(dcauchy(eta)^2, .Machine$double.eps)
           },
           "cloglog" = function(eta) {
               (1 - exp(eta)) * pmax(exp(eta) * exp(-exp(eta)), .Machine$double.eps)
           },
           "identity" = function(eta) {
               rep.int(0, length(eta))
           },
           "log" = function(eta) {
               pmax(exp(eta), .Machine$double.eps)
           },
           "sqrt" = function(eta) {
               rep.int(2, length(eta))
           },
           "1/mu^2" = function(eta) {
               3/(4 * eta^2.5)
                                          },
           "inverse" = function(eta) {
               2/(eta^3)
           },
           ## else :
           stop(sQuote(link), " link not recognised")
           )# end switch(.)
}


`compute_d2mu.deta` <- function(object, ...) {
    UseMethod('compute_d2mu.deta')
}

#' @method compute_d3mu.deta link-glm
`compute_d3mu.deta.link-glm` <- function(object, ...) {
    mu.eta <- object$mu.eta
    linkinv <- object$linkinv
    link <- object$name
    if (grepl("mu\\^", link) & (link != "1/mu^2")) {
        d3mu.deta <- function(eta) {
            (1/lambda) * (1/lambda - 1) * (1/lambda - 2) * eta^(1/lambda - 3)
        }
        ## set the environment so lambda can be found
        environment(d3mu.deta) <- environment(object$linkfun)
        return(d3mu.deta)
    }
    switch(link,
           "logit" = function(eta) {
               mu.eta(eta) * (1 - 6 * mu.eta(eta))
           },
           "probit" = function(eta) {
               (eta^2 - 1) * pmax(dnorm(eta),.Machine$double.eps)
           },
           "cauchit" = function(eta) {
               -2 * pi * pmax(dcauchy(eta)^2, .Machine$double.eps) + 8 * eta^2 * pi^2 * pmax(dcauchy(eta)^3, .Machine$double.eps)
           },
           "cloglog" = function(eta) {
               ((1 - exp(eta))^2 - exp(eta))  * pmax(exp(eta) * exp(-exp(eta)), .Machine$double.eps)
           },
           "identity" = function(eta) {
               rep.int(0, length(eta))
           },
           "log" = function(eta) {
               pmax(exp(eta), .Machine$double.eps)
           },
           "sqrt" = function(eta) {
               rep.int(0, length(eta))
           },
           "1/mu^2" = function(eta) {
               -15/(8 * eta^3.5)
           },
           "inverse" = function(eta) {
               -6/(eta^4)
           },
           ## else :
           stop(sQuote(link), " link not recognised")
           )# end switch(.)
}


`compute_d3mu.deta` <- function(object, ...) {
    UseMethod('compute_d3mu.deta')
}

if (getRversion() >= "2.15.1") globalVariables(c("lambda"))






## ## Call that produced the enrichwith template for the current script:
## create_enrichwith_skeleton(class = "link-glm", option = c("d2mu.deta",
##     "d3mu.deta", "inverse link derivatives"), description = c("2nd derivative of the inverse link function",
##     "3rd derivative of the inverse link function", "2nd and 3rd derivative of the inverse link function"),
##     component = list("d2mu.deta", "d3mu.deta", c("d2mu.deta",
##         "d3mu.deta")), path = "~/Downloads", attempt_rename = TRUE)

#' Enrich objects of class \code{\link{glm}}
#'
#' Enrich objects of class \code{\link{glm}} with any or all of a set
#' of auxiliary functions, the maximum likelihood estimate of the
#' dispersion parameter, the expected or observed information at the
#' maximum likelihood estimator, and the first term in the expansion
#' of the bias of the maximum likelihood estimator.
#'
#' @param object an object of class glm
#' @param with a character vector of options for the enrichment of \code{object}
#' @param ... extra arguments to be passed to the
#'     \code{compute_*} functions
#'
#' @details
#'
#' The \code{auxiliary_functions} component consists of any or all of the following functions:
#' \itemize{
#' \item \code{score}: the log-likelihood derivatives as a function of the model parameters; see \code{\link{get_score_function.glm}}
#' \item \code{information}: the expected or observed information as a function of the model parameters; see \code{\link{get_information_function.glm}}
#' \item \code{bias}: the first-order term in the expansion of the bias of the maximum likelihood estimator as a function of the model parameters; see \code{\link{get_bias_function.glm}}
#' \item \code{simulate}: a \code{\link{simulate}} function for \code{\link{glm}} objects that can simulate variates from the model at user-supplied parameter values for the regression parameters and the dispersion (default is the maximum likelihood estimates); see \code{\link{get_simulate_function.glm}}
#' \item \code{dmodel}: computes densities or probability mass functions under the model at user-supplied \code{\link{data.frame}}s and at user-supplied values for the regression parameters and the dispersion, if any (default is at the maximum likelihood estimates); see \code{\link{get_dmodel_function.glm}}
#' \item \code{pmodel}: computes distribution functions under the model at user-supplied \code{\link{data.frame}}s and at user-supplied values for the regression parameters and the dispersion, if any (default is at the maximum likelihood estimates); see \code{\link{get_pmodel_function.glm}}
#' \item \code{qmodel}: computes quantile functions under the model at user-supplied \code{\link{data.frame}}s and at user-supplied values for the regression parameters and the dispersion, if any (default is at the maximum likelihood estimates); see \code{\link{get_qmodel_function.glm}}
#' \item \code{Pmat}: the matrices \eqn{P_t} as a function of the model parameters (see, Kosmidis, 2014, expression (5))
#' \item \code{Qmat}: the matrices \eqn{Q_t} as a function of the model parameters (see, Kosmidis, 2014, expression (5))
#' }
#'
#' @return
#'
#' The object \code{object} of class \code{\link{glm}} with extra
#' components. See \code{get_enrichment_options.glm()} for the
#' components and their descriptions.
#'
#'
#' @references
#'
#' Kosmidis I (2014). Bias in parametric estimation: reduction and
#' useful side-effects. *WIRE Computational Statistics*, **6**,
#' 185-196. \doi{10.1002/wics.1296}.
#'
#' @export
#' @examples
#'
#' \dontrun{
#' # A Gamma example, from McCullagh & Nelder (1989, pp. 300-2)
#' clotting <- data.frame(
#'    u = c(5,10,15,20,30,40,60,80,100, 5,10,15,20,30,40,60,80,100),
#'    time = c(118,58,42,35,27,25,21,19,18,69,35,26,21,18,16,13,12,12),
#'    lot = factor(c(rep(1, 9), rep(2, 9))))
#' cML <- glm(time ~ lot*log(u), data = clotting, family = Gamma)
#'
#' # The simulate method for the above fit would simulate at coef(cML)
#' # for the regression parameters and MASS::gamma.dispersion(cML) for
#' # the dispersion. It is not possible to simulate at different
#' # parameter values than those, at least not, without "hacking" the
#' # cML object.
#'
#' # A general simulator for cML results via its enrichment with
#' # auxiliary functions:
#' cML_functions <- get_auxiliary_functions(cML)
#' # which is a shorthand for
#' # enriched_cML <- enrich(cML, with = "auxiliary functions")
#' # cML_functions <- enriched_cML$auxiliary_functions
#'
#' # To simulate 2 samples at the maximum likelihood estimator do
#' dispersion_mle <- MASS::gamma.dispersion(cML)
#' cML_functions$simulate(coef = coef(cML),
#'                        dispersion = dispersion_mle,
#'                        nsim = 2, seed = 123)
#' # To simulate 5 samples at c(0.1, 0.1, 0, 0) and dispersion 0.2 do
#' cML_functions$simulate(coef = c(0.1, 0.1, 0, 0),
#'                        dispersion = 0.2,
#'                        nsim = 5, seed = 123)
#'
#' }
#'
#' \dontrun{
#'
#' ## Reproduce left plot in Figure 4.1 in Kosimdis (2007)
#' ## (see https://ikosmidis.com/files/ikosmidis_thesis.pdf)
#' mod <- glm(1 ~ 1, weights = 10, family = binomial())
#' enriched_mod <- enrich(mod, with = "auxiliary functions")
#' biasfun <- enriched_mod$auxiliary_functions$bias
#' probabilities <- seq(1e-02, 1 - 1e-02, length = 100)
#' biases <- Vectorize(biasfun)(qlogis(probabilities))
#' plot(probabilities, biases, type = "l", ylim = c(-0.5, 0.5),
#'      xlab = expression(pi), ylab = "first-order bias")
#' abline(h = 0, lty = 2); abline(v = 0.5, lty = 2)
#' title("First-order bias of the MLE of the log-odds", sub = "m = 10")
#' }
`enrich.glm` <- function(object, with = "all", ...) {
    if (is.null(with)) {
        return(object)
    }
    dots <- list(...)
    enrichment_options <- get_enrichment_options(object, option = with)
    component <- unlist(enrichment_options$component)
    compute <- unlist(enrichment_options$compute_function)
    for (j in seq.int(length(component))) {
        ccall <- call(compute[j], object = object)
        for (nam in names(dots)) {
            ccall[[nam]] <- dots[[nam]]
        }
        object[[component[j]]] <- eval(ccall)
    }
    if (is.null(attr(object, "enriched"))) {
        attr(object, "enriched") <- TRUE
        classes <- class(object)
        class(object) <- c(paste0("enriched_", classes[1]), classes)
    }
    object
}



#' Available options for the enrichment objects of class
#' \code{\link{glm}}
#'
#' @param object the object to be enriched
#' @param option a character vector listing the options for enriching
#'     the object
#' @param all_options if \code{TRUE} then output a data frame with the
#'     available enrichment options, their descriptions, the names of
#'     the components that each option results in, and the names of
#'     the corresponding \code{compute_*} functions.
#' @return an object of class \code{enrichment_options}
#'
#' @details A check is being made whether the requested option is
#'     available. No check is being made on whether the functions that
#'     produce the components exist.
#' @examples
#' \dontrun{
#' get_enrichment_options.glm(option = "all")
#' get_enrichment_options.glm(all_options = TRUE)
#' }
#' @export
`get_enrichment_options.glm` <- function(object, option, all_options = missing(option)) {
    ## List the enrichment options that you would like to make
    ## available for objects of class
    out <- list()
    out$option <- c('auxiliary functions', 'score vector', 'mle of dispersion', 'expected information', 'observed information', 'first-order bias')
    ## Provide the descriptions of the enrichment options
    out$description <- c('various likelihood-based quantities (gradient of the log-likelihood, expected and observed information matrix and first term in the expansion of the bias of the mle) and a simulate method as functions of the model parameters', 'gradient of the log-likelihood at the mle', 'mle of the dispersion parameter', 'expected information matrix evaluated at the mle', 'observed information matrix evaluated at the mle', 'first term in the expansion of the bias of the mle at the mle')
    ## Add all as an option
    out$option <- c(out$option, 'all')
    out$description <- c(out$description, 'all available options')
    out$component <- list('auxiliary_functions', 'score_mle', 'dispersion_mle', 'expected_information_mle', 'observed_information_mle', 'bias_mle')
    out$component[[length(out$component) + 1]] <- unique(unlist(out$component))
    names(out$component) <- names(out$description) <- out$option
    out$compute_function <- lapply(out$component, function(z) paste0('compute_', z))
    class(out) <- 'enrichment_options'
    if (all_options) {
        return(out)
    }
    invalid_options <- !(option %in% out$option)
    if (any(invalid_options)) {
        stop(gettextf('some options have not been implemented: %s', paste0('"', paste(option[invalid_options], collapse = ', '), '"')))
    }

    out <- list(option = option,
                description = out$description[option],
                component = out$component[option],
                compute_function = out$compute_function[option])
    class(out) <- 'enrichment_options'
    out
}

#' @method compute_auxiliary_functions glm
`compute_auxiliary_functions.glm` <- function(object, ...) {
    if (is.null(object$model)) {
        object <- update(object, model = TRUE)
    }

    ## Extract formula
    formula <- formula(object)

    ## Enrich link-glm and family objects
    link <- enrich(make.link(object$family$link), with = "all")
    family <- enrich(object$family, with = "all")

    ## Extract functions from link
    linkfun <- link$linkfun
    linkinv <- link$linkinv
    d1mu <- link$mu.eta
    d2mu <- link$d2mu.deta

    ## Extract functions from family
    variance <- family$variance
    d1variance <- family$d1variance
    d1afun <- family$d1afun
    d2afun <- family$d2afun
    d3afun <- family$d3afun
    dev.resids <- family$dev.resids
    aic <- family$aic

    ## Get x, y, ...
    y <- object$y
    x <- model.matrix(object)
    nobs <- nobs(object)
    off <- model.offset(object$model)
    prior_weights <- weights(object, type = "prior")
    keep <- prior_weights > 0
    df_residual <- sum(keep) - object$rank

    ## Take care of aliasing
    na_coefficients <- is.na(coef(object))
    has_na <- any(na_coefficients)

    nvar <- sum(!na_coefficients)

    if (is.null(off)) {
        off <- rep(0, nobs)
    }

    score <- function(coefficients, dispersion, contributions = FALSE) {
        if (missing(coefficients)) {
            coefficients <- coef(object)
        }
        if (missing(dispersion)) {
            dispersion <- enrich(object, with = "mle of dispersion")$dispersion_mle
        }
        if (has_na) {
            predictors <- drop(x[, !na_coefficients] %*% coefficients[!na_coefficients] + off)
        }
        else {
            predictors <- drop(x %*% coefficients + off)
        }
        fitted_values <- linkinv(predictors)
        d1mus <- d1mu(predictors)
        variances <- variance(fitted_values)
        ## Score for coefficients
        ## score_beta <- colSums(prior_weights * d1mus * (y - fitted_values) * x / variances)/dispersion
        score_beta <- prior_weights * d1mus * (y - fitted_values) * x / variances /dispersion
        ## Score for dispersion
        if (family$family %in% c("poisson", "binomial")) {
            score_dispersion <- NULL
            vnames <- colnames(score_beta)
        }
        else {
            zetas <- -prior_weights/dispersion
            d1afuns <- rep(NA, nobs)
            d1afuns[keep] <- d1afun(zetas[keep])
            if (family$family == "Gamma") d1afuns <- d1afuns - 2
            devianceResiduals <- family$dev.resids(y, fitted_values, prior_weights)
            EdevianceResiduals <- prior_weights * d1afuns
            ## score_dispersion <- sum(devianceResiduals - EdevianceResiduals, na.rm = TRUE) / (2 * dispersion^2)
            score_dispersion <- (devianceResiduals - EdevianceResiduals) / (2 * dispersion^2)
            vnames <- c(colnames(score_beta), "dispersion")
        }
        if (has_na) {
            score_beta[, na_coefficients] <- NA
        }
        ## Overall score
        out <- cbind(score_beta, score_dispersion)
        colnames(out) <- vnames

        out <- if (contributions) out else colSums(out)
        attr(out, "coefficients") <- coefficients
        attr(out, "dispersion") <- dispersion
        out
    }

    information <- function(coefficients, dispersion,
                            type = c("expected", "observed"), QR = FALSE, CHOL = FALSE) {
        if (missing(coefficients)) {
            coefficients <- coef(object)
        }
        if (missing(dispersion)) {
            dispersion <- enrich(object, with = "mle of dispersion")$dispersion_mle
        }
        if (has_na) {
            predictors <- drop(x[, !na_coefficients] %*% coefficients[!na_coefficients] + off)
        }
        else {
            predictors <- drop(x %*% coefficients + off)
        }
        type <- match.arg(type)
        fitted_values <- linkinv(predictors)
        d1mus <- d1mu(predictors)
        variances <- variance(fitted_values)
        working_weights <- prior_weights * d1mus^2 / variances
        wx <- x * sqrt(working_weights)
        if (QR) {
            return(qr(wx))
        }
        ## expected info coefficients-coefficients
        info_beta <- crossprod(wx) / dispersion
        if (type == "observed") {
            d2mus <- d2mu(predictors)
            d1variances <- d1variance(fitted_values)
            w <- prior_weights * (d2mus / variances - d1mus^2 * d1variances / variances^2) * (y - fitted_values)
            ## observed info coefficients-coefficients
            info_beta <- info_beta - t(x * w) %*% x / dispersion
        }
        rownames(info_beta) <- colnames(info_beta) <- colnames(x)
        if (has_na) {
            info_beta[na_coefficients, ] <- NA
            info_beta[, na_coefficients] <- NA
        }
        ## If there is no dispersion parameter then return the
        ## information for the coefficients only
        if (family$family %in% c("poisson", "binomial")) {
            out <- info_beta
            colnames(out) <- rownames(out) <- colnames(x)
        }
        ## If there is a dispersion parameter then return the
        ## information on the coefficients and the dispersion
        else {
            ## expected info coefficients-dispersion
            info_cross <- rep(0, ncol(info_beta))
            ## expected info dispersion-dispersion
            zetas <- -prior_weights/dispersion
            d2afuns <- rep(NA, nobs)
            d2afuns[keep] <- d2afun(zetas[keep])
            info_dispe <- sum(prior_weights^2 * d2afuns, na.rm = TRUE) / (2 * dispersion^4)
            if (type == "observed") {
                ## observed info coefficients-dispersion
                info_cross <- info_cross + colSums(prior_weights * d1mus * (y - fitted_values) * x / variances)/dispersion^2
                ## observed info dispersion-dispersion
                zetas <- -prior_weights/dispersion
                d1afuns <- rep(NA, nobs)
                d1afuns[keep] <- d1afun(zetas[keep])
                if (family$family == "Gamma") d1afuns <- d1afuns - 2
                devianceResiduals <- family$dev.resids(y, fitted_values, prior_weights)
                EdevianceResiduals <- prior_weights * d1afuns
                info_dispe <- info_dispe + sum(devianceResiduals - EdevianceResiduals, na.rm = TRUE) / dispersion^3
            }
            if (has_na) {
                info_cross[na_coefficients] <- NA
            }
            out <- rbind(cbind(info_beta, info_cross),
                         c(info_cross, info_dispe))
            colnames(out) <- rownames(out) <- c(colnames(x), "dispersion")
        }
        if (CHOL)
            out <- chol(out)
        attr(out, "coefficients") <- coefficients
        attr(out, "dispersion") <- dispersion
        out
    }

    ## E(score %*% t(score) %*% score[t])
    Pmat <- function(coefficients, dispersion) {
        if (missing(coefficients)) {
            coefficients <- coef(object)
        }
        if (missing(dispersion)) {
            dispersion <- enrich(object, with = "mle of dispersion")$dispersion_mle
        }
        if (has_na) {
            predictors <- drop(x[, !na_coefficients] %*% coefficients[!na_coefficients] + off)
        }
        else {
            predictors <- drop(x %*% coefficients + off)
        }
        fitted_values <- linkinv(predictors)
        d1mus <- d1mu(predictors)
        d2mus <- d2mu(predictors)
        variances <- variance(fitted_values)
        d1variances <- d1variance(fitted_values)
        working_weights <- prior_weights * d1mus^2 / variances

        coefnames <- colnames(x)

        Pbb <- lapply(1:ncol(x), function(t) {
            out <- crossprod(x, x * working_weights * d1mus * d1variances / variances * x[, t]) / dispersion
            out[na_coefficients, ] <- out[, na_coefficients] <- NA
            dimnames(out) <- list(coefnames, coefnames)
            out
        })

        if (family$family %in% c("poisson", "binomial")) {
            P <- Pbb
            names(P) <- coefnames
        } else {
            wx <- x * sqrt(working_weights)
            xwx <- crossprod(wx)
            zeros <- rep(0, ncol(x))
            zeros[na_coefficients] <- NA
            xwx[na_coefficients, ] <- xwx[, na_coefficients] <- NA
            Pbp <- xwx / dispersion^2
            P <- lapply(seq_along(Pbb), function(r) {
                Pbphi <- Pbp[r, , drop = TRUE]
                Pbphi[na_coefficients] <- NA
                out <- rbind(cbind(Pbb[[r]], Pbphi), c(Pbphi, 0))
                colnames(out) <- rownames(out) <- c(coefnames, "dispersion")
                out
            })
            zetas <- -prior_weights / dispersion
            d3afuns <- rep(NA, nobs)
            d3afuns[keep] <- d3afun(zetas[keep])
            Pa <- rbind(cbind(xwx / dispersion^2, zeros),
                        c(zeros, 0.5 * sum(prior_weights^3 * d3afuns, na.rm = TRUE) / dispersion^6))
            dimnames(Pa) <- dimnames(P[[1]])
            P <- c(P, list(Pa))
            names(P) <- c(coefnames, "dispersion")
        }
        attr(P, "coefficients") <- coefficients
        attr(P, "dispersion") <- dispersion
        P
    }

    ## -E(observed_information %*% score[t])
    Qmat <- function(coefficients, dispersion) {
        if (missing(coefficients)) {
            coefficients <- coef(object)
        }
        if (missing(dispersion)) {
            dispersion <- enrich(object, with = "mle of dispersion")$dispersion_mle
        }
        if (has_na) {
            predictors <- drop(x[, !na_coefficients] %*% coefficients[!na_coefficients] + off)
        }
        else {
            predictors <- drop(x %*% coefficients + off)
        }
        fitted_values <- linkinv(predictors)
        d1mus <- d1mu(predictors)
        d2mus <- d2mu(predictors)
        variances <- variance(fitted_values)
        d1variances <- d1variance(fitted_values)
        working_weights <- prior_weights * d1mus^2 / variances

        coefnames <- colnames(x)

        Qbb <- lapply(1:ncol(x), function(t) {
            out <- - crossprod(x, x * working_weights * d1mus * (d1variances / variances - d2mus / d1mus^2) * x[, t]) / dispersion
            out[na_coefficients, ] <- out[, na_coefficients] <- NA
            dimnames(out) <- list(coefnames, coefnames)
            out
        })

        if (family$family %in% c("poisson", "binomial")) {
            Q <- Qbb
            names(Q) <- coefnames
        } else {
            wx <- x * sqrt(working_weights)
            xwx <- crossprod(wx)
            zeros <- rep(0, ncol(x))
            zeros[na_coefficients] <- NA
            xwx[na_coefficients, ] <- xwx[, na_coefficients] <- NA
            Qbp <- - xwx / dispersion^2
            Q <- lapply(seq_along(Qbb), function(r) {
                Qbphi <- Qbp[r, , drop = TRUE]
                Qbphi[na_coefficients] <- NA
                out <- rbind(cbind(Qbb[[r]], Qbphi), c(Qbphi, 0))
                out[na_coefficients, ] <- out[, na_coefficients] <- NA
                colnames(out) <- rownames(out) <- c(coefnames, "dispersion")
                out
            })
            zetas <- -prior_weights / dispersion
            d2afuns <- rep(NA, nobs)
            d2afuns[keep] <- d2afun(zetas[keep])
            Qa <- matrix(0, ncol(x) + 1, ncol(x) + 1)
            Qa[ncol(x) + 1, ncol(x) + 1] <- -sum(prior_weights^2 * d2afuns, na.rm = TRUE) / dispersion^5
            Qa[na_coefficients, ] <- Qa[, na_coefficients] <- NA
            dimnames(Qa) <- dimnames(Q[[1]])
            Q <- c(Q, list(Qa))
            names(Q) <- c(coefnames, "dispersion")
        }
        attr(Q, "coefficients") <- coefficients
        attr(Q, "dispersion") <- dispersion
        Q
    }

    bias <- function(coefficients, dispersion) {
        if (missing(coefficients)) {
            coefficients <- coef(object)
        }
        if (missing(dispersion)) {
            dispersion <- enrich(object, with = "mle of dispersion")$dispersion_mle
        }
        if (has_na) {
            predictors <- drop(x[, !na_coefficients] %*% coefficients[!na_coefficients] + off)
        }
        else {
            predictors <- drop(x %*% coefficients + off)
        }
        fitted_values <- linkinv(predictors)
        d1mus <- d1mu(predictors)
        d2mus <- d2mu(predictors)
        variances <- variance(fitted_values)
        working_weights <- prior_weights * d1mus^2 / variances
        Qr <- information(coefficients, dispersion = dispersion, QR = TRUE)
        inds <- seq.int(Qr$rank)
        Q <- qr.Q(Qr)[, inds, drop = FALSE]
        if (all(dim(Q) == c(1, 1))) {
            hats <- 1
        }
        else {
            hats <- rowSums(Q * Q)
        }
        ksi <- -0.5 * dispersion * d2mus * hats / (d1mus * sqrt(working_weights))

        bias_beta <- numeric(ncol(x))
        ## coefnames <- names(coefficients)
        ## if (is.null(coefnames)) {
        coefnames <- colnames(x)
        ## }
        names(bias_beta) <- coefnames
        biases <- drop(tcrossprod(ksi %*% Q, solve(qr.R(Qr)[inds, inds, drop = FALSE])))
        bias_beta[names(biases)] <- biases
        if (family$family %in% c("poisson", "binomial")) {
            bias_dispersion <- NULL
            vnames <- names(bias_beta)
        }
        else {
            if (df_residual > 0) {
                ## Enrich family object with the the derivatives of the a
                ## function (see ?enrich.family for details)
                zetas <- -prior_weights/dispersion
                d2afuns <- d3afuns <- rep(NA, nobs)
                d2afuns[keep] <- d2afun(zetas[keep])
                d3afuns[keep] <- d3afun(zetas[keep])
                ## As in brglm2
                s3 <- sum(prior_weights^3 * d3afuns, na.rm = TRUE)
                s2 <- sum(prior_weights^2 * d2afuns, na.rm = TRUE)
                bias_dispersion <- - (nvar - 2) * dispersion^3 / s2  - dispersion^2 * s3 / s2^2
            }
            else {
                bias_dispersion <- NA
            }
            vnames <- c(names(bias_beta), "dispersion")
        }
        if (has_na) {
            bias_beta[na_coefficients] <- NA
        }
        out <- c(bias_beta, bias_dispersion)
        names(out) <- vnames
        attr(out, "coefficients") <- coefficients
        attr(out, "dispersion") <- dispersion
        out
    }

    simulate <- function(coefficients, dispersion, nsim = 1, seed = NULL) {
        if (missing(coefficients)) {
            coefficients <- coef(object)
        }
        else {
            if (!isTRUE(identical(length(coefficients), length(coef(object))))) {
                stop("`coefficients` does not have the right length")
            }

        }
        if (missing(dispersion)) {
            dispersion <- enrich(object, with = "mle of dispersion")$dispersion_mle
        }
        if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
            runif(1)
        if (is.null(seed))
            RNGstate <- get(".Random.seed", envir = .GlobalEnv)
        else {
            R.seed <- get(".Random.seed", envir = .GlobalEnv)
            set.seed(seed)
            RNGstate <- structure(seed, kind = as.list(RNGkind()))
            on.exit(assign(".Random.seed", R.seed, envir = .GlobalEnv))
        }
        predictors <- drop(x %*% coefficients + off)
        fitted_values <- linkinv(predictors)
        fitted_names <- names(fitted_values)
        n <- length(fitted_values)
        variates <- switch(family$family,
                           "gaussian" = {
            rnorm(nsim * n, mean = fitted_values, sd = sqrt(dispersion/prior_weights))
        },
        "Gamma" = {
            if (any(prior_weights!= 1)) {
                message("using prior weights in the shape parameters")
            }
            rgamma(nsim * n, shape = prior_weights/dispersion, scale = fitted_values*dispersion)
        },
        "binomial" = {
            if (any(prior_weights %% 1 != 0))
                stop("cannot simulate from non-integer prior.weights")
            if (!is.null(mf <- object$model)) {
                y <- model.response(mf)
                if (is.factor(y)) {
                    yy <- factor(1 + rbinom(n * nsim, size = 1, prob = fitted_values),
                                 labels = levels(y))
                    split(yy, rep(seq_len(nsim), each = n))
                }
                else if (is.matrix(y) && ncol(y) == 2) {
                    yy <- vector("list", nsim)
                    for (i in seq_len(nsim)) {
                        Y <- rbinom(n, size = prior_weights, prob = fitted_values)
                        YY <- cbind(Y, prior_weights - Y)
                        colnames(YY) <- colnames(y)
                        yy[[i]] <- YY
                    }
                    yy
                }
                else rbinom(n * nsim, size = prior_weights, prob = fitted_values)/prior_weights
            }
            else rbinom(n * nsim, size = prior_weights, prob = fitted_values)/prior_weights
        },
        "poisson" = {
            if (any(prior_weights != 1)) {
                warning("ignoring prior weights")
            }
            rpois(n * nsim, lambda = fitted_values)
        },
        "inverse.gaussian" = {
            if (!requireNamespace("SuppDists", quietly = TRUE))
                stop("need CRAN package 'SuppDists' for simulation from the 'inverse.gaussian' family")
            SuppDists::rinvGauss(n * nsim, nu = fitted_values, lambda = prior_weights/dispersion)
        },
        NULL)
        ## Inspired by stats:::simulate.lm
        if (!is.list(variates)) {
            dim(variates) <- c(n, nsim)
            variates <- as.data.frame(variates)
        }
        else {
            class(variates) <- "data.frame"
        }
        names(variates) <-  paste("sim", seq_len(nsim), sep = "_")
        if (!is.null(fitted_names)) {
            row.names(variates) <- fitted_names
        }
        attr(variates, "seed") <- RNGstate
        attr(variates, "coefficients") <- coefficients
        attr(variates, "dispersion") <- dispersion
        variates
    }

    ## data should have a response
    dmodel <- function(data, coefficients, dispersion, log = FALSE) {
        if (missing(coefficients)) {
            coefficients <- coef(object)
        }
        if (missing(dispersion)) {
            dispersion <- enrich(object, with = "mle of dispersion")$dispersion_mle
        }
        if (missing(data)) {
            mf <- object$model
        }
        else  {
            mf <- model.frame(formula = formula, data = data)
        }
        contr <- attr(model.matrix(object), "contrasts")
        new_x <- model.matrix(object = formula, data = mf, terms = terms, contrasts.arg = contr)
        new_y <- model.response(mf)
        new_off <- model.offset(mf)
        if (is.null(new_off)) {
            new_off <- rep(0, nrow(mf))
        }
        if (missing(data)) {
            new_prior_weights <- model.weights(mf)
        }
        else {
            new_prior_weights <- with(data, eval(object$call$weights))
        }
        if (is.null(new_prior_weights)) {
            new_prior_weights <- rep(1, nrow(mf))
        }
        if (missing(coefficients)) {
            coefficients <- coef(object)
        }
        if (missing(dispersion)) {
            dispersion <- enrich(object, with = "mle of dispersion")$dispersion_mle
        }
        if (has_na) {
            predictors <- drop(new_x[, !na_coefficients] %*% coefficients[!na_coefficients] + new_off)
        }
        else {
            predictors <- drop(new_x %*% coefficients + new_off)
        }
        fitted_values <- linkinv(predictors)
        d1mus <- d1mu(predictors)
        variances <- variance(fitted_values)
        dfun <- switch(family$family,
                       "gaussian" = {
            dnorm(new_y, mean = fitted_values, sd = sqrt(dispersion/new_prior_weights), log = log)
        },
        "Gamma" = {
            if (any(new_prior_weights!= 1)) {
                message("using prior weights in the shape parameters")
            }
            dgamma(new_y, shape = new_prior_weights/dispersion, scale = fitted_values*dispersion, log = log)
        },
        "binomial" = {
            if (any(new_prior_weights %% 1 != 0)) {
                stop("cannot simulate from non-integer prior.weights")
            }
            if (is.matrix(new_y) && ncol(new_y)) {
                new_prior_weights <- rowSums(new_y)
                new_y <- new_y[, 1]
                dbinom(new_y, size = new_prior_weights, prob = fitted_values, log = log)
            }
            else {
                if (is.factor(new_y)) {
                    new_y <- as.numeric(new_y) - 1
                    dbinom(new_y, size = 1, prob = fitted_values, log = log)
                }
                else {
                    dbinom(new_y*new_prior_weights, size = new_prior_weights, prob = fitted_values, log = log)
                }
            }
        },
        "poisson" = {
            if (any(new_prior_weights != 1)) {
                warning("ignoring prior weights")
            }
            dpois(new_y, lambda = fitted_values, log = log)
        },
        "inverse.gaussian" = {
            if (!requireNamespace("SuppDists", quietly = TRUE))
                stop("need CRAN package 'SuppDists' for simulation from the 'inverse.gaussian' family")
            SuppDists::dinvGauss(new_y, nu = fitted_values, lambda = new_prior_weights/dispersion, log = log)
        },
        NULL)
        attr(dfun, "coefficients") <- coefficients
        attr(dfun, "dispersion") <- dispersion
        dfun
    }

    ## data should have a response
    pmodel <- function(data, coefficients, dispersion, lower.tail = TRUE, log.p = FALSE) {
        if (missing(coefficients)) {
            coefficients <- coef(object)
        }
        if (missing(dispersion)) {
            dispersion <- enrich(object, with = "mle of dispersion")$dispersion_mle
        }
        contr <- attr(model.matrix(object), "contrasts")
        if (missing(data)) {
            mf <- object$model
        }
        else  {
            mf <- model.frame(formula = formula, data = data)
        }
        new_x <- model.matrix(object = formula, data = mf, terms = terms, contrasts.arg = contr)
        new_y <- model.response(mf)
        new_off <- model.offset(mf)
        if (is.null(new_off)) {
            new_off <- rep(0, nrow(mf))
        }
        if (missing(data)) {
            new_prior_weights <- model.weights(mf)
        }
        else {
            new_prior_weights <- with(data, eval(object$call$weights))
        }
        if (is.null(new_prior_weights)) {
            new_prior_weights <- rep(1, nrow(mf))
        }
        if (missing(coefficients)) {
            coefficients <- coef(object)
        }
        if (missing(dispersion)) {
            dispersion <- enrich(object, with = "mle of dispersion")$dispersion_mle
        }
        if (has_na) {
            predictors <- drop(new_x[, !na_coefficients] %*% coefficients[!na_coefficients] + new_off)
        }
        else {
            predictors <- drop(new_x %*% coefficients + new_off)
        }
        fitted_values <- linkinv(predictors)
        d1mus <- d1mu(predictors)
        variances <- variance(fitted_values)
        pfun <- switch(family$family,
                       "gaussian" = {
            pnorm(new_y, mean = fitted_values, sd = sqrt(dispersion/new_prior_weights), lower.tail = lower.tail, log.p = log.p)
        },
        "Gamma" = {
            if (any(new_prior_weights!= 1)) {
                message("using prior weights in the shape parameters")
            }
            pgamma(new_y, shape = new_prior_weights/dispersion, scale = fitted_values*dispersion, lower.tail = lower.tail, log.p = log.p)
        },
        "binomial" = {
            if (any(new_prior_weights %% 1 != 0)) {
                stop("cannot simulate from non-integer prior.weights")
            }
            if (is.matrix(new_y) && ncol(new_y)) {
                new_prior_weights <- rowSums(new_y)
                new_y <- new_y[, 1]
                pbinom(new_y, size = new_prior_weights, prob = fitted_values, log.p = log.p)
            }
            else {
                if (is.factor(new_y)) {
                    new_y <- as.numeric(new_y) - 1
                    pbinom(new_y, size = 1, prob = fitted_values, log.p = log.p)
                }
                else {
                    pbinom(new_y*new_prior_weights, size = new_prior_weights, prob = fitted_values, log.p = log.p)
                }
            }
        },
        "poisson" = {
            if (any(new_prior_weights != 1)) {
                warning("ignoring prior weights")
            }
            ppois(new_y, lambda = fitted_values, log.p = log.p)
        },
        "inverse.gaussian" = {
            if (!requireNamespace("SuppDists", quietly = TRUE))
                stop("need CRAN package 'SuppDists' for simulation from the 'inverse.gaussian' family")
            SuppDists::pinvGauss(new_y, nu = fitted_values, lambda = new_prior_weights/dispersion, lower.tail = lower.tail, log.p = log.p)
        },
        NULL)
        attr(pfun, "coefficients") <- coefficients
        attr(pfun, "dispersion") <- dispersion
        pfun
    }


    ## any response in the data is ignored
    qmodel <- function(p, data, coefficients, dispersion, lower.tail = TRUE, log.p = FALSE) {
        if (missing(coefficients)) {
            coefficients <- coef(object)
        }
        if (missing(dispersion)) {
            dispersion <- enrich(object, with = "mle of dispersion")$dispersion_mle
        }
        ## output an function that takes as input a data frame and returns densities
        contr <- attr(model.matrix(object), "contrasts")
        if (missing(data)) {
            mf <- object$model
        }
        else  {
            mf <- model.frame(formula = formula, data = data)
        }
        if (length(p) != nrow(mf)) {
            stop("length(p) must be equal to nrow(data)")
        }
        new_x <- model.matrix(object = formula, data = mf, terms = terms, contrasts.arg = contr)
        new_y <- model.response(mf)
        new_off <- model.offset(mf)
        if (is.null(new_off)) {
            new_off <- rep(0, nrow(mf))
        }
        if (missing(data)) {
            new_prior_weights <- model.weights(mf)
        }
        else {
            new_prior_weights <- with(data, eval(object$call$weights))
        }
        if (is.null(new_prior_weights)) {
            new_prior_weights <- rep(1, nrow(mf))
        }
        if (missing(coefficients)) {
            coefficients <- coef(object)
        }
        if (missing(dispersion)) {
            dispersion <- enrich(object, with = "mle of dispersion")$dispersion_mle
        }
        if (has_na) {
            predictors <- drop(new_x[, !na_coefficients] %*% coefficients[!na_coefficients] + new_off)
        }
        else {
            predictors <- drop(new_x %*% coefficients + new_off)
        }
        fitted_values <- linkinv(predictors)
        d1mus <- d1mu(predictors)
        variances <- variance(fitted_values)
        qfun <- switch(family$family,
                       "gaussian" = {
            qnorm(p, mean = fitted_values, sd = sqrt(dispersion/new_prior_weights), lower.tail = lower.tail, log.p = log.p)
        },
        "Gamma" = {c
            if (any(new_prior_weights!= 1)) {
                message("using prior weights in the shape parameters")
            }
            qgamma(p, shape = new_prior_weights/dispersion, scale = fitted_values*dispersion, lower.tail = lower.tail, log.p = log.p)
        },
        "binomial" = {
            if (any(new_prior_weights %% 1 != 0)) {
                stop("cannot simulate from non-integer prior.weights")
            }
            if (is.matrix(new_y) && ncol(new_y)) {
                new_prior_weights <- rowSums(new_y)
                new_y <- new_y[, 1]
                qbinom(p, size = new_prior_weights, prob = fitted_values, lower.tail = lower.tail, log.p = log.p)
            }
            else {
                if (is.factor(new_y)) {
                    new_y <- as.numeric(new_y) - 1
                    qbinom(p, size = 1, prob = fitted_values, lower.tail = lower.tail, log.p = log.p)
                }
                else {
                    qbinom(p, size = new_prior_weights, prob = fitted_values, lower.tail = lower.tail, log.p = log.p)
                }
            }
        },
        "poisson" = {
            if (any(new_prior_weights != 1)) {
                warning("ignoring prior weights")
            }
            qpois(p, lambda = fitted_values, log.p = log.p)
        },
        "inverse.gaussian" = {
            if (!requireNamespace("SuppDists", quietly = TRUE))
                stop("need CRAN package 'SuppDists' for simulation from the 'inverse.gaussian' family")
            SuppDists::qinvGauss(p, nu = fitted_values, lambda = new_prior_weights/dispersion, lower.tail = lower.tail, log.p = log.p)
        },
        NULL)
        attr(qfun, "coefficients") <- coefficients
        attr(qfun, "dispersion") <- dispersion
        qfun
    }


    ## any response in the data is ignored
    ## To be implemented at a later release
    rmodel <- function(n, data, coefficients, dispersion, nsim = 1, seed = NULL) {
        if (missing(coefficients)) {
            coefficients <- coef(object)
        }
        if (missing(dispersion)) {
            dispersion <- enrich(object, with = "mle of dispersion")$dispersion_mle
        }
    }

    return(list(score = score,
                information = information,
                bias = bias,
                simulate = simulate,
                dmodel = dmodel,
                pmodel = pmodel,
                qmodel = qmodel,
                Pmat = Pmat,
                Qmat = Qmat))

}


`compute_auxiliary_functions` <- function(object, ...) {
    UseMethod('compute_auxiliary_functions')
}

#' @method compute_score_mle glm
`compute_score_mle.glm` <- function(object, ...) {
    get_score_function(object)()
}


`compute_score_mle` <- function(object, ...) {
    UseMethod('compute_score_mle')
}

#' @method compute_dispersion_mle glm
`compute_dispersion_mle.glm` <- function(object, ...) {
    if (object$family$family %in% c("poisson", "binomial")) {
        dispersion_mle <- 1
    }
    else {
        object <- enrich(object, with = "auxiliary functions")
        nobs <- nobs(object)
        prior_weights <- weights(object, type = "prior")
        keep <- prior_weights > 0
        df_residual <- sum(keep) - object$rank

        gradfun <- function(logdispersion) {
            beta <- coef(object, model = "mean")
            object$auxiliary_functions$score(beta, exp(logdispersion))[length(beta) + 1]
        }
        if (df_residual > 0) {
            dispFit <- try(uniroot(f = gradfun, lower = 0.5*log(.Machine$double.eps), upper = 20, tol = 1e-08, maxiter = 10000), silent = FALSE)
            if (inherits(dispFit, "try-error")) {
                warning("the mle of dispersion could not be calculated")
                dispersion_mle <- NA
            }
            else {
                dispersion_mle <- exp(dispFit$root)
            }
        }
        else {
            ## if the model is saturated dispersion_mle is NA
            dispersion_mle <- NA
    }
        names(dispersion_mle) <- "dispersion"
        dispersion_mle
    }
}

`compute_dispersion_mle` <- function(object, ...) {
    UseMethod('compute_dispersion_mle')
}

#' @method compute_expected_information_mle glm
`compute_expected_information_mle.glm` <- function(object, dispersion, ...) {
    get_information_function(object)(dispersion = dispersion, type = "expected")
}


`compute_expected_information_mle` <- function(object, ...) {
    UseMethod('compute_expected_information_mle')
}

#' @method compute_observed_information_mle glm
`compute_observed_information_mle.glm` <- function(object, dispersion = dispersion, ...) {
    get_information_function(object)(dispersion = dispersion, type = "observed")
}


`compute_observed_information_mle` <- function(object, ...) {
    UseMethod('compute_observed_information_mle')
}

#' @method compute_bias_mle glm
`compute_bias_mle.glm` <- function(object, ...) {
    get_bias_function(object)()
}


`compute_bias_mle` <- function(object, ...) {
    UseMethod('compute_bias_mle')
}


## Other extractor functions

#' Function to extract model coefficients from objects of class \code{enriched_glm}
#'
#' @param object an object of class \code{enriched_glm}
#' @param model either "mean" for the estimates of the parameters in the linear predictor, or "dispersion" for the estimate of the dispersion, or "full" for all estimates
#' @param ... currently unused
#' @export
coef.enriched_glm <- function(object, model = c("mean", "full", "dispersion"), ...) {
    beta <- object$coefficients
    switch(match.arg(model),
           mean = {
               beta
           },
           dispersion = {
               object <- enrich(object, with = "mle of dispersion")
               object$dispersion_mle
           },
           full = {
               object <- enrich(object, with = "mle of dispersion")
               c(beta, object$dispersion_mle)
           })
}

#' Function to compute/extract auxiliary functions from objects of
#' class \code{glm}/\code{enriched_glm}
#'
#' @param object an object of class \code{glm} or\code{enriched_glm}
#' @param ... currently not used
#'
#' @details
#'
#' See \code{\link{enrich.glm}} for details.
#'
#' @export
get_auxiliary_functions.glm <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        enriched_object <- enrich(object, with = "auxiliary functions")
        enriched_object$auxiliary_functions
    }
    else {
        object$auxiliary_functions
    }
}

#' Function to compute/extract a simulate function for response
#' vectors from an object of class \code{glm}/\code{enriched_glm}
#'
#' @param object an object of class \code{glm} or\code{enriched_glm}
#' @param ... currently not used
#'
#' @details
#' The computed/extracted simulate function has arguments
#' \describe{
#'
#' \item{coefficients}{the regression coefficients at which the
#' response vectors are simulated. If missing then the maximum
#' likelihood estimates are used}
#'
#' \item{dispersion}{the dispersion parameter at which the response
#' vectors are simulated. If missing then the maximum likelihood
#' estimate is used}
#'
#' \item{nsim}{number of response vectors to simulate.  Defaults to \code{1}}
#'
#' \item{seed}{an object specifying if and how the random number
#' generator should be initialized ('seeded'). It can be either
#' \code{NULL} or an integer that will be used in a call to
#' \code{set.seed} before simulating the response vectors.  If set,
#' the value is saved as the \code{seed} attribute of the returned
#' value.  The default, \code{NULL} will not change the random
#' generator state, and return \code{.Random.seed} as the \code{seed}
#' attribute, see \code{Value}}
#'
#' }
#'
#' @export
get_simulate_function.glm <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        get_auxiliary_functions(object)$simulate
    }
    else {
        object$auxiliary_functions$simulate
    }
}


#' Function to compute/extract a function that returns the scores
#' (derivatives of the log-likelihood) for an object of class
#' \code{glm}/\code{enriched_glm}
#'
#' @param object an object of class \code{glm} or\code{enriched_glm}
#' @param ... currently not used
#'
#' @details
#' The computed/extracted function has arguments
#' \describe{
#'
#' \item{coefficients}{the regression coefficients at which the scores
#' are computed. If missing then the maximum likelihood estimates are
#' used}
#'
#' \item{dispersion}{the dispersion parameter at which the score
#' function is evaluated. If missing then the maximum likelihood
#' estimate is used}
#'
#' }
#'
#' @export
get_score_function.glm <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        get_auxiliary_functions(object)$score
    }
    else {
        object$auxiliary_functions$score
    }
}

#' Function to compute/extract a function that returns the information
#' matrix for an object of class \code{glm}/\code{enriched_glm}
#'
#' @param object an object of class \code{glm} or\code{enriched_glm}
#' @param ... currently not used
#'
#' @details
#' The computed/extracted function has arguments
#' \describe{
#'
#' \item{coefficients}{the regression coefficients at which the
#' information matrix is evaluated. If missing then the maximum
#' likelihood estimates are used}
#'
#' \item{dispersion}{the dispersion parameter at which the information
#' matrix is evaluated. If missing then the maximum likelihood estimate
#' is used}
#'
#' \item{type}{should the function return the 'expected' or 'observed' information? Default is \code{expected}}
#'
#' \item{QR}{If \code{TRUE}, then the QR decomposition of \deqn{W^{1/2} X} is returned, where \deqn{W} is a diagonal matrix with the working weights (\code{object$weights}) and \deqn{X} is the model matrix.}
#'
#' \item{CHOL}{If \code{TRUE}, then the Cholesky decomposition of the information matrix at the coefficients is returned}
#'
#' }
#'
#' @export
get_information_function.glm <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        get_auxiliary_functions(object)$information
    }
    else {
        object$auxiliary_functions$information
    }
}


#' Function to compute/extract a function that returns the first term
#' in the expansion of the bias of the MLE for the parameters of an
#' object of class \code{glm}/\code{enriched_glm}
#'
#' @param object an object of class \code{glm} or\code{enriched_glm}
#' @param ... currently not used
#'
#' @details
#' The computed/extracted function has arguments
#' \describe{
#'
#' \item{coefficients}{the regression coefficients at which the
#' first-order bias is evacuated. If missing then the maximum
#' likelihood estimates are used}
#'
#' \item{dispersion}{the dispersion parameter at which the first-order
#' bias is evaluated. If missing then the maximum likelihood estimate
#' is used}
#'
#' }
#'
#' @export
get_bias_function.glm <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        get_auxiliary_functions(object)$bias
    }
    else {
        object$auxiliary_functions$bias
    }
}


#' Function to compute/extract a \code{dmodel} function
#'
#' @param object an object of class \code{glm} or\code{enriched_glm}
#' @param ... currently not used
#'
#' @details
#' The computed/extracted function has arguments
#' \describe{
#'
#' \item{data}{a data frame with observations at which to compute
#' densities. If missing then densities are computed at the model
#' frame extracted from the object (see \code{\link{glm}})}
#'
#' \item{coefficients}{the regression coefficients at which the
#' densities are computed. If missing then the maximum likelihood
#' estimates are used}
#'
#' \item{dispersion}{the dispersion parameter at which the densities
#' function is computed. If missing then the maximum likelihood
#' estimate is used}
#'
#' \item{log}{logical; if \code{TRUE}, the logarithm of the density is
#' returned}
#'
#' }
#'
#' @export
get_dmodel_function.glm <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        get_auxiliary_functions(object)$dmodel
    }
    else {
        object$auxiliary_functions$dmodel
    }
}

#' Function to compute/extract a \code{pmodel} function
#'
#' @param object an object of class \code{glm} or\code{enriched_glm}
#' @param ... currently not used
#'
#' @details
#' The computed/extracted function has arguments
#' \describe{
#'
#' \item{data}{a data frame with observations at which to compute the
#' distribution function. If missing then probabilities are computed
#' at the model frame extracted from the object (see
#' \code{\link{glm}})}
#'
#' \item{coefficients}{the regression coefficients at which the
#' distribution function are computed. If missing then the maximum
#' likelihood estimates are used}
#'
#' \item{dispersion}{the dispersion parameter at which the
#' distribution function is computed. If missing then the maximum
#' likelihood estimate is used}
#'
#' \item{log.p}{logical; if \code{TRUE}, the logarithm of the
#' distribution function is returned}
#'
#' \item{lower.tail}{logical; if \code{TRUE} (default), probabilities
#' are P[X <= x] otherwise, P[X > x]}
#' }
#'
#' @export
get_pmodel_function.glm <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        get_auxiliary_functions(object)$pmodel
    }
    else {
        object$auxiliary_functions$pmodel
    }
}

#' Function to compute/extract a \code{qmodel} function
#'
#' @param object an object of class \code{glm} or\code{enriched_glm}
#' @param ... currently not used
#'
#' @details
#' The computed/extracted function has arguments
#' \describe{
#'
#' \item{p}{a vector of probabilities with \code{length(p)} equal to
#' \code{nrow(data)} at which to evaluate quantiles}
#'
#' \item{data}{a data frame with observations at which to compute the
#' quantiles. If missing then quantiles are computed at the model
#' frame extracted from the object (see \code{\link{glm}})}
#'
#' \item{coefficients}{the regression coefficients at which the
#' quantiles are computed. If missing then the maximum likelihood
#' estimates are used}
#'
#' \item{dispersion}{the dispersion parameter at which the
#' quantiles are computed. If missing then the maximum
#' likelihood estimate is used}
#'
#' \item{log.p}{logical; if \code{TRUE}, the logarithm of the
#' probabilities is used}
#'
#' \item{lower.tail}{logical; if \code{TRUE} (default), probabilities
#' are P[X <= x] otherwise, P[X > x]}
#'
#' }
#'
#' @export
get_qmodel_function.glm <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        get_auxiliary_functions(object)$qmodel
    }
    else {
        object$auxiliary_functions$qmodel
    }
}


## ## ## Call that produced the enrichwith template for the current script:
## create_enrichwith_skeleton(class = "glm", option = c("auxiliary functions",
##     "score vector", "mle of dispersion", "expected information",
##     "observed information", "first-order bias"), description = c("various likelihood-based quantities (gradient of the log-likelihood, expected and observed information matrix and first term in the expansion of the bias of the mle) and a simulate method as functions of the model parameters",
##     "gradient of the log-likelihood at the mle", "mle of the dispersion parameter",
##     "expected information matrix evaluated at the mle", "observed information matrix evaluated at the mle",
##     "first term in the expansion of the bias of the mle at the mle"),
##     component = list("auxiliary_functions", "score_mle", "dispersion_mle",
##         "expected_information_mle", "observed_information_mle",
##         "bias_mle"), path = "~/Downloads", attempt_rename = FALSE)

#' Enrich objects of class \code{\link{lm}}
#'
#' Enrich objects of class \code{\link{lm}} with any or all of a set
#' auxiliary functions, the maximum likelihood estimate of the
#' dispersion parameter, the expected or observed information at the
#' maximum likelihood estimator, and the first term in the expansion
#' of the bias of the maximum likelihood estimator.
#'
#' @param object an object of class lm
#' @param with a character vector of options for the enrichment of \code{object}
#' @param ... extra arguments to be passed to the
#'     \code{compute_*} functions
#'
#' @details
#'
#' The auxiliary functions consist of the score functions, the
#' expected or observed information, the first-order bias of the
#' maximum likelihood estimator as functions of the model parameters,
#' and a \code{simulate} function that takes as input the model
#' parameters (including the dispersion if any). The result from the
#' \code{simulate} auxiliary function has the same structure to that
#' of the \code{\link{simulate}} method for \code{\link{lm}} objects.
#'
#' @return
#'
#' The object \code{object} of class \code{\link{lm}} with extra
#' components. \code{get_enrichment_options.lm()} returns the
#' components and their descriptions.
#'
#' @method enrich lm
#' @export
`enrich.lm` <- function(object, with = "all", ...) {
    if (is.null(with)) {
        return(object)
    }
    dots <- list(...)
    enrichment_options <- get_enrichment_options(object, option = with)
    component <- unlist(enrichment_options$component)
    compute <- unlist(enrichment_options$compute_function)
    for (j in seq.int(length(component))) {
        ccall <- call(compute[j], object = object)
        for (nam in names(dots)) {
            ccall[[nam]] <- dots[[nam]]
        }
        object[[component[j]]] <- eval(ccall)
    }
    if (is.null(attr(object, "enriched"))) {
        attr(object, "enriched") <- TRUE
        classes <- class(object)
        class(object) <- c(paste0("enriched_", classes[1]), classes)
    }
    object
}



#' Available options for the enrichment objects of class
#' \code{\link{lm}}
#'
#' @param object the object to be enriched
#' @param option a character vector listing the options for enriching
#'     the object
#' @param all_options if \code{TRUE} then output a data frame with the
#'     available enrichment options, their descriptions, the names of
#'     the components that each option results in, and the names of
#'     the corresponding \code{compute_*} functions.
#' @return an object of class \code{enrichment_options}
#'
#' @details A check is being made whether the requested option is
#'     available. No check is being made on whether the functions that
#'     produce the components exist.
#' @examples
#' \dontrun{
#' get_enrichment_options.lm(option = "all")
#' get_enrichment_options.lm(all_options = TRUE)
#' }
#' @export
`get_enrichment_options.lm` <- function(object, option, all_options = missing(option)) {
    ## List the enrichment options that you would like to make
    ## available for objects of class
    out <- list()
    out$option <- c('auxiliary functions', 'score vector', 'mle of dispersion', 'expected information', 'observed information', 'first-order bias')
    ## Provide the descriptions of the enrichment options
    out$description <- c('various likelihood-based quantities (gradient of the log-likelihood, expected and observed information matrix and first term in the expansion of the bias of the mle) and a simulate method as functions of the model parameters', 'gradient of the log-likelihood at the mle', 'mle of the dispersion parameter', 'expected information matrix evaluated at the mle', 'observed information matrix evaluated at the mle', 'first term in the expansion of the bias of the mle at the mle')
    ## Add all as an option
    out$option <- c(out$option, 'all')
    out$description <- c(out$description, 'all available options')
    out$component <- list('auxiliary_functions', 'score_mle', 'dispersion_mle', 'expected_information_mle', 'observed_information_mle', 'bias_mle')
    out$component[[length(out$component) + 1]] <- unique(unlist(out$component))
    names(out$component) <- names(out$description) <- out$option
    out$compute_function <- lapply(out$component, function(z) paste0('compute_', z))
    class(out) <- 'enrichment_options'
    if (all_options) {
        return(out)
    }
    invalid_options <- !(option %in% out$option)
    if (any(invalid_options)) {
        stop(gettextf('some options have not been implemented: %s', paste0('"', paste(option[invalid_options], collapse = ', '), '"')))
    }

    out <- list(option = option,
                description = out$description[option],
                component = out$component[option],
                compute_function = out$compute_function[option])
    class(out) <- 'enrichment_options'
    out
}


#' @method compute_auxiliary_functions lm
`compute_auxiliary_functions.lm` <- function(object, ...) {
    if (is.null(object$model)) {
        object <- update(object, model = TRUE)
    }

    ## Get x, y, ...
    y <- model.response(object$model)
    x <- model.matrix(object)
    nobs <- nobs(object)
    nvar <- ncol(x)
    off <- model.offset(object$model)
    prior_weights <- weights(object)
    if (is.null(prior_weights)) {
        prior_weights <- rep(1, nobs)
    }

    keep <- prior_weights > 0
    df_residual <- sum(keep) - object$rank
    disp <- sum(prior_weights * residuals(object)^2)/nobs
    names(disp) <- "dispersion"

    if (is.null(off)) {
        off <- rep(0, nobs)
    }

    score <- function(coefficients, dispersion, contributions = FALSE) {
        if (missing(coefficients)) {
            coefficients <- coef(object)
        }
        if (missing(dispersion)) {
            dispersion <- disp
        }
        fitted_values <- drop(x %*% coefficients + off)
        ## Residuals
        resid <- y - fitted_values
        ## Score for coefficients
        ## score_beta <- colSums(prior_weights * resid * x)/dispersion
        score_beta <- prior_weights * resid * x / dispersion
        ## Score for dispersion
        ## score_dispersion <- - nobs/(2 * dispersion) + sum(prior_weights * resid^2)/(2*dispersion^2)
        score_dispersion <- - 1/(2 * dispersion) + prior_weights * resid^2 / (2 * dispersion^2)
        vnames <- c(colnames(score_beta), "dispersion")
        ## Overall score
        out <- cbind(score_beta, score_dispersion)
        colnames(out) <- vnames
        out <- if (contributions) out else colSums(out)
        attr(out, "coefficients") <- coefficients
        attr(out, "dispersion") <- dispersion
        out
    }

    information <- function(coefficients, dispersion,
                            type = c("expected", "observed"), QR = FALSE, CHOL = FALSE) {
        if (missing(coefficients)) {
            coefficients <- coef(object)
        }
        if (missing(dispersion)) {
            dispersion <- disp
        }
        type <- match.arg(type)
        fitted_values <- drop(x %*% coefficients + off)
        resid <- y - fitted_values
        wx <- x * sqrt(prior_weights)
        if (QR) {
            return(qr(wx))
        }
        ## expected info coefficients-coefficients
        info_beta <- crossprod(wx) / dispersion
        rownames(info_beta) <- colnames(info_beta) <- colnames(x)
        ## expected info coefficients-dispersion
        info_cross <- rep(0, ncol(info_beta))
        ## expected info dispersion-dispersion
        info_dispe <- sum(prior_weights)/dispersion^2 - nobs / (2 * dispersion^2)
        if (type == "observed") {
            ## observed info coefficients-dispersion
            info_cross <- info_cross + colSums(prior_weights * resid * x) / dispersion^2
            ## observed info dispersion-dispersion
            info_dispe <- sum(resid^2 * prior_weights) / dispersion^3 - nobs / (2 * dispersion^2)
        }
        out <- rbind(cbind(info_beta, info_cross),
                     c(info_cross, info_dispe))
        colnames(out) <- rownames(out) <- c(colnames(x), "dispersion")
        if (CHOL)
            out <- chol(out)
        attr(out, "coefficients") <- coefficients
        attr(out, "dispersion") <- dispersion
        out
    }

    bias <- function(coefficients, dispersion) {
        if (missing(coefficients)) {
            coefficients <- coef(object)
        }
        if (missing(dispersion)) {
            dispersion <- disp
        }
        fitted_values <- drop(x %*% coefficients + off)
        bias_beta <- numeric(ncol(x))
        names(bias_beta) <- names(coefficients)

        if (df_residual > 0) {
            bias_dispersion <- -nvar/nobs * dispersion
        }
        else {
            bias_dispersion <- NA
        }
        out <- c(bias_beta, bias_dispersion)
        names(out) <- c(names(bias_beta), "dispersion")
        attr(out, "coefficients") <- coefficients
        attr(out, "dispersion") <- dispersion
        out
    }

    simulate <- function(coefficients, dispersion, nsim = 1, seed = NULL) {
        if (missing(coefficients)) {
            coefficients <- coef(object)
        }
        else {
            if (!isTRUE(identical(length(coefficients), length(coef(object))))) {
                stop("`coefficients` does not have the right length")
            }
        }
        if (missing(dispersion)) {
            dispersion <- enrich(object, with = "mle of dispersion")$dispersion_mle
        }
        if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
            runif(1)
        if (is.null(seed))
            RNGstate <- get(".Random.seed", envir = .GlobalEnv)
        else {
            R.seed <- get(".Random.seed", envir = .GlobalEnv)
            set.seed(seed)
            RNGstate <- structure(seed, kind = as.list(RNGkind()))
            on.exit(assign(".Random.seed", R.seed, envir = .GlobalEnv))
        }
        predictors <- drop(x %*% coefficients + off)
        fitted_values <- predictors
        fitted_names <- names(fitted_values)
        n <- length(fitted_values)
        variates <- rnorm(nsim * n, mean = fitted_values, sd = sqrt(dispersion/prior_weights))
        ## Inspired by stats:::simulate.lm
        if (!is.list(variates)) {
            dim(variates) <- c(n, nsim)
            variates <- as.data.frame(variates)
        }
        else {
            class(variates) <- "data.frame"
        }
        names(variates) <-  paste("sim", seq_len(nsim), sep = "_")
        if (!is.null(fitted_names)) {
            row.names(variates) <- fitted_names
        }
        attr(variates, "seed") <- RNGstate
        attr(variates, "coefficients") <- coefficients
        attr(variates, "dispersion") <- dispersion
        variates
    }

    return(list(score = score,
                information = information,
                bias = bias,
                simulate = simulate))

}


`compute_auxiliary_functions` <- function(object, ...) {
    UseMethod('compute_auxiliary_functions')
}


#' @method compute_score_mle lm
`compute_score_mle.lm` <- function(object, ...) {
    get_score_function(object)()
}


`compute_score_mle` <- function(object, ...) {
    UseMethod('compute_score_mle')
}

#' @method compute_dispersion_mle lm
`compute_dispersion_mle.lm` <- function(object, ...) {
    prior_weights <- weights(object)
    nobs <- nobs(object)
    if (is.null(prior_weights)) {
        prior_weights <- rep(1, nobs)
    }
    keep <- prior_weights > 0
    df_residual <- sum(keep) - object$rank
    disp <- sum(prior_weights * residuals(object)^2)/nobs
    disp
}

`compute_dispersion_mle` <- function(object, ...) {
    UseMethod('compute_dispersion_mle')
}

#' @method compute_expected_information_mle lm
`compute_expected_information_mle.lm` <- function(object, dispersion, ...) {
    get_information_function(object)(type = "expected")
}

`compute_expected_information_mle` <- function(object, ...) {
    UseMethod('compute_expected_information_mle')
}

#' @method compute_observed_information_mle lm
`compute_observed_information_mle.lm` <- function(object, dispersion, ...) {
    get_information_function(object)(type = "observed")
}

`compute_observed_information_mle` <- function(object, ...) {
    UseMethod('compute_observed_information_mle')
}


#' @method compute_bias_mle lm
`compute_bias_mle.lm` <- function(object, ...) {
    get_bias_function(object)()
}

`compute_bias_mle` <- function(object, ...) {
    UseMethod('compute_bias_mle')
}


## Other extractor functions

#' Function to extract model coefficients from objects of class \code{enriched_lm}
#'
#' @param object an object of class \code{enriched_lm}
#' @param model either "mean" for the estimates of the parameters in the linear predictor, or "dispersion" for the estimate of the dispersion, or "full" for all estimates
#' @param ... currently unused
#' @export
coef.enriched_lm <- function(object, model = c("mean", "full", "dispersion"), ...) {
    beta <- object$coefficients
    switch(match.arg(model),
           mean = {
               beta
           },
           dispersion = {
               object$dispersion
           },
           full = {
               c(beta, object$dispersion)
           })
}

#' Function to compute/extract auxiliary functions from objects of
#' class \code{lm}/\code{enriched_lm}
#'
#' @param object an object of class \code{lm} or\code{enriched_lm}
#' @param ... currently not used
#'
#' @details
#'
#' See \code{\link{enrich.lm}} for details.
#'
#' @export
get_auxiliary_functions.lm <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        enriched_object <- enrich(object, with = "auxiliary functions")
        enriched_object$auxiliary_functions
    }
    else {
        object$auxiliary_functions
    }
}

#' Function to compute/extract a simulate function for response
#' vectors from an object of class \code{lm}/\code{enriched_lm}
#'
#' @param object an object of class \code{lm} or\code{enriched_lm}
#' @param ... currently not used
#'
#' @details
#' The computed/extracted simulate function has arguments
#' \describe{
#'
#' \item{coefficients}{the regression coefficients at which the
#' response vectors are simulated. If missing then the maximum
#' likelihood estimates are used}
#'
#' \item{dispersion}{the dispersion parameter at which the response
#' vectors are simulated. If missing then the maximum likelihood
#' estimate is used}
#'
#' \item{nsim}{number of response vectors to simulate.  Defaults to \code{1}}
#'
#' \item{seed}{an object specifying if and how the random number
#' generator should be initialized ('seeded'). It can be either
#' \code{NULL} or an integer that will be used in a call to
#' \code{set.seed} before simulating the response vectors.  If set,
#' the value is saved as the \code{seed} attribute of the returned
#' value.  The default, \code{NULL} will not change the random
#' generator state, and return \code{.Random.seed} as the \code{seed}
#' attribute, see \code{Value}}
#'
#' }
#'
#' @export
get_simulate_function.lm <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        get_auxiliary_functions(object)$simulate
    }
    else {
        object$auxiliary_functions$simulate
    }
}


#' Function to compute/extract a function that returns the scores
#' (derivatives of the log-likelihood) for an object of class
#' \code{lm}/\code{enriched_lm}
#'
#' @param object an object of class \code{lm} or\code{enriched_lm}
#' @param ... currently not used
#'
#' @details
#' The computed/extracted function has arguments
#' \describe{
#'
#' \item{coefficients}{the regression coefficients at which the scores
#' are computed. If missing then the maximum likelihood estimates are
#' used}
#'
#' \item{dispersion}{the dispersion parameter at which the score
#' function is evaluated. If missing then the maximum likelihood
#' estimate is used}
#'
#' }
#'
#' @export
get_score_function.lm <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        get_auxiliary_functions(object)$score
    }
    else {
        object$auxiliary_functions$score
    }
}

#' Function to compute/extract a function that returns the information
#' matrix for an object of class \code{lm}/\code{enriched_lm}
#'
#' @param object an object of class \code{lm} or\code{enriched_lm}
#' @param ... currently not used
#'
#' @details
#' The computed/extracted function has arguments
#' \describe{
#'
#' \item{coefficients}{the regression coefficients at which the
#' information matrix is evaluated. If missing then the maximum
#' likelihood estimates are used}
#'
#' \item{dispersion}{the dispersion parameter at which the information
#' matrix is evaluated. If missing then the maximum likelihood estimate
#' is used}
#'
#' \item{type}{should the function return the 'expected' or 'observed' information? Default is \code{expected}}
#'
#' \item{QR}{If \code{TRUE}, then the QR decomposition of \deqn{W^{1/2} X} is returned, where \deqn{W} is a diagonal matrix with the working weights (\code{object$weights}) and \deqn{X} is the model matrix.}
#'
#' \item{CHOL}{If \code{TRUE}, then the Cholesky decomposition of the information matrix at the coefficients is returned}
#'
#' }
#'
#' @export
get_information_function.lm <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        get_auxiliary_functions(object)$information
    }
    else {
        object$auxiliary_functions$information
    }
}


#' Function to compute/extract a function that returns the first term
#' in the expansion of the bias of the MLE for the parameters of an
#' object of class \code{lm}/\code{enriched_lm}
#'
#' @param object an object of class \code{lm} or\code{enriched_lm}
#' @param ... currently not used
#'
#' @details
#' The computed/extracted function has arguments
#' \describe{
#'
#' \item{coefficients}{the regression coefficients at which the
#' first-order bias is evacuated. If missing then the maximum
#' likelihood estimates are used}
#'
#' \item{dispersion}{the dispersion parameter at which the first-order
#' bias is evaluated. If missing then the maximum likelihood estimate
#' is used}
#'
#' }
#'
#' @export
get_bias_function.lm <- function(object, ...) {
    if (is.null(object$auxiliary_functions)) {
        get_auxiliary_functions(object)$bias
    }
    else {
        object$auxiliary_functions$bias
    }
}




## ## ## Call that produced the enrichwith template for the current script:
## create_enrichwith_skeleton(class = "lm", option = c("auxiliary functions",
##     "score vector", "mle of dispersion", "expected information",
##     "observed information", "first-order bias"), description = c("various likelihood-based quantities (gradient of the log-likelihood, expected and observed information matrix and first term in the expansion of the bias of the mle) and a simulate method as functions of the model parameters",
##     "gradient of the log-likelihood at the mle", "mle of the dispersion parameter",
##     "expected information matrix evaluated at the mle", "observed information matrix evaluated at the mle",
##     "first term in the expansion of the bias of the mle at the mle"),
##     component = list("auxiliary_functions", "score_mle", "dispersion_mle",
##         "expected_information_mle", "observed_information_mle",
##         "bias_mle"), path = "~/Downloads", attempt_rename = FALSE)

#' Methods to enrich list-like R objects with extra components
#'
#' The enrichwith package provides the \code{\link{enrich}} method to
#' enrich list-like R objects with new, relevant components. The
#' resulting objects preserve their class, so all methods associated
#' with them still apply. The package can also be used to produce
#' customizable source code templates for the structured
#' implementation of methods to compute new components
#'
#' Depending on the object, enriching it can be a tedious task. The
#' \pkg{enrichwith} package aims to streamline the task into 3 simple
#' steps:
#'
#' \enumerate{
#'
#' \item Use \code{\link{create_enrichwith_skeleton}} to produce a
#' customizable enrichwith template.
#'
#' \item Edit the \code{compute_*} functions by adding the specific
#' code that calculates the components.
#'
#' \item Finalize the documentation and/or include more examples.
#' }
#'
#' The first step results in a template that includes all necessary
#' functions to carry out the enrichment. The second step is where the
#' user edits the template and implements the calculation of the
#' components that the object will be enriched with. Specifically,
#' each \code{compute_*} function takes as input the object to be
#' enriched and returns the corresponding new component to be added to
#' the object.
#'
#' Everything else (for example, mapping between the enrichment
#' options and the components that the enriched object will have,
#' checks that an enrichment option exists, listing enrichment
#' options, enriching the object, and so on) is taken care of by the
#' methods in \pkg{enrichwith}.
#'
#' Developers can either put their enrichwith templates in their
#' packages or are welcome to contribute their template to enrichwith,
#' particularly if that extends core R objects.
#'
#' @seealso \code{\link{enrich.glm}} and \code{\link{enriched_glm}}, \code{\link{enrich.family}}, \code{\link{enrich.link-glm}}, \code{\link{enrich.betareg}}
#'
#' @docType package
#' @name enrichwith
#' @import stats
#'
"_PACKAGE"
#> NULL

## register S3 methods
#' Generic method for enriching objects
#'
#' @param object the object to be enriched
#' @param with a character vector with enrichment options for \code{object}
#' @param ... Arguments to be passed to other methods
#'
#' @seealso \code{\link{enrich.glm}}, \code{\link{enriched_glm}}, \code{\link{enrich.family}}, \code{\link{enrich.link-glm}}, \code{\link{enrich.betareg}}
#'
#' @export
enrich <- function(object, with, ...) {
    UseMethod("enrich")
}

#' Generic method for getting available options for the enrichment
#' of objects
#'
#' @aliases print.enrichment_options
#'
#' @param object the object to be enriched
#' @param option a character vector listing the options for enriching
#'     the object
#' @param all_options if \code{TRUE} then output a data frame with the
#'     available enrichment options, their descriptions, the names of
#'     the components that each option results in, and the names of
#'     the corresponding \code{compute} functions.
#' @return if \code{all_options = TRUE} then an object of class
#'     \code{enrichment_options} is returned, otherwise if
#'     \code{option} is specified the output is a character vector
#'     with the names of the functions that compute the enrichment
#'     components
#'
#' @details A check is being made whether the requested option is
#'     available. No check is being made on whether the functions that
#'     produce the components exist.
#'
#' @seealso \code{\link{enrich.glm}}, \code{\link{enrich.family}}, \code{\link{enrich.link-glm}}, \code{\link{enrich.betareg}}
#'
#' @export
get_enrichment_options <- function(object, option, all_options) {
    UseMethod("get_enrichment_options")
}


#' Generic method for extracting or computing auxiliary functions for
#' objects
#' @param object the object to be enriched or the enriched object
#' @param ... currently not used
#' @export
get_auxiliary_functions <- function(object, ...) {
    UseMethod("get_auxiliary_functions")
}

#' Generic method for extracting or computing a function that
#' returns the scores for modelling objects
#' @param object the object to be enriched or the enriched object
#' @param ... currently not used
#' @export
get_score_function <- function(object, ...) {
    UseMethod("get_score_function")
}


#' Generic method for extracting or computing a function that returns
#' the information matrix for modelling objects
#' @param object the object to be enriched or the enriched object
#' @param ... currently not used
#' @export
get_information_function <- function(object, ...) {
    UseMethod("get_information_function")
}


#' Generic method for extracting or computing a function that returns
#' the bias for the parameters in modelling objects
#' @param object the object to be enriched or the enriched object
#' @param ... currently not used
#' @export
get_bias_function <- function(object, ...) {
    UseMethod("get_bias_function")
}

#' Generic method for extracting or computing a simulate function for
#' modelling objects
#' @param object the object to be enriched or the enriched object
#' @param ... currently not used
#' @export
get_simulate_function <- function(object, ...) {
    UseMethod("get_simulate_function")
}

#' Generic method for extracting or computing a dmodel function for
#' modelling objects
#' @param object the object to be enriched or the enriched object
#' @param ... currently not used
#' @export
get_dmodel_function <- function(object, ...) {
    UseMethod("get_dmodel_function")
}

#' Generic method for extracting or computing a pmodel function for
#' modelling objects
#' @param object the object to be enriched or the enriched object
#' @param ... currently not used
#' @export
get_pmodel_function <- function(object, ...) {
    UseMethod("get_pmodel_function")
}

#' Generic method for extracting or computing a qmodel function for
#' modelling objects
#' @param object the object to be enriched or the enriched object
#' @param ... currently not used
#' @export
get_qmodel_function <- function(object, ...) {
    UseMethod("get_qmodel_function")
}

---
title: "Enriching `glm` objects"
author: "[Ioannis Kosmidis](https://www.ikosmidis.com)"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
bibliography: enrichwith.bib
vignette: >
  %\VignetteIndexEntry{enriching glm objects}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

# Introduction

The [**enrichwith**](https://github.com/ikosmidis/enrichwith) R
package provides the `enrich` method to enrich list-like R objects
with new, relevant components. The resulting objects preserve their
class, so all methods associated with them still apply.

This vignette is a demo of the available enrichment options for `glm`
objects.

# Clotting data set

The following data set is provided in @mccullagh:89 [Section 8.4] and
consists of observations on $n = 18$ mean clotting times of blood in
seconds (`time`) for each combination of nine percentage
concentrations of normal plasma (`conc`) and two lots of clotting
agent (`lot`).
```{r, echo = TRUE, eval = TRUE}
clotting <- data.frame(conc = c(5,10,15,20,30,40,60,80,100,
                                5,10,15,20,30,40,60,80,100),
                       time = c(118, 58, 42, 35, 27, 25, 21, 19, 18,
                                69, 35, 26, 21, 18, 16, 13, 12, 12),
                       lot = factor(c(rep(1, 9), rep(2, 9))))
```

@mccullagh:89 [Section 8.4] fitted a series of nested generalized
linear models assuming that the times are realizations of independent
[Gamma](https://en.wikipedia.org/wiki/Gamma_distribution) random
variables whose mean varies appropriately with concentration and
lot. In particular, @mccullagh:89 linked the inverse mean of the gamma
random variables to the linear predictors `~ 1`, `~ log(conc)`, `~
log(conc) + lot`, `~ log(conc) * lot` and carried out an analysis of
deviance to conclude that the `~ log(u) * lot` provides the best
explanation of clotting times. The really close fit of that model to
the data can be seen in the figure below.
```{r, echo = TRUE, eval = TRUE}
library("ggplot2")
clottingML <- glm(time ~ log(conc) * lot, family = Gamma, data = clotting)
alpha <- 0.01
pr_out <- predict(clottingML, type = "response", se.fit = TRUE)
new_data <- clotting
new_data$time <- pr_out$fit
new_data$type <- "fitted"
clotting$type <- "observed"
all_data <- rbind(clotting, new_data)
new_data <- within(new_data, {
    low <- pr_out$fit - qnorm(1 - alpha/2) * pr_out$se.fit
    upp <- pr_out$fit + qnorm(1 - alpha/2) * pr_out$se.fit
}) |> merge(all_data, all = TRUE)
ggplot(all_data) +
    geom_point(aes(conc, time, col = type), alpha = 0.8) +
    geom_segment(data = new_data, aes(x = conc, y = low, xend = conc, yend = upp, col = type)) +
    facet_grid(. ~ lot) +
    theme_bw() +
    theme(legend.position = "top")
```

# Key quantities in likelihood inference
## Score function

The [score function](`r URLencode("https://en.wikipedia.org/wiki/Score_(statistics)")`)
is the gradient of the log-likelihood and is a key object for
likelihood inference.

The **enrichwith** R package provides methods for the enrichment of
`glm` objects with the corresponding score function. This can either
be done by enriching the `glm` object with `auxiliary_functions` and
then extracting the score function
```{r, echo = TRUE, eval = TRUE}
library("enrichwith")
enriched_clottingML <- enrich(clottingML, with = "auxiliary functions")
scores_clottingML <- enriched_clottingML$auxiliary_functions$score
```
or directly using the `get_score_function` convenience method
```{r, echo = TRUE, eval = TRUE}
scores_clottingML <- get_score_function(clottingML)
```

By definition, the score function has to have zero components when evaluated at the
maximum likelihood estimates (only numerically zero here).
```{r, echo = TRUE, eval = TRUE}
scores_clottingML()
```

## Information matrix

Another key quantity in likelihood inference is the [expected
information](https://en.wikipedia.org/wiki/Fisher_information).


The `auxiliary_functions` enrichment option of the `enrich` method
enriches a `glm` object with a function for the evaluation of the
expected information.
```{r, echo = TRUE, eval = TRUE}
info_clottingML <- enriched_clottingML$auxiliary_functions$information
```
and can also be computed directly using the `get_information_function` method
```{r, echo = TRUE, eval = TRUE}
info_clottingML <- get_information_function(clottingML)
```

One of the uses of the expected information is the calculation of
standard errors for the model parameters. The `stats::summary.glm`
function already does that for `glm` objects, estimating the
dispersion parameter (if any) using Pearson residuals.
```{r, echo = TRUE, eval = TRUE}
summary_clottingML <- summary(clottingML)
```

Duly, `info_clottingML` returns (numerically) the same standard errors
as the `summary` method does.
```{r, echo = TRUE, eval = TRUE}
summary_std_errors <- coef(summary_clottingML)[, "Std. Error"]
einfo <- info_clottingML(dispersion = summary_clottingML$dispersion)
all.equal(sqrt(diag(solve(einfo)))[1:4], summary_std_errors, tolerance = 1e-05)
```

Another estimate of the standard errors results by the [observed
information](https://en.wikipedia.org/wiki/Observed_information),
which is the negative [Hessian
matrix](https://en.wikipedia.org/wiki/Hessian_matrix) of the
log-likelihood.

At least at the time of writing the current vignette, there appears
to be no general implementation of the observed information for `glm`
objects. I guess the reason for that is the dependence of the observed
information on higher-order derivatives of the inverse link and
variance functions, which are not readily available in base R.

**enrichwith** provides options for the enrichment of `link-glm` and
`family` objects with such derivatives (see `?enrich.link-glm` and
`?enrich.family` for details), and, based on those allows the
enrichment of `glm` objects with a function to compute the observed
information.

The observed and the expected information for the regression
parameters coincide for GLMs with canonical link, like `clottingML`
```{r, echo = TRUE, eval = TRUE}
oinfo <- info_clottingML(dispersion = summary_clottingML$dispersion, type = "observed")
all.equal(oinfo[1:4, 1:4], einfo[1:4, 1:4])
```
which is not generally true for a `glm` with a non-canonical link, as
seen below.
```{r, echo = TRUE, eval = TRUE}
clottingML_log <- update(clottingML, family = Gamma("log"))
summary_clottingML_log <- summary(clottingML_log)
info_clottingML_log <- get_information_function(clottingML_log)
einfo_log <- info_clottingML_log(dispersion = summary_clottingML_log$dispersion, type = "expected")
oinfo_log <- info_clottingML_log(dispersion = summary_clottingML_log$dispersion, type = "observed")
round(einfo_log, 3)
round(oinfo_log, 3)
```

## Score tests

We can now use `scores_clottingML` and `info_clottingML` to carry out
[a score test](https://en.wikipedia.org/wiki/Score_test) to compare
nested models.

If $l_\psi(\psi, \lambda)$ is the gradient of the log-likelihood with
respect to $\psi$ evaluated at $\psi$ and $\lambda$, $i^{\psi\psi}(\psi,
\lambda)$ is the $(\psi, \psi)$ block of the inverse of the expected
information matrix, and $\hat\lambda_\psi$ is the maximum likelihood
estimator of $\lambda$ for fixed $\psi$, then, assuming that the model
is adequate
\[
l_\psi(\psi, \hat\lambda_\psi)^\top i^{\psi\psi}(\psi,\hat\lambda_\psi) l_\psi(\psi, \hat\lambda_\psi)
\]
has an asymptotic $\chi^2_{dim(\psi)}$ distribution.

The following code chunk computes the maximum likelihood estimates
when the parameters for `lot` and `log(conc):lot` are fixed to zero
($\hat\lambda\psi$ for $\psi = 0$ above), along with the score and
expected information matrix evaluated at them
($l(\psi,\hat\lambda_\psi) and $i(\psi,\hat\lambda_\psi)$, above).
```{r, echo = TRUE, eval = TRUE}
clottingML_nested <- update(clottingML, . ~ log(conc))
enriched_clottingML_nested <- enrich(clottingML_nested, with = "mle of dispersion")
coef_full <- coef(clottingML)
coef_hypothesis <- structure(rep(0, length(coef_full)), names = names(coef_full))
coef_hypothesis_short <- coef(enriched_clottingML_nested, model = "mean")
coef_hypothesis[names(coef_hypothesis_short)] <- coef_hypothesis_short
disp_hypothesis <- coef(enriched_clottingML_nested, model = "dispersion")
scores <- scores_clottingML(coef_hypothesis, disp_hypothesis)
info <- info_clottingML(coef_hypothesis, disp_hypothesis)
```
The object `enriched_clottingML_nested` inherits from `enriched_glm`,
`glm` and `lm`, and, as illustrated above, **enrichwith** provides a
corresponding `coef` method to extract the estimates for the mean
regression parameters and/or the estimate for the dispersion
parameter.

The score statistic is then
```{r, echo = TRUE, eval = TRUE}
(score_statistic <- drop(scores%*%solve(info)%*%scores))
```
which gives a p-value of
```{r, echo = TRUE, eval = TRUE}
pchisq(score_statistic, 2, lower.tail = FALSE)
```

For comparison, the Wald statistic for the same hypothesis is
```{r, echo = TRUE, eval = TRUE}
coef_full[3:4]%*%solve(solve(info)[3:4, 3:4])%*%coef_full[3:4]
```
and the log-likelihood ratio statistic is
```{r, echo = TRUE, eval = TRUE}
(deviance(clottingML_nested) - deviance(clottingML))/disp_hypothesis
```
which is close to the score statistic


## Simulating from `glm` objects at parameter values

**enrichwith** also provides the `get_simulate_function` method for
`glm` or `lm` objects. The `get_simulate_function` computes a function
to simulate response vectors at /arbitrary/ values of the model
parameters, which can be useful when setting up simulation experiments
and for various inferential procedures (e.g. indirect inference).

For example, the following code chunk simulates three response vectors
at the maximum likelihood estimates of the parameters for `clottingML`.
```{r, echo = TRUE, eval = TRUE}
simulate_clottingML <- get_simulate_function(clottingML)
simulate_clottingML(nsim = 3, seed = 123)
```
The result is the same to what the `simulate` method returns
```{r, echo = TRUE, eval = TRUE}
simulate(clottingML, nsim = 3, seed = 123)
```
but `simulate_clottingML` can also be used to simulate at any given
parameter value
```{r, echo = TRUE, eval = TRUE}
coefficients <- c(0, 0.01, 0, 0.01)
dispersion <- 0.001
samples <- simulate_clottingML(coefficients = coefficients, dispersion = dispersion, nsim = 500000, seed = 123)
```
The empirical means and variances based on `samples` agree with the   exact means and variances at `coefficients` and `dispersion`
```{r, echo = TRUE, eval = TRUE}
means <- 1/(model.matrix(clottingML) %*% coefficients)
variances <- dispersion * means^2
max(abs(rowMeans(samples) - means))
max(abs(apply(samples, 1, var) - variances))
```

# pmodel, dmodel, qmodel
**enrichwith** also provides the `get_dmodel_function`, `get_pmodel_function` and `get_qmodel_function` methods, which can be used to evaluate densities or probability mass functions, distribution functions, and quantile functions, respectively at arbitrary parameter values and data settings.

For example, the following code chunk computes the density at the observations in the `clotting` data set under then maximum likelihood fit
```{r, echo = TRUE, eval = TRUE}
cML_dmodel <- get_dmodel_function(clottingML)
cML_dmodel()
```
We can also compute the densities at user-supplied parameter values
```{r, echo = TRUE, eval = TRUE}
cML_dmodel(coefficients = c(-0.01, 0.02, -0.01, 0), dispersion = 0.1)
```
or even at different data points
```{r, echo = TRUE, eval = TRUE}
new_data <- data.frame(conc = 5:10, time = 50:45, lot = factor(c(1, 1, 1, 2, 2, 2)))
cML_dmodel(data = new_data, coefficients = c(-0.01, 0.02, -0.01, 0), dispersion = 0.1)
```
We can also compute cumulative probabilities and quantile functions. So
```{r, echo = TRUE, eval = TRUE}
cML_qmodel <- get_qmodel_function(clottingML)
cML_pmodel <- get_pmodel_function(clottingML)
probs <- cML_pmodel(data = new_data, coefficients = c(-0.01, 0.02, -0.01, 0), dispersion = 0.1)
cML_qmodel(probs, data = new_data, coefficients = c(-0.01, 0.02, -0.01, 0), dispersion = 0.1)
```
returns `new_data$time`.

For example the fitted densities when `(conc, lot)` is `c(15, 1)`, `c(15, 2)`, `c(40, 1)` or `c(40, 2)` are
```{r, echo = TRUE, eval = TRUE}
new_data <- expand.grid(conc = c(15, 40),
                        lot = factor(1:2),
                        time = seq(0, 50, length = 500))
new_data$density <- cML_dmodel(new_data)
ggplot(data = new_data) + 
     geom_line(aes(time, density)) + facet_grid(conc ~ lot) +
     theme_bw() +
     geom_vline(data = clotting[c(3, 6, 12, 15), ], aes(xintercept = time), lty = 2)
```

The dashed vertical lines are the observed values of time for the `(conc, lot)` combinations in the figure.

# `enriched_glm`
**enrichwith** provides a wrapper to the `glm` interface that results directly in `enriched_glm` objects that carry all the components described above. For example,
```{r, echo = TRUE, eval = TRUE}
enriched_clottingML <- enriched_glm(time ~ log(conc) * lot, family = Gamma, data = clotting)
names(enriched_clottingML$auxiliary_functions)
enriched_clottingML$score_mle
enriched_clottingML$expected_information_mle
enriched_clottingML$observed_information_mle
enriched_clottingML$bias_mle
enriched_clottingML$dispersion_mle
```

# Links to other resources

The vignette
[bias](https://CRAN.R-project.org/package=enrichwith/vignettes/bias.html)
of **enrichwith** describes how to use the `auxiliary_functions` to
reduce the bias of the maximum likelihood estimator for generalized
linear models.

# References

---
title: "Bias reduction in generalized linear models using **enrichwith**"
author: "[Ioannis Kosmidis](https://www.ikosmidis.com)"
date: 2016-09-02
output: rmarkdown::html_vignette
bibliography: enrichwith.bib
vignette: >
  %\VignetteIndexEntry{Bias reduction in generalized linear models using enrichwith}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

# Introduction

The [**enrichwith**](https://github.com/ikosmidis/enrichwith) package
provides the `enrich` method to enrich list-like R objects with new,
relevant components. The resulting objects preserve their class, so
all methods associated with them still apply.

This vignette is a short case study demonstrating how enriched `glm`
objects can be used to implement a *quasi [Fisher
scoring](https://en.wikipedia.org/wiki/Scoring_algorithm)* procedure
for computing reduced-bias estimates in [generalized linear
models](https://en.wikipedia.org/wiki/Generalized_linear_model)
(GLMs). @kosmidis:10 describe a parallel quasi
[Newton-Raphson](https://en.wikipedia.org/wiki/Newton%27s_method)
procedure.

------

# Endometrial cancer data

@heinze:02 used a logistic regression model to analyse data from a
study on endometrial cancer. @agresti:15 [Section 5.7] provide details on the data set. Below, we
fit a probit regression model with the same linear predictor as the
logistic regression model in @heinze:02.
```{r, echo = TRUE, eval = TRUE}
# Get the data from the online supplmementary material of Agresti (2015)
data("endometrial", package = "enrichwith")
modML <- glm(HG ~ NV + PI + EH, family = binomial("probit"), data = endometrial)
theta_mle <- coef(modML)
summary(modML)
```
As is the case for the logistic regression in @heinze:02, the maximum
likelihood (ML) estimate of the parameter for `NV` is actually
infinite. The reported, apparently finite value is merely due to false
convergence of the iterative estimation procedure. The same is true
for the estimated standard error, and, hence the value `r
round(coef(summary(modML))["NV", "z value"], 3)` for the $z$-statistic
cannot be trusted for inference on the size of the effect for `NV`.

In categorical-response models like the above, the bias reduction
method in @firth:93 has been found to result in finite estimates even
when the ML ones are infinite [see, @heinze:02, for logistic
regressions; @kosmidis:11, for multinomial regressions; @kosmidis:14a,
for cumulative link models].

One of the variants of that bias reduction method is implemented in
the
[**brglm**](https://CRAN.R-project.org/package=brglm) R
package, which estimates binomial-response GLMs using iterative ML fits
on binomial pseudo-data [see, @kosmidis:07, Chapter 5, for
details]. The reduced-bias estimates for the probit regression on the
endometrial data can be computed as follows.
``` {r, echo = TRUE, eval = TRUE, messages = FALSE}
library("brglm")
modBR <- brglm(HG ~ NV + PI + EH, family = binomial("probit"), data = endometrial)
theta_brglm <- coef(modBR)
summary(modBR)
```
The $z$-statistic for `NV` has value `r round(coef(summary(modBR))["NV", "z value"], 3)` when based on the reduced-bias estimates, providing some evidence for the existence of an effect.

In the following, we use **enrichwith** to implement two variants of the bias reduction method via a unifying quasi Fisher scoring estimation procedure.

------


# Quasi Fisher scoring for bias reduction

Consider a parametric statistical model $\mathcal{P}_\theta$ with
unknown parameter $\theta \in \Re^p$ and the iteration $$\theta^{(k +
1)} := \theta^{(k)} + \left\{i(\theta^{(k)})\right\}^{-1}
s(\theta^{(k)}) - c(\theta^{(k)}) b(\theta^{(k)})$$ where
$\theta^{(k)}$ is the value of $\theta$ at the $k$th iteration,
$s(\theta)$ is the gradient of the
log-[likelihood](https://en.wikipedia.org/wiki/Likelihood_function)
for $\mathcal{P}_\theta$, $i(\theta)$ is the [expected
information](https://en.wikipedia.org/wiki/Fisher_information) matrix,
and $b(\theta)$ is the $O(n^{-1})$ term in the expansion of the
[bias](https://en.wikipedia.org/wiki/Bias_of_an_estimator) of the [ML
estimator](https://en.wikipedia.org/wiki/Maximum_likelihood_estimation)
of $\theta$ [see, for example, @cox:68].

The above iteration defines a *quasi* Fisher scoring estimation
procedure in general, and reduces to exact Fisher scoring for ML
estimation when $c(\theta^{(k)}) = 0_p$, where $0_p$ is a vector of
$p$ zeros.

For either $c(\theta) = I_p$ or $c(\theta) =
\left\{i(\theta)\right\}^{-1} j(\theta)$, where $I_p$ is the $p \times
p$ identity matrix and $j(\theta)$ is the [observed
information](https://en.wikipedia.org/wiki/Observed_information),
$\theta^{(\infty)}$ (if it exists) is a reduced-bias estimate, in the
sense that it corresponds an estimator with bias of smaller asymptotic
order than that of the ML estimator [see, @firth:93;
@kosmidis:10]. The **brglm** estimates correspond to $c(\theta) = I_p$.

The asymptotic distribution of the reduced-bias estimators is the same
to that of the ML estimator [see, @firth:93 for
details]. So, the reduced-bias estimates can be readily used to
calculate $z$-statistics.

------

# Implementation using **enrichwith**

For implementing the iteration for bias reduction, we need functions that can compute the gradient of the log-likelihood, the observed and expected information matrix, and $b(\theta)$ at arbitrary values of $\theta$.

The **enrichwith** package can produce those functions for any `glm` object through the `auxiliary_functions` enrichment option (to see all available enrichment options for `glm` objects run `get_enrichment_options(modML)`.
``` {r, echo = TRUE, eval = TRUE}
library("enrichwith")
enriched_modML <- enrich(modML, with = "auxiliary functions")
```
Let's extract the functions from the `enriched_modML` object.
``` {r, echo = TRUE, eval = TRUE}
# Extract the ingredients for the quasi Fisher scoring iteration from the enriched glm object
gradient <- enriched_modML$auxiliary_functions$score # gradient of the log-likelihood
information <- enriched_modML$auxiliary_functions$information # information matrix
bias <- enriched_modML$auxiliary_functions$bias # first-order bias
```
For the more technically minded, note here that the above functions are specific to `modML` in the sense that they look into a special environment for necessary objects like the model matrix, the model weights, the response vector, and so on.

This stems from the way **enrichwith** has been implemented. In particular, if `create_enrichwith_skeleton` is used, the user/developer can directly implement enrichment options to enrich objects with functions that directly depend on other components in the object to be enriched.

The following code chunk uses `enriched_modML` to implement the quasi Fisher scoring procedure for the analysis of the endometrial cancer data. For `p <- length(theta_mle)` `r p <- length(theta_mle)`, the starting value for the parameter vector is set to `theta_current <- rep(0, p)` `r theta_current <- rep(0, p)`, and the maximum number of iterations to `maxit <- 100` `r maxit <- 100`. As stopping criterion, we use the absolute change in each parameter value with tolerance `epsilon <- 1e-06` `r epsilon <- 1e-06`
``` {r, echo = TRUE, eval = TRUE}
# The quasi Fisher scoring iteration using c(theta) = identity
for (k in seq.int(maxit)) {
    s_vector <- gradient(theta_current)
    i_matrix <- information(theta_current, type = "expected")
    b_vector <- bias(theta_current)
    step <- solve(i_matrix) %*% s_vector - b_vector
    theta_current <- theta_current + step
    # A stopping criterion
    if (all(abs(step) < epsilon)) {
        break
    }
}
(theta_e <- drop(theta_current))
```
The estimation procedure took `r k` iterations to converge, and, as expected, the estimates are numerically the same to the ones that **brglm** returned.
``` {r, echo = TRUE, eval = TRUE}
all.equal(theta_e, theta_brglm, check.attributes = FALSE, tolerance = epsilon)
```

A set of alternative reduced-bias estimates can be obtained using $c(\theta) = \left\{i(\theta)\right\}^{-1} j(\theta)$. Starting again at `theta_current <- rep(0, p)` `r theta_current <- rep(0, p)`
``` {r, echo = TRUE, eval = TRUE}
# The quasi Fisher scoring iteration using c(theta) = solve(i(theta)) %*% j(theta)
for (k in seq.int(maxit)) {
    s_vector <- gradient(theta_current)
    i_matrix <- information(theta_current, type = "expected")
    j_matrix <- information(theta_current, type = "observed")
    b_vector <- bias(theta_current)
    step <- solve(i_matrix) %*% (s_vector - j_matrix %*% b_vector)
    theta_current <- theta_current + step
    # A stopping criterion
    if (all(abs(step) < epsilon)) {
        break
    }
}
(theta_o <- drop(theta_current))
```
The estimation procedure took `r k` iterations to converge.

The ML estimates and the estimates from the two variants of the bias reduction method are
``` {r, echo = TRUE, eval = TRUE}
round(data.frame(theta_mle, theta_e, theta_o), 3)
```
Note that the reduced-bias estimates have shrunk towards zero. This is typical for reduced-bias estimation in binomial-response GLMs [see, for example, @cordeiro:91, Section 8; @kosmidis:07, Section 5.2; @kosmidis:14a for shrinkage in cumulative link models].

The corresponding $z$-statistics are
``` {r, echo = TRUE, eval = TRUE}
se_theta_mle <- sqrt(diag(solve(information(theta_mle, type = "expected"))))
se_theta_e <- sqrt(diag(solve(information(theta_e, type = "expected"))))
se_theta_o <- sqrt(diag(solve(information(theta_o, type = "expected"))))
round(data.frame(z_mle = theta_mle/se_theta_mle,
                 z_br_e = theta_e/se_theta_e,
                 z_br_o = theta_o/se_theta_o), 3)
```
The two variants for bias reduction result in slightly different reduced-bias estimates and $z$-statistics, though the $z$-statistics from both variants provide some evidence for the existence of an effect for `NV`.

------

# Notes
A general family of bias reduction methods is described in @kosmidis:09.

The quasi Fisher scoring iteration that has been described here is at the core of the [**brglm2**](https://github.com/ikosmidis/brglm2) R package, which provides various bias reduction methods for GLMs.

------

# References

---
title: "Enriching `family` objects: exponential family of distributions"
author: "[Ioannis Kosmidis](https://www.ikosmidis.com)"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
bibliography: enrichwith.bib
vignette: >
  %\VignetteIndexEntry{enriching family objects}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

# Introduction

The [**enrichwith**](https://github.com/ikosmidis/enrichwith) R
package provides the `enrich` method to enrich list-like R objects
with new, relevant components. The resulting objects preserve their
class, so all methods associated with them still apply.

This vignette is a demo of the available enrichment options for
`family` objects, focusing on objects that correspond to members of
the exponential family of distributions.

# Exponential family and `family` objects

`family` objects specify characteristics of the models used by
functions such as `glm`. The families implemented in the `stats`
package include `binomial`, `gaussian`, `Gamma`, `inverse.gaussian`,
and `poisson`, which obvious corresponding distributions. These
distributions are all special cases of the exponential family of
distributions with probability mass or density function of the
form
$$
f_{Y}(y) = \exp\left\{\frac{y \theta - b(\theta) - c_1(y)}{\phi/m} - \frac{1}{2}a\left(-\frac{m}{\phi}\right) + c_2(y) \right\}
$$
for some sufficiently smooth functions $b(.)$, $c_1(.)$, $a(.)$ and $c_2(.)$, and a fixed weight $m$. The expected value and the variance of $Y$ are then
\begin{align*}
      E(Y) & = \mu =  b'(\theta) \\
      Var(Y) & = \frac{\phi}{m}b''(\theta) = \frac{\phi}{m}V(\mu)
\end{align*}
where $V(\mu)$ and $\phi$ are the variance function and the dispersion parameter, respectively. Below we list the characteristics of the distributions supported by `family` objects.

#### [Normal](https://en.wikipedia.org/wiki/Normal_distribution) with mean $\mu$ and variance $\phi/m$
$\theta = \mu$, $\displaystyle b(\theta) = \frac{\theta^2}{2}$, $\displaystyle  c_1(y) = \frac{y^2}{2}$,
$\displaystyle a(\zeta) = -\log(-\zeta)$, $\displaystyle c_2(y) = -\frac{1}{2}\log(2\pi)$

#### [Binomial](https://en.wikipedia.org/wiki/Binomial_distribution) with index $m$ and probability $\mu$
$\displaystyle \theta = \log\frac{\mu}{1- \mu}$, $\displaystyle b(\theta) = \log(1 + e^\theta)$,
$\displaystyle \phi = 1$, $\displaystyle c_1(y) = 0$, $\displaystyle a(\zeta) = 0$, $\displaystyle c_2(y) = \log{m\choose{my}}$

#### [Poisson](https://en.wikipedia.org/wiki/Poisson_distribution) with mean $\mu$
$\displaystyle \theta = \log\mu$,
$\displaystyle b(\theta) = e^\theta$,
$\displaystyle \phi = 1$,
$\displaystyle c_1(y) = 0$,
$\displaystyle a(\zeta) = 0$,
$\displaystyle c_2(y) = -\log\Gamma(y + 1)$

#### [Gamma](https://en.wikipedia.org/wiki/Gamma_distribution) with mean $\mu$ and shape $1/\phi$
$\displaystyle \theta = -\frac{1}{\mu}$,
$\displaystyle b(\theta) = -\log(-\theta)$,
$\displaystyle c_1(y) = -\log y$,
$\displaystyle a(\zeta) = 2 \log \Gamma(-\zeta) + 2 \zeta \log\left(-\zeta\right)$,
$\displaystyle c_2(y) = -\log y$

#### [Inverse Gaussian](https://en.wikipedia.org/wiki/Inverse_Gaussian_distribution) with mean $\mu$ and variance $\phi\mu^3$
$\displaystyle \theta = -\frac{1}{2\mu^2}$,
$\displaystyle b(\theta) = -\sqrt{-2\theta}$,
$\displaystyle c_1(y) = \frac{1}{2y}$,
$\displaystyle a(\zeta) = -\log(-\zeta)$,
$\displaystyle c_2(y) = -\frac{1}{2}\log\left(\pi y^3\right)$


# Components in `family` objects
`family` objects provide functions for the variance function (`variance`), a specification of deviance residuals (`dev.resids`) and the Akaike information criterion (`aic`). For example
```{r, echo = TRUE, eval = TRUE}
inverse.gaussian()$dev.resids
inverse.gaussian()$variance
inverse.gaussian()$aic
```

# Enrichment options for `family` objects
The **enrichwith** R package provides methods for the enrichment of
`family` objects with a function that links the natural parameter
$\theta$ with $\mu$, the function $b(\theta)$, the first two
derivatives of $V(\mu)$, $a(\zeta)$ and its first four derivatives,
and $c_1(y)$ and $c_2(y)$. To illustrate, let's write a function that
reconstructs the densities and probability mass functions from the
components that result from enrichment
```{r, echo = TRUE, eval = TRUE}
library("enrichwith")
dens <- function(y, m = 1, mu, phi, family) {
    object <- enrich(family)
    with(object, {
        c2 <- if (family == "binomial") c2fun(y, m) else c2fun(y)
        exp(m * (y * theta(mu) - bfun(theta(mu)) - c1fun(y))/phi -
            0.5 * afun(-m/phi) + c2)
    })
}
```
The following chunks test `dens` for a few distributions against the
standard `d*` functions
```{r, echo = TRUE, eval = TRUE}
## Normal
all.equal(dens(y = 0.2, m = 3, mu = 1, phi = 3.22, gaussian()),
          dnorm(x = 0.2, mean = 1, sd = sqrt(3.22/3)))

## Gamma
all.equal(dens(y = 3, m = 1.44, mu = 2.3, phi = 1.3, Gamma()),
          dgamma(x = 3, shape = 1.44/1.3, 1.44/(1.3 * 2.3)))

## Inverse gaussian
all.equal(dens(y = 0.2, m = 7.23, mu = 1, phi = 3.22, inverse.gaussian()),
          SuppDists::dinvGauss(0.2, nu = 1, lambda = 7.23/3.22))

## Binomial
all.equal(dens(y = 0.34, m = 100, mu = 0.32, phi = 1, binomial()),
          dbinom(x = 34, size = 100, prob = 0.32))

```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich-package.R
\name{enrich}
\alias{enrich}
\title{Generic method for enriching objects}
\usage{
enrich(object, with, ...)
}
\arguments{
\item{object}{the object to be enriched}

\item{with}{a character vector with enrichment options for \code{object}}

\item{...}{Arguments to be passed to other methods}
}
\description{
Generic method for enriching objects
}
\seealso{
\code{\link{enrich.glm}}, \code{\link{enriched_glm}}, \code{\link{enrich.family}}, \code{\link{enrich.link-glm}}, \code{\link{enrich.betareg}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{endometrial}
\alias{endometrial}
\title{Histology grade and risk factors for 79 cases of endometrial cancer}
\format{
A data frame with 79 rows and 4 variables:
\describe{

\item{NV}{neovasculization with coding 0 for absent and 1 for present}

\item{PI}{pulsality index of arteria uterina}

\item{EH}{endometrium height}

\item{HG}{histology grade with coding 0 for low grade and 1 for high grade}

}
}
\source{
The packaged data set was downloaded in \code{.dat} format
    from \url{https://users.stat.ufl.edu/~aa/glm/data/}. The latter
    link provides the data sets used in Agresti (2015).

    The endometrial data set was first analyzed in Heinze and
    Schemper (2002), and was originally provided by Dr
    E. Asseryanis from the Medical University of Vienna.


Agresti, A. 2015. *Foundations of Linear and Generalized Linear
Models*.  Wiley Series in Probability and Statistics. Wiley.

Heinze G, Schemper M (2002). A Solution to the Problem of
Separation in Logistic Regression. *Statistics in Medicine*,
**21**, 2409–2419. \doi{10.1002/sim.1047}.
}
\usage{
endometrial
}
\description{
Histology grade and risk factors for 79 cases of endometrial cancer
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.lm.R
\name{enrich.lm}
\alias{enrich.lm}
\title{Enrich objects of class \code{\link{lm}}}
\usage{
\method{enrich}{lm}(object, with = "all", ...)
}
\arguments{
\item{object}{an object of class lm}

\item{with}{a character vector of options for the enrichment of \code{object}}

\item{...}{extra arguments to be passed to the
\code{compute_*} functions}
}
\value{
The object \code{object} of class \code{\link{lm}} with extra
components. \code{get_enrichment_options.lm()} returns the
components and their descriptions.
}
\description{
Enrich objects of class \code{\link{lm}} with any or all of a set
auxiliary functions, the maximum likelihood estimate of the
dispersion parameter, the expected or observed information at the
maximum likelihood estimator, and the first term in the expansion
of the bias of the maximum likelihood estimator.
}
\details{
The auxiliary functions consist of the score functions, the
expected or observed information, the first-order bias of the
maximum likelihood estimator as functions of the model parameters,
and a \code{simulate} function that takes as input the model
parameters (including the dispersion if any). The result from the
\code{simulate} auxiliary function has the same structure to that
of the \code{\link{simulate}} method for \code{\link{lm}} objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.lm.R
\name{get_information_function.lm}
\alias{get_information_function.lm}
\title{Function to compute/extract a function that returns the information
matrix for an object of class \code{lm}/\code{enriched_lm}}
\usage{
\method{get_information_function}{lm}(object, ...)
}
\arguments{
\item{object}{an object of class \code{lm} or\code{enriched_lm}}

\item{...}{currently not used}
}
\description{
Function to compute/extract a function that returns the information
matrix for an object of class \code{lm}/\code{enriched_lm}
}
\details{
The computed/extracted function has arguments
\describe{

\item{coefficients}{the regression coefficients at which the
information matrix is evaluated. If missing then the maximum
likelihood estimates are used}

\item{dispersion}{the dispersion parameter at which the information
matrix is evaluated. If missing then the maximum likelihood estimate
is used}

\item{type}{should the function return the 'expected' or 'observed' information? Default is \code{expected}}

\item{QR}{If \code{TRUE}, then the QR decomposition of \deqn{W^{1/2} X} is returned, where \deqn{W} is a diagonal matrix with the working weights (\code{object$weights}) and \deqn{X} is the model matrix.}

\item{CHOL}{If \code{TRUE}, then the Cholesky decomposition of the information matrix at the coefficients is returned}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich-package.R
\name{get_dmodel_function}
\alias{get_dmodel_function}
\title{Generic method for extracting or computing a dmodel function for
modelling objects}
\usage{
get_dmodel_function(object, ...)
}
\arguments{
\item{object}{the object to be enriched or the enriched object}

\item{...}{currently not used}
}
\description{
Generic method for extracting or computing a dmodel function for
modelling objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.glm.R
\name{get_enrichment_options.glm}
\alias{get_enrichment_options.glm}
\title{Available options for the enrichment objects of class
\code{\link{glm}}}
\usage{
\method{get_enrichment_options}{glm}(object, option, all_options = missing(option))
}
\arguments{
\item{object}{the object to be enriched}

\item{option}{a character vector listing the options for enriching
the object}

\item{all_options}{if \code{TRUE} then output a data frame with the
available enrichment options, their descriptions, the names of
the components that each option results in, and the names of
the corresponding \code{compute_*} functions.}
}
\value{
an object of class \code{enrichment_options}
}
\description{
Available options for the enrichment objects of class
\code{\link{glm}}
}
\details{
A check is being made whether the requested option is
    available. No check is being made on whether the functions that
    produce the components exist.
}
\examples{
\dontrun{
get_enrichment_options.glm(option = "all")
get_enrichment_options.glm(all_options = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.betareg.R
\name{get_information_function.betareg}
\alias{get_information_function.betareg}
\title{Function to compute/extract a function that returns the information
matrix for an object of class \code{betareg}/\code{enriched_betareg}}
\usage{
\method{get_information_function}{betareg}(object, ...)
}
\arguments{
\item{object}{an object of class \code{betareg} or\code{enriched_betareg}}

\item{...}{currently not used}
}
\description{
Function to compute/extract a function that returns the information
matrix for an object of class \code{betareg}/\code{enriched_betareg}
}
\details{
The computed/extracted function has arguments
\describe{

\item{coefficients}{the regression coefficients at which the
information matrix is evaluated. If missing then the maximum
likelihood estimates are used}


\item{type}{should the function return the 'expected' or 'observed' information? Default is \code{expected}}

\item{QR}{Currently not used}

\item{CHOL}{If \code{TRUE}, then the Cholesky decomposition of the information matrix at the coefficients is returned}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.lm.R
\name{get_simulate_function.lm}
\alias{get_simulate_function.lm}
\title{Function to compute/extract a simulate function for response
vectors from an object of class \code{lm}/\code{enriched_lm}}
\usage{
\method{get_simulate_function}{lm}(object, ...)
}
\arguments{
\item{object}{an object of class \code{lm} or\code{enriched_lm}}

\item{...}{currently not used}
}
\description{
Function to compute/extract a simulate function for response
vectors from an object of class \code{lm}/\code{enriched_lm}
}
\details{
The computed/extracted simulate function has arguments
\describe{

\item{coefficients}{the regression coefficients at which the
response vectors are simulated. If missing then the maximum
likelihood estimates are used}

\item{dispersion}{the dispersion parameter at which the response
vectors are simulated. If missing then the maximum likelihood
estimate is used}

\item{nsim}{number of response vectors to simulate.  Defaults to \code{1}}

\item{seed}{an object specifying if and how the random number
generator should be initialized ('seeded'). It can be either
\code{NULL} or an integer that will be used in a call to
\code{set.seed} before simulating the response vectors.  If set,
the value is saved as the \code{seed} attribute of the returned
value.  The default, \code{NULL} will not change the random
generator state, and return \code{.Random.seed} as the \code{seed}
attribute, see \code{Value}}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.betareg.R
\name{get_score_function.betareg}
\alias{get_score_function.betareg}
\title{Function to compute/extract a function that returns the scores
(derivatives of the log-likelihood) for an object of class
\code{betareg}/\code{enriched_betareg}}
\usage{
\method{get_score_function}{betareg}(object, ...)
}
\arguments{
\item{object}{an object of class \code{betareg} or\code{enriched_betareg}}

\item{...}{currently not used}
}
\description{
Function to compute/extract a function that returns the scores
(derivatives of the log-likelihood) for an object of class
\code{betareg}/\code{enriched_betareg}
}
\details{
The computed/extracted function has arguments
\describe{

\item{coefficients}{the regression coefficients at which the scores
are computed. If missing then the maximum likelihood estimates are
used}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.family.R
\name{get_enrichment_options.family}
\alias{get_enrichment_options.family}
\title{Available options for the enrichment objects of class family}
\usage{
\method{get_enrichment_options}{family}(object, option, all_options = missing(option))
}
\arguments{
\item{object}{the object to be enriched}

\item{option}{a character vector listing the options for enriching
the object}

\item{all_options}{if \code{TRUE} then output a data frame with the
available enrichment options, their descriptions, the names of
the components that each option results in, and the names of
the corresponding \code{compute_*} functions.}
}
\value{
an object of class \code{enrichment_options}
}
\description{
Available options for the enrichment objects of class family
}
\details{
A check is being made whether the requested option is
    available. No check is being made on whether the functions that
    produce the components exist.
}
\examples{
\dontrun{
get_enrichment_options.family(option = "all")
get_enrichment_options.family(all_options = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.glm.R
\name{coef.enriched_glm}
\alias{coef.enriched_glm}
\title{Function to extract model coefficients from objects of class \code{enriched_glm}}
\usage{
\method{coef}{enriched_glm}(object, model = c("mean", "full", "dispersion"), ...)
}
\arguments{
\item{object}{an object of class \code{enriched_glm}}

\item{model}{either "mean" for the estimates of the parameters in the linear predictor, or "dispersion" for the estimate of the dispersion, or "full" for all estimates}

\item{...}{currently unused}
}
\description{
Function to extract model coefficients from objects of class \code{enriched_glm}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich-package.R
\name{get_bias_function}
\alias{get_bias_function}
\title{Generic method for extracting or computing a function that returns
the bias for the parameters in modelling objects}
\usage{
get_bias_function(object, ...)
}
\arguments{
\item{object}{the object to be enriched or the enriched object}

\item{...}{currently not used}
}
\description{
Generic method for extracting or computing a function that returns
the bias for the parameters in modelling objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich-package.R
\name{get_enrichment_options}
\alias{get_enrichment_options}
\alias{print.enrichment_options}
\title{Generic method for getting available options for the enrichment
of objects}
\usage{
get_enrichment_options(object, option, all_options)
}
\arguments{
\item{object}{the object to be enriched}

\item{option}{a character vector listing the options for enriching
the object}

\item{all_options}{if \code{TRUE} then output a data frame with the
available enrichment options, their descriptions, the names of
the components that each option results in, and the names of
the corresponding \code{compute} functions.}
}
\value{
if \code{all_options = TRUE} then an object of class
    \code{enrichment_options} is returned, otherwise if
    \code{option} is specified the output is a character vector
    with the names of the functions that compute the enrichment
    components
}
\description{
Generic method for getting available options for the enrichment
of objects
}
\details{
A check is being made whether the requested option is
    available. No check is being made on whether the functions that
    produce the components exist.
}
\seealso{
\code{\link{enrich.glm}}, \code{\link{enrich.family}}, \code{\link{enrich.link-glm}}, \code{\link{enrich.betareg}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.family.R
\name{enrich.family}
\alias{enrich.family}
\title{Enrich objects of class \code{\link{family}}}
\usage{
\method{enrich}{family}(object, with = "all", ...)
}
\arguments{
\item{object}{an object of class \code{\link{family}}}

\item{with}{a character vector with enrichment options for \code{object}}

\item{...}{extra arguments to be passed to the \code{compute_*}
functions}
}
\value{
The object \code{object} of class \code{\link{family}} with
    extra components. \code{get_enrichment_options.family()}
    returns the components and their descriptions.
}
\description{
Enrich objects of class \code{\link{family}} with family-specific
mathematical functions
}
\details{
\code{\link[stats]{family}} objects specify characteristics of the
models used by functions such as \code{\link[stats]{glm}}. The
families implemented in the \code{stats} package include
\code{\link[stats]{binomial}}, \code{\link[stats]{gaussian}},
\code{\link[stats]{Gamma}}, \code{\link[stats]{inverse.gaussian}},
and \code{\link[stats]{poisson}}, which are all special cases of
the exponential family of distributions that have probability mass
or density function of the form \deqn{f(y; \theta, \phi) =
\exp\left\{\frac{y\theta - b(\theta) - c_1(y)}{\phi/m} -
\frac{1}{2}a\left(-\frac{m}{\phi}\right) + c_2(y)\right\} \quad y
\in Y \subset \Re\,, \theta \in \Theta \subset \Re\, , \phi >
0}{f(y, theta, phi) = exp((y * theta - b(theta) - c_1(y))/(phi/m) -
a(-m/phi)/2 + c_2(y))} where \eqn{m > 0}{m > 0} is an observation
weight, and \eqn{a(.)}{a(.)}, \eqn{b(.)}{b(.)},
\eqn{c_1(.)}{c_1(.)} and \eqn{c_2(.)}{c_2(.)} are sufficiently
smooth, real-valued functions.

The current implementation of \code{\link[stats]{family}} objects
includes the variance function (\code{variance}), the deviance
residuals (\code{dev.resids}), and the Akaike information criterion
(\code{aic}). See, also \code{\link{family}}.

The \code{enrich} method can further enrich exponential
\code{\link{family}} distributions with \eqn{\theta}{theta} in
terms of \eqn{\mu}{mu} (\code{theta}), the functions
\eqn{b(\theta)}{b(theta)} (\code{bfun}), \eqn{c_1(y)}{c_1(y)}
(\code{c1fun}), \eqn{c_2(y)}{c_2(y)} (\code{c2fun}),
\eqn{a(\zeta)}{a(zeta)} (\code{fun}), the first two derivatives of
\eqn{V(\mu)}{V(mu)} (\code{d1variance} and \code{d2variance},
respectively), and the first four derivatives of
\eqn{a(\zeta)}{a(zeta)} (\code{d1afun}, \code{d2afun},
\code{d3afun}, \code{d4afun}, respectively).

Corresponding enrichment options are also available for
\code{\link[stats]{quasibinomial}},
\code{\link[stats]{quasipoisson}} and \code{\link[gnm]{wedderburn}}
families.

The \code{\link[stats]{quasi}} families are enriched with
\code{d1variance} and \code{d2variance}.

See \code{\link{enrich.link-glm}} for the enrichment of
\code{\link[=make.link]{link-glm}} objects.
}
\examples{

## An example from ?glm to illustrate that things still work with
## enriched families
counts <- c(18,17,15,20,10,20,25,13,12)
outcome <- gl(3,1,9)
treatment <- gl(3,3)
print(d.AD <- data.frame(treatment, outcome, counts))
glm.D93 <- glm(counts ~ outcome + treatment, family = enrich(poisson()))
anova(glm.D93)
summary(glm.D93)
}
\seealso{
\code{\link{enrich.link-glm}}, \code{\link{make.link}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich-package.R
\name{get_auxiliary_functions}
\alias{get_auxiliary_functions}
\title{Generic method for extracting or computing auxiliary functions for
objects}
\usage{
get_auxiliary_functions(object, ...)
}
\arguments{
\item{object}{the object to be enriched or the enriched object}

\item{...}{currently not used}
}
\description{
Generic method for extracting or computing auxiliary functions for
objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.glm.R
\name{get_pmodel_function.glm}
\alias{get_pmodel_function.glm}
\title{Function to compute/extract a \code{pmodel} function}
\usage{
\method{get_pmodel_function}{glm}(object, ...)
}
\arguments{
\item{object}{an object of class \code{glm} or\code{enriched_glm}}

\item{...}{currently not used}
}
\description{
Function to compute/extract a \code{pmodel} function
}
\details{
The computed/extracted function has arguments
\describe{

\item{data}{a data frame with observations at which to compute the
distribution function. If missing then probabilities are computed
at the model frame extracted from the object (see
\code{\link{glm}})}

\item{coefficients}{the regression coefficients at which the
distribution function are computed. If missing then the maximum
likelihood estimates are used}

\item{dispersion}{the dispersion parameter at which the
distribution function is computed. If missing then the maximum
likelihood estimate is used}

\item{log.p}{logical; if \code{TRUE}, the logarithm of the
distribution function is returned}

\item{lower.tail}{logical; if \code{TRUE} (default), probabilities
are P[X <= x] otherwise, P[X > x]}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.glm.R
\name{get_dmodel_function.glm}
\alias{get_dmodel_function.glm}
\title{Function to compute/extract a \code{dmodel} function}
\usage{
\method{get_dmodel_function}{glm}(object, ...)
}
\arguments{
\item{object}{an object of class \code{glm} or\code{enriched_glm}}

\item{...}{currently not used}
}
\description{
Function to compute/extract a \code{dmodel} function
}
\details{
The computed/extracted function has arguments
\describe{

\item{data}{a data frame with observations at which to compute
densities. If missing then densities are computed at the model
frame extracted from the object (see \code{\link{glm}})}

\item{coefficients}{the regression coefficients at which the
densities are computed. If missing then the maximum likelihood
estimates are used}

\item{dispersion}{the dispersion parameter at which the densities
function is computed. If missing then the maximum likelihood
estimate is used}

\item{log}{logical; if \code{TRUE}, the logarithm of the density is
returned}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.betareg.R
\name{enrich.betareg}
\alias{enrich.betareg}
\title{Enrich objects of class betareg}
\usage{
\method{enrich}{betareg}(object, with = "all", ...)
}
\arguments{
\item{object}{an object of class \code{\link[betareg]{betareg}}}

\item{with}{a character vector of options for the enrichment of \code{object}}

\item{...}{extra arguments to be passed to the
\code{compute_*} functions}
}
\value{
The object \code{object} of class \code{\link[betareg]{betareg}}
    with extra components. \code{get_enrichment_options.betareg()}
    returns the components and their descriptions.
}
\description{
Enrich objects of class \code{\link[betareg]{betareg}} with any or all of a
set of auxiliary functions, the expected information at the maximum
likelihood estimator, and the first term in the expansion of the
bias of the maximum likelihood estimator.
}
\details{
The \code{auxiliary_functions} component consists of any or all of the following functions:
\itemize{
\item \code{score}: the log-likelihood derivatives as a function of the model parameters; see \code{get_score_function.betareg}
\item \code{information}: the expected information as a function of the model parameters; see \code{\link{get_information_function.betareg}}
\item \code{bias}: the first-order term in the expansion of the bias of the maximum likelihood estimator as a function of the model parameters; see \code{\link{get_bias_function.betareg}}
\item \code{simulate}: a \code{\link{simulate}} function for \code{\link[betareg]{betareg}} objects that can simulate variates from the model at user-supplied parameter values for the regression parameters (default is the maximum likelihood estimates); see \code{\link{get_simulate_function.betareg}}
}
}
\examples{

\dontrun{
if (require("betareg")) {

   data("GasolineYield", package = "betareg")
   gy <- betareg(yield ~ batch + temp, data = GasolineYield)

   # Get a function that evaluates the expected information for gy at supplied parameter values
   gy_info <- get_information_function(gy)
.   # compare standard errors with what `summary` returns
   all.equal(sqrt(diag(solve(gy_info())))[1:11],
             coef(summary(gy))$mean[, 2], check.attributes = FALSE)
.   # evaluating at different parameter values
   gy_info(rep(1, length = 12))

   # Get a function that evaluates the first-order bias of gy at supplied parameter values
   gy_bias <- get_bias_function(gy)
   # compare with internal betareg implementation of bias correction
   gy_bc <- update(gy, type = "BC")
   all.equal(gy_bias(coef(gy)), gy_bc$bias, check.attributes = FALSE)

 }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.betareg.R
\name{get_simulate_function.betareg}
\alias{get_simulate_function.betareg}
\title{Function to compute/extract a simulate function for response
vectors from an object of class \code{betareg}/\code{enriched_betareg}}
\usage{
\method{get_simulate_function}{betareg}(object, ...)
}
\arguments{
\item{object}{an object of class \code{betareg} or\code{enriched_betareg}}

\item{...}{currently not used}
}
\description{
Function to compute/extract a simulate function for response
vectors from an object of class \code{betareg}/\code{enriched_betareg}
}
\details{
The computed/extracted simulate function has arguments
\describe{

\item{coefficients}{the regression coefficients at which the
response vectors are simulated. If missing then the maximum
likelihood estimates are used}

\item{nsim}{number of response vectors to simulate.  Defaults to \code{1}}

\item{seed}{an object specifying if and how the random number
generator should be initialized ('seeded'). It can be either
\code{NULL} or an integer that will be used in a call to
\code{set.seed} before simulating the response vectors.  If set,
the value is saved as the \code{seed} attribute of the returned
value.  The default, \code{NULL} will not change the random
generator state, and return \code{.Random.seed} as the \code{seed}
attribute, see \code{Value}}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.glm.R
\name{get_score_function.glm}
\alias{get_score_function.glm}
\title{Function to compute/extract a function that returns the scores
(derivatives of the log-likelihood) for an object of class
\code{glm}/\code{enriched_glm}}
\usage{
\method{get_score_function}{glm}(object, ...)
}
\arguments{
\item{object}{an object of class \code{glm} or\code{enriched_glm}}

\item{...}{currently not used}
}
\description{
Function to compute/extract a function that returns the scores
(derivatives of the log-likelihood) for an object of class
\code{glm}/\code{enriched_glm}
}
\details{
The computed/extracted function has arguments
\describe{

\item{coefficients}{the regression coefficients at which the scores
are computed. If missing then the maximum likelihood estimates are
used}

\item{dispersion}{the dispersion parameter at which the score
function is evaluated. If missing then the maximum likelihood
estimate is used}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.glm.R
\name{get_simulate_function.glm}
\alias{get_simulate_function.glm}
\title{Function to compute/extract a simulate function for response
vectors from an object of class \code{glm}/\code{enriched_glm}}
\usage{
\method{get_simulate_function}{glm}(object, ...)
}
\arguments{
\item{object}{an object of class \code{glm} or\code{enriched_glm}}

\item{...}{currently not used}
}
\description{
Function to compute/extract a simulate function for response
vectors from an object of class \code{glm}/\code{enriched_glm}
}
\details{
The computed/extracted simulate function has arguments
\describe{

\item{coefficients}{the regression coefficients at which the
response vectors are simulated. If missing then the maximum
likelihood estimates are used}

\item{dispersion}{the dispersion parameter at which the response
vectors are simulated. If missing then the maximum likelihood
estimate is used}

\item{nsim}{number of response vectors to simulate.  Defaults to \code{1}}

\item{seed}{an object specifying if and how the random number
generator should be initialized ('seeded'). It can be either
\code{NULL} or an integer that will be used in a call to
\code{set.seed} before simulating the response vectors.  If set,
the value is saved as the \code{seed} attribute of the returned
value.  The default, \code{NULL} will not change the random
generator state, and return \code{.Random.seed} as the \code{seed}
attribute, see \code{Value}}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.betareg.R
\name{get_auxiliary_functions.betareg}
\alias{get_auxiliary_functions.betareg}
\title{Function to compute/extract auxiliary functions from objects of
class \code{betreg}/\code{enriched_betareg}}
\usage{
\method{get_auxiliary_functions}{betareg}(object, ...)
}
\arguments{
\item{object}{an object of class \code{betareg} or\code{enriched_betareg}}

\item{...}{currently not used}
}
\description{
Function to compute/extract auxiliary functions from objects of
class \code{betreg}/\code{enriched_betareg}
}
\details{
See \code{\link{enrich.betareg}} for details.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.glm.R
\name{enrich.glm}
\alias{enrich.glm}
\title{Enrich objects of class \code{\link{glm}}}
\usage{
\method{enrich}{glm}(object, with = "all", ...)
}
\arguments{
\item{object}{an object of class glm}

\item{with}{a character vector of options for the enrichment of \code{object}}

\item{...}{extra arguments to be passed to the
\code{compute_*} functions}
}
\value{
The object \code{object} of class \code{\link{glm}} with extra
components. See \code{get_enrichment_options.glm()} for the
components and their descriptions.
}
\description{
Enrich objects of class \code{\link{glm}} with any or all of a set
of auxiliary functions, the maximum likelihood estimate of the
dispersion parameter, the expected or observed information at the
maximum likelihood estimator, and the first term in the expansion
of the bias of the maximum likelihood estimator.
}
\details{
The \code{auxiliary_functions} component consists of any or all of the following functions:
\itemize{
\item \code{score}: the log-likelihood derivatives as a function of the model parameters; see \code{\link{get_score_function.glm}}
\item \code{information}: the expected or observed information as a function of the model parameters; see \code{\link{get_information_function.glm}}
\item \code{bias}: the first-order term in the expansion of the bias of the maximum likelihood estimator as a function of the model parameters; see \code{\link{get_bias_function.glm}}
\item \code{simulate}: a \code{\link{simulate}} function for \code{\link{glm}} objects that can simulate variates from the model at user-supplied parameter values for the regression parameters and the dispersion (default is the maximum likelihood estimates); see \code{\link{get_simulate_function.glm}}
\item \code{dmodel}: computes densities or probability mass functions under the model at user-supplied \code{\link{data.frame}}s and at user-supplied values for the regression parameters and the dispersion, if any (default is at the maximum likelihood estimates); see \code{\link{get_dmodel_function.glm}}
\item \code{pmodel}: computes distribution functions under the model at user-supplied \code{\link{data.frame}}s and at user-supplied values for the regression parameters and the dispersion, if any (default is at the maximum likelihood estimates); see \code{\link{get_pmodel_function.glm}}
\item \code{qmodel}: computes quantile functions under the model at user-supplied \code{\link{data.frame}}s and at user-supplied values for the regression parameters and the dispersion, if any (default is at the maximum likelihood estimates); see \code{\link{get_qmodel_function.glm}}
\item \code{Pmat}: the matrices \eqn{P_t} as a function of the model parameters (see, Kosmidis, 2014, expression (5))
\item \code{Qmat}: the matrices \eqn{Q_t} as a function of the model parameters (see, Kosmidis, 2014, expression (5))
}
}
\examples{

\dontrun{
# A Gamma example, from McCullagh & Nelder (1989, pp. 300-2)
clotting <- data.frame(
   u = c(5,10,15,20,30,40,60,80,100, 5,10,15,20,30,40,60,80,100),
   time = c(118,58,42,35,27,25,21,19,18,69,35,26,21,18,16,13,12,12),
   lot = factor(c(rep(1, 9), rep(2, 9))))
cML <- glm(time ~ lot*log(u), data = clotting, family = Gamma)

# The simulate method for the above fit would simulate at coef(cML)
# for the regression parameters and MASS::gamma.dispersion(cML) for
# the dispersion. It is not possible to simulate at different
# parameter values than those, at least not, without "hacking" the
# cML object.

# A general simulator for cML results via its enrichment with
# auxiliary functions:
cML_functions <- get_auxiliary_functions(cML)
# which is a shorthand for
# enriched_cML <- enrich(cML, with = "auxiliary functions")
# cML_functions <- enriched_cML$auxiliary_functions

# To simulate 2 samples at the maximum likelihood estimator do
dispersion_mle <- MASS::gamma.dispersion(cML)
cML_functions$simulate(coef = coef(cML),
                       dispersion = dispersion_mle,
                       nsim = 2, seed = 123)
# To simulate 5 samples at c(0.1, 0.1, 0, 0) and dispersion 0.2 do
cML_functions$simulate(coef = c(0.1, 0.1, 0, 0),
                       dispersion = 0.2,
                       nsim = 5, seed = 123)

}

\dontrun{

## Reproduce left plot in Figure 4.1 in Kosimdis (2007)
## (see https://ikosmidis.com/files/ikosmidis_thesis.pdf)
mod <- glm(1 ~ 1, weights = 10, family = binomial())
enriched_mod <- enrich(mod, with = "auxiliary functions")
biasfun <- enriched_mod$auxiliary_functions$bias
probabilities <- seq(1e-02, 1 - 1e-02, length = 100)
biases <- Vectorize(biasfun)(qlogis(probabilities))
plot(probabilities, biases, type = "l", ylim = c(-0.5, 0.5),
     xlab = expression(pi), ylab = "first-order bias")
abline(h = 0, lty = 2); abline(v = 0.5, lty = 2)
title("First-order bias of the MLE of the log-odds", sub = "m = 10")
}
}
\references{
Kosmidis I (2014). Bias in parametric estimation: reduction and
useful side-effects. *WIRE Computational Statistics*, **6**,
185-196. \doi{10.1002/wics.1296}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.glm.R
\name{get_auxiliary_functions.glm}
\alias{get_auxiliary_functions.glm}
\title{Function to compute/extract auxiliary functions from objects of
class \code{glm}/\code{enriched_glm}}
\usage{
\method{get_auxiliary_functions}{glm}(object, ...)
}
\arguments{
\item{object}{an object of class \code{glm} or\code{enriched_glm}}

\item{...}{currently not used}
}
\description{
Function to compute/extract auxiliary functions from objects of
class \code{glm}/\code{enriched_glm}
}
\details{
See \code{\link{enrich.glm}} for details.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.link-glm.R
\name{enrich.link-glm}
\alias{enrich.link-glm}
\title{Enrich objects of class \code{\link[=make.link]{link-glm}}}
\usage{
\method{enrich}{`link-glm`}(object, with = "all", ...)
}
\arguments{
\item{object}{an object of class \code{\link[=make.link]{link-glm}}}

\item{with}{a character vector with enrichment options for \code{object}}

\item{...}{extra arguments to be passed to the \code{compute_*}
functions}
}
\value{
The object \code{object} of class \code{\link[=make.link]{link-glm}}
    with extra components. \code{get_enrichment_options.link-glm()}
    returns the components and their descriptions.
}
\description{
Enrich objects of class \code{\link[=make.link]{link-glm}} with
further derivatives of \code{linkinv} with respect to \code{eta}.
}
\details{
The \code{enrich.link-glm} method supports \code{logit},
\code{probit}, \code{cauchit}, \code{cloglog}, \code{identity},
\code{log}, \code{sqrt}, \code{1/mu^2}, \code{inverse}, as well as
the \code{\link{power}} family of links.
}
\examples{

# Example
elogit <- enrich(make.link("logit"), with = "inverse link derivatives")
str(elogit)
elogit$d2mu.deta
elogit$d3mu.deta
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_enrichwith_skeleton.R
\name{create_enrichwith_skeleton}
\alias{create_enrichwith_skeleton}
\title{Create a enrichwith skeleton}
\usage{
create_enrichwith_skeleton(
  class,
  option,
  description,
  component,
  path,
  filename = paste0(class, "_options.R"),
  attempt_rename = TRUE
)
}
\arguments{
\item{class}{the class of the objects to be enriched}

\item{option}{a character vector with components the enrichment
options}

\item{description}{a character vector of length
\code{length(options)} with components the description of the
enrichment options}

\item{component}{a list of as many character vectors as
\code{length(option)}, specifying the names of the components
that each option will add to the object after enrichment}

\item{path}{the path where the skeleton file will be created}

\item{filename}{the name of the skeleton file}

\item{attempt_rename}{attempt to rename syntactically incorrect
component names? Default is \code{TRUE}}
}
\value{
A file with the necessary functions to use
    \code{enrichwith} infrastructure. The skeleton consists of the
    following functions
\itemize{

\item One \code{compute_component.class} function per component
name from \code{unique(unlist(component))}. The function takes as
input the object to be enriched and returns as output the component
to be added to the object.

\item The \code{get_enrichment_options.class} function, that takes
as input the object to be enriched and an enrichment option, and
returns the names of the components that will be appended to the
object for this option. This function can also be used to list the
available options and print their description.

\item The \code{enrich.class} function

}
}
\description{
Create an enrichwith skeleton file for the structured
implementation of methods to compute new components for objects of
a specific class
}
\examples{
\dontrun{
# Set the directory where the skeleton is placed
my_path <- "~/Downloads"
# This is the call that created the enrichment skeleton for glms
# that ships with the package
create_enrichwith_skeleton(class = "glm",
      option = c("auxiliary functions", "score vector",
       "mle of dispersion", "expected information",
       "observed information", "first-order bias"),
      description = c("various likelihood-based quantities
        (gradient of the log-likelihood, expected and observed
        information matrix and first term in the expansion of
        the bias of the mle) and a simulate method as functions
        of the model parameters",
       "gradient of the log-likelihood at the mle",
       "mle of the dispersion parameter",
       "expected information matrix evaluated at the mle",
       "observed information matrix evaluated at the mle",
       "first term in the expansion of the bias of the mle
        at the mle"),
      component = list("auxiliary_functions", "score_mle",
                       "dispersion_mle",
                       "expected_information_mle",
                       "observed_information_mle",
                       "bias_mle"),
      path = my_path,
      attempt_rename = FALSE)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich-package.R
\docType{package}
\name{enrichwith}
\alias{enrichwith-package}
\alias{enrichwith}
\title{Methods to enrich list-like R objects with extra components}
\description{
The enrichwith package provides the \code{\link{enrich}} method to
enrich list-like R objects with new, relevant components. The
resulting objects preserve their class, so all methods associated
with them still apply. The package can also be used to produce
customizable source code templates for the structured
implementation of methods to compute new components
}
\details{
Depending on the object, enriching it can be a tedious task. The
\pkg{enrichwith} package aims to streamline the task into 3 simple
steps:

\enumerate{

\item Use \code{\link{create_enrichwith_skeleton}} to produce a
customizable enrichwith template.

\item Edit the \code{compute_*} functions by adding the specific
code that calculates the components.

\item Finalize the documentation and/or include more examples.
}

The first step results in a template that includes all necessary
functions to carry out the enrichment. The second step is where the
user edits the template and implements the calculation of the
components that the object will be enriched with. Specifically,
each \code{compute_*} function takes as input the object to be
enriched and returns the corresponding new component to be added to
the object.

Everything else (for example, mapping between the enrichment
options and the components that the enriched object will have,
checks that an enrichment option exists, listing enrichment
options, enriching the object, and so on) is taken care of by the
methods in \pkg{enrichwith}.

Developers can either put their enrichwith templates in their
packages or are welcome to contribute their template to enrichwith,
particularly if that extends core R objects.
}
\seealso{
\code{\link{enrich.glm}} and \code{\link{enriched_glm}}, \code{\link{enrich.family}}, \code{\link{enrich.link-glm}}, \code{\link{enrich.betareg}}
}
\author{
\strong{Maintainer}: Ioannis Kosmidis \email{ioannis.kosmidis@warwick.ac.uk}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enriched_glm.R
\name{enriched_glm}
\alias{enriched_glm}
\title{Fitting generalized linear models enriched with useful components}
\usage{
enriched_glm(formula, family = gaussian, ...)
}
\arguments{
\item{formula}{an object of class \code{"\link[stats]{formula}"} (or one that
    can be coerced to that class): a symbolic description of the
    model to be fitted.  The details of model specification are given
    under \sQuote{Details}.}

\item{family}{a description of the error distribution and link
    function to be used in the model.  For \code{glm} this can be a
    character string naming a family function, a family function or the
    result of a call to a family function.  For \code{glm.fit} only the
    third option is supported.  (See \code{\link[stats]{family}} for details of
    family functions.)}

\item{...}{other arguments passed to \code{\link{glm}}}
}
\value{
An object of class \code{enriched_glm} that contains all the
components of a \code{\link{glm}} object, along with a set of
auxiliary functions (score function, information matrix, a simulate
method, first term in the expansion of the bias of the maximum
likelihood estimator, and dmodel, pmodel, qmodel), the maximum
likelihood estimate of the dispersion parameter, the expected or
observed information at the maximum likelihood estimator, and the
first term in the expansion of the bias of the maximum likelihood
estimator.

See \code{\link{enrich.glm}} for more details and links for the
auxiliary functions.
}
\description{
\code{\link{enriched_glm}} fits generalized linear models using
\code{\link{glm}} and then enriches the resulting object with all
enrichment options.
}
\details{
\code{enriched_glm} has the same interface as \code{\link{glm}}
}
\examples{
\dontrun{
# A Gamma example, from McCullagh & Nelder (1989, pp. 300-2)
clotting <- data.frame(
   u = c(5,10,15,20,30,40,60,80,100, 5,10,15,20,30,40,60,80,100),
   time = c(118,58,42,35,27,25,21,19,18,69,35,26,21,18,16,13,12,12),
   lot = factor(c(rep(1, 9), rep(2, 9))))

# Fit a generalized linear model
cML <- enriched_glm(time ~ lot*log(u), data = clotting, family = Gamma("log"))

# Evaluate the densities at the data points in clotting at the
# maximum likelihood estimates
cML_dmodel <- get_dmodel_function(cML) # same as cML$auxiliary_functions$dmodel
cML_dmodel()

# Evaluate the densities at supplied data points
new_data <- data.frame(u = c(15:17, 15:17),
                       time = c(30:32, 15:17),
                       lot = factor(c(1, 1, 1, 2, 2, 2)))
cML_dmodel(data = new_data)

# Get pmodel and qmodel function
cML_pmodel <- get_pmodel_function(cML) # same as cML$auxiliary_functions$pmodel
cML_qmodel <- get_qmodel_function(cML) # same as cML$auxiliary_functions$qmodel

# The following should return c(30:32, 15:17)
probs <- cML_pmodel(data = new_data)
cML_qmodel(probs, data = new_data)

# Evaluate the observed information matrix at the MLE
cML_info <- get_information_function(cML)
cML_info(type = "observed")

# Wald tests based on the observed information at the
# moment based esimator of the dispersion
dispersion <- summary(cML)$dispersion
cML_vcov_observed <- solve(cML_info(dispersion = dispersion, type = "observed"))
lmtest::coeftest(cML, vcov = cML_vcov_observed)

# Wald tests based on the expected information at the
# moment based esimator of the dispersion
cML_vcov_expected <- solve(cML_info(dispersion = dispersion, type = "expected"))
lmtest::coeftest(cML, vcov = cML_vcov_expected)
# Same statistics as coef(summary(cML))[, "t value"]

}


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.betareg.R
\name{get_bias_function.betareg}
\alias{get_bias_function.betareg}
\title{Function to compute/extract a function that returns the first term
in the expansion of the bias of the MLE for the parameters of an
object of class \code{betareg}/\code{enriched_betareg}}
\usage{
\method{get_bias_function}{betareg}(object, ...)
}
\arguments{
\item{object}{an object of class \code{betareg} or\code{enriched_betareg}}

\item{...}{currently not used}
}
\description{
Function to compute/extract a function that returns the first term
in the expansion of the bias of the MLE for the parameters of an
object of class \code{betareg}/\code{enriched_betareg}
}
\details{
The computed/extracted function has arguments
\describe{

\item{coefficients}{the regression coefficients at which the
first-order bias is evacuated. If missing then the maximum
likelihood estimates are used}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.lm.R
\name{get_bias_function.lm}
\alias{get_bias_function.lm}
\title{Function to compute/extract a function that returns the first term
in the expansion of the bias of the MLE for the parameters of an
object of class \code{lm}/\code{enriched_lm}}
\usage{
\method{get_bias_function}{lm}(object, ...)
}
\arguments{
\item{object}{an object of class \code{lm} or\code{enriched_lm}}

\item{...}{currently not used}
}
\description{
Function to compute/extract a function that returns the first term
in the expansion of the bias of the MLE for the parameters of an
object of class \code{lm}/\code{enriched_lm}
}
\details{
The computed/extracted function has arguments
\describe{

\item{coefficients}{the regression coefficients at which the
first-order bias is evacuated. If missing then the maximum
likelihood estimates are used}

\item{dispersion}{the dispersion parameter at which the first-order
bias is evaluated. If missing then the maximum likelihood estimate
is used}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.lm.R
\name{get_auxiliary_functions.lm}
\alias{get_auxiliary_functions.lm}
\title{Function to compute/extract auxiliary functions from objects of
class \code{lm}/\code{enriched_lm}}
\usage{
\method{get_auxiliary_functions}{lm}(object, ...)
}
\arguments{
\item{object}{an object of class \code{lm} or\code{enriched_lm}}

\item{...}{currently not used}
}
\description{
Function to compute/extract auxiliary functions from objects of
class \code{lm}/\code{enriched_lm}
}
\details{
See \code{\link{enrich.lm}} for details.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/enrich.glm.R
\name{get_qmodel_function.glm}
\alias{get_qmodel_function.glm}
\title{Function to compute/extract a \code{qmodel} function}
\usage{
\method{get_qmodel_function}{glm}(object, ...)
}
\arguments{
\item{object}{an object of class \code{glm} or\code{enriched_glm}}

\item{...}{currently not used}
}
\description{
Function to compute/extract a \code{qmodel} function
}
\details{
The computed/extracted function has arguments
\describe{

\item{p}{a vector of probabilities with \code{length(p)} equal to
\code{nrow(data)} at which to evaluate quantiles}

\item{data}{a data frame with observations at which to compute the
quantiles. If missing then quantiles are computed at the model
frame extracted from the object (see \code{\link{glm}})}

\item{coefficients}{the regression coefficients at which the
quantiles are computed. If missing then the maximum likelihood
estimates are used}

\item{dispersion}{the dispersion parameter at which the
quantiles are computed. If missing then the maximum
likelihood estimate is used}

\item{log.p}{logical; if \code{TRUE}, the logarithm of the
probabilities is used}

\item{lower.tail}{logical; if \code{TRUE} (default), probabilities
are P[X <= x] otherwise, P[X > x]}

}
}
